require 'notebook_core.rb'

class Document < NotebookObject
	attr_accessor :page_list, :name

	def Document.current
		Document.new
	end
	
	def delay_draw!
	  Document.current.draw!

	  #StoppableThread.new do
	  #  sleep 0.15
	  #  Document.current.draw!
	  #end
	end
end

class Page < NotebookObject
	attr_accessor :parent, :annotation_list
end

class Annotation < NotebookObject
  attr_accessor :annotation_id
  attr_accessor :guid
  attr_accessor :demoing_animation
	def initialize(annotation_id = -1)
		super
		@size = Size.new
		@position = Point.new
		@transparency = 1.0
		@annotation_id = annotation_id
		@guid = ""
		@onclick_procedure = nil
		@xml = ""
		@demoing_animation = false
		
		@saved_attributes = {}
	end

	def offset_position distance
		pos = self.position
		pos.x = pos.x + distance.x
		pos.y = pos.y + distance.y
		self.position = pos
	end

  def scale(s)
    self.size = [self.size.width * s, self.size.height * s]
  end
   
  def scale_in_place(s)
    pos = self.position
    pos.x += self.size.width / 2.0 * (1 - s)
    pos.y += self.size.height / 2.0 * (1 - s)
    self.scale(s)
  end
  
  def do_animation
    if self.properties.has_key? "animation_type"
      if not self.properties["animation_type"].empty?
	  
        animation_name = self.properties["animation_type"]
        ObjectAnimationFactory.animation(animation_name).animate(self)
      end
    end
  end
  
  def do_delayed_animation
#	StoppableThread.new do
#		sleep 1.0
		self.do_animation
#	end
  end
  
  def demo_animation
    if self.properties.has_key? "animation_type"
      if not self.properties["animation_type"].empty?
        animation_name = self.properties["animation_type"]
        @demoing_animation = true
		
        case self.properties["animation_type"]
		when "Fly Out"
		  # Need to make sure we move the annotation back to it's original position
		  # before we demo the animation
		  self.position.x = self.properties["original_positionX"]
		  self.position.y = self.properties["original_positionY"]
        when "Fade In"
          self.transparency = 0.02
        when "Fade Out"
          self.transparency = 1.0
		when "Shrink and Grow"
		  self.set_property({"animation_options" => 2})
        end
	
		#BUGZID 1139309, need to stop the annotation if it is alread animating. 
		# for example, if last selection is spin, forever. Need to stop first
		if ObjectAnimation.already_animating? self
			#already animating, check if we need to stop
			if self.properties.has_key? "animation_repeats"
				ObjectAnimation.stop_animation_for_annotation(self)
				did_stop = true
			end
		end
 	
		#BUGZID: 1151962
		return if self.properties["animation_repeats"] == -2

		ObjectAnimationFactory.animation(animation_name).animate(self)
      end
    end
  end
  
  attr_accessor :parent, :size, :position, :annotation_id
end

class Annotation
	def Annotation.create_from_xml(xml)
		ann = Annotation.new
		return ann.from_xml(xml)
	end
end

class Line < Annotation
	def initialize
	  super
	  @points = Array.new
	  self.create_c_object
	end

	def add_point point
		@points << point if point.is_a? Array and point.size == 2
		self.update
	end

	def points
		@points
	end
end

class Ink < Annotation
	def initialize
	  super
	  @points = Array.new
	  self.create_c_object
	end

	def add_point point
		@points << point if point.is_a? Array and point.size == 2
		self.update
	end

	def points
		@points
	end
end

class CalligraphicLine < Annotation
	def initialize
	  super
	  @points = Array.new
	  self.create_c_object
	end

	def add_point point
		@points << point if point.is_a? Array and point.size == 2
		self.update
	end

	def points
		@points
	end
end

class Shape < Annotation
end

class Text < Annotation
	def initialize text=nil
		super
		self.create_c_object
		@text = ""
		@text = text.to_s if text.respond_to? :to_s
	end

	def text= text
		@text = text.to_s if text.respond_to? :to_s
		self.update
	end
	
	def text
		@text
	end
end

class Image < Annotation
	def initialize(path_to_image)
		super
		@path = path_to_image
		self.create_c_object
	end
end

class App
  include Singleton
  
	def App.add_menu (title, callback)
    App.instance().add_menu title, callback
	end

	def App.alert (msg)
		App.instance().alert msg
	end
end

#class Runner
#	def Runner.run
#		eval($notebook_script) if $notebook_script and $notebook_script.respond_to? :to_s
#	end
#end

if __FILE__ == $0
	#ann = Line.new
	#print ann.size
end
