
# May move these to animator namespace
# Event.add_observer(:annotation_clicked, Proc.new {|args| slide_from_side_animation args[:annotation]})

debug_puts "Loading notebook_animation.rb... "

require "rexml/document"
include REXML


require "notebook_core.rb"


DEBUG_ON = false
LOOP_DEBUG = false

if defined? c_debugger_on and c_debugger_on()
  DEBUG_ON = true
  LOOP_DEBUG = true

  if DEBUG_ON 
	begin
	  require "rubygems"
	  require "ruby-debug"
	  Debugger.start_server
	  Debugger.start
	  puts "Debug enabled"
	  LOOP_DEBUG = true
  	rescue
	  puts "Unable to load rdebug"
	end
  end
end

class StoppableThread
  def StoppableThread.initialize
    @@current_threads = []
    @@curr_thread = nil
  end
  
  def StoppableThread.stop_all_threads
    @@current_threads.each do |t|
	  t.reset_object_animation	# reset the animation if the running object is an ObjectAnimation
	  t.kill					# set the kill status to the thread. it's used to detect a terminated thread.
    end
    @@current_threads = []
  end
  
  def StoppableThread.threads_run_once
	@@current_threads.delete_if{|t| t==nil or t.should_kill?}
	
	@@current_threads.each do |t| 
		if not t.should_pause?
			t.add_time_slice(0.04) #the timer is set to 40 ms in foreground of notebook
		end
	end
	
	should_process = true
	begin_time = Time.now.to_f
	while should_process
		should_process = false
		@@current_threads.delete_if{|t| t==nil or t.should_kill?}
		i = 0
		@@current_threads.each do |t|
			if t.should_pause?
				next
			end

			if t.get_time_slice <= 0
				next
			end

			i += 1
			debug_puts "running thread: " + i.to_s
			@@curr_thread = t
			if t.should_stop?
				t.kill
			end
			if t.run_once == false
				t.kill
			else
				if t.get_time_slice >0
					should_process = true
				end
			end
		end
		@@curr_thread = nil
		
		if Time.now.to_f - begin_time > 0.015 # we don't want the notebook timer to run too long.
			should_process = false
		end
	end
	
  end
  
  def StoppableThread.current
	return @@curr_thread
  end
  
  def initialize
    @stop_myself = false
    @pause_myself = false
    @@current_threads << self
    @running_object = nil
    @kill_myself = false
    @time_slice = 0.0
  end
  
  def sleep(sec)
	@time_slice -= sec
  end
  
  def get_time_slice
	return @time_slice
  end
  
  def add_time_slice(sec)
	@time_slice += sec
  end
  def spend_time_slice(sec)
	@time_slice -= sec
  end
  
  # the thread will be killed if this funtion return false
  def run_once
	if @running_object == nil
		debug_puts "running object == nil"
		return false
	end
	if not @running_object.is_valid #check if the annotation had been deleted in notebook
		debug_puts "kill the invalid thread..."
		return false
	end
	
	return @running_object.run_once
  end
  
  def set_running_object(o)
	if o == nil or o.is_a? Animation or o.is_a? ObjectAnimation
		@running_object = o
	end
  end
  def get_running_object
	return @running_object
  end
  def set_stop_thread
	@stop_myself = true
  end
  
  def stop_thread
    @stop_myself = true
  end
 
  def pause_thread
    @pause_myself = true
  end

  def should_stop?
    return @stop_myself
  end
 
  def should_pause?
    return @pause_myself
  end

  def wakeup
    @pause_myself = false
  end
  
  def kill
	if not @running_object.is_a? ObjectAnimation # page recording restore page once recording is finished
		Document.current.current_page.restore_delta
	end
	@kill_myself = true
  end

  def reset_object_animation
	if @running_object.is_a? ObjectAnimation
		@running_object.reset
	end
  end

  def should_kill?
	return @kill_myself
  end
  
  # implementation of the function in ruby Thread class and it is different with the set_stop_thread function.
  def stop
	@pause_myself = true
  end
end

StoppableThread.initialize


# Stores an animation
class Animation
  attr_accessor :animation_steps, :original_state, :original_bgcolor, :original_state_order, :last_event_time, :animation_created_objects
  
  def initialize
    @animation_steps = []
    @animation_created_objects = {}
    @last_event_time = 0
    @original_state = nil
	@original_state_order = nil
    @original_bgcolor = [1.0, 1.0, 1.0]
    
    #args for play
    @counter = 0
    @document = nil
    @current_page = nil
    @current_pen_annotations = {}
    @current_pen_annotation_initialx = {}
    @current_pen_annotation_initialy = {}
    @rect_width_to_height_ratio = {}
  end
  
  def set_initial_state(state, bgcolor, state_order)
    @original_state = state
    @original_bgcolor = bgcolor
	@original_state_order = state_order
  end
  
  def add_event(e)
    #key = e[:annotation].annotation_id
    @animation_steps.last[:next_delay] = Time.now.to_f - @last_event_time unless @animation_steps.empty?
    @last_event_time = Time.now.to_f
    #@animation_steps[key] = [] unless @animation_steps.has_key? key
    #@animation_steps[key] << {:x => e[:x], :y => e[:y], :guid => e[:guid]}
    @animation_steps << e
  end
  
  def rewind
	Document.current.current_page.backup_before_rewind
    Animator.restore_state(@original_state, @original_bgcolor, @original_state_order)
	#then after rewind, there is no backup unless you modify something the page again
	
  end
   
   def is_valid
	return true
   end
  def play
  	#KVZ 2010-NOV-29 Added to fix bug#1222176
	Document.current.current_page.backup_page

    #save original positions (still might need this!)
    self.rewind
    
    #original_state.each do |k, v|
    #   annotation = Annotation.new(k)
    #   annotation.position = v
    #end
    
    @document = Document.current
    @current_page = Document.current.current_page
    @counter = 0
    @current_pen_annotations = {}
   	@current_pen_annotation_initialx = {}
   	@current_pen_annotation_initialy = {}

    #begin
    #  if animation_created_objects.has_key? nil
    #    debugger if DEBUG_ON
    #    xml = animation_created_objects[nil][:xml]
    #    if xml.include? "<defs"
    #      annotation = Annotation.create_from_xml(xml)
    #    end
    #  end
    #rescue Exception => e
    #end
    
    a_play_thread = StoppableThread.new
    a_play_thread.set_running_object(self)
    return a_play_thread
    end
    
    
    # if the run_once function return false the thread will be terminated.
    def run_once
        if @counter >= @animation_steps.size
			StoppableThread.current.set_stop_thread
			return false
        end
        index = @counter % @animation_steps.size
        
        debugger if DEBUG_ON and LOOP_DEBUG
        
      #  if StoppableThread.current.should_stop?
      #    Thread.stop
      #  end
       
	#	if StoppableThread.current.should_pause?
	#	  Thread.stop
	#	end

        debug_puts "checking c_address"
        if Document.current.current_page.c_address != @current_page.c_address
          debug_puts "changed page, STOP NOW"
        #  Thread.kill(self)
			return false
        end
        
        debug_puts "done."
        
        debug_puts "playing.."
        #annotation = @animation_steps[index][:annotation]
        #annotation = Annotation.new(@animation_steps[index][:annotation_id])
        annotation = Annotation.new
		    annotation.guid = @animation_steps[index][:guid]
		    midgroup_pen_strokes = []

		    debug_puts @animation_steps[index][:type]
		guid = @animation_steps[index][:guid]		  
		
		

        case @animation_steps[index][:type]
        when "moved"
          annotation.position = [@animation_steps[index][:x], @animation_steps[index][:y]] 
        when "rotated"
          annotation.rotation = @animation_steps[index][:angle]
        when "sized"
          annotation.size = [@animation_steps[index][:w], @animation_steps[index][:h]]
        when "pen_creation"
          if not @current_pen_annotations.has_key? guid
            debug_puts "creating faux annotation"          
            debug_puts "scanning for created annotation"
          if guid == "annotation.empty"
			debug_puts "annotation.empty---------------------------old scripts"
			j = 0;
			for i in index..(@animation_steps.size - 1) do  
				if @animation_steps[i][:type] == "pen_creation_finished"
					guid =  @animation_steps[i][:guid]
					j = i
					debug_puts "convert annotation.empty to a valid guid"
					break
				end
			end
			for i in index..j do  
				animation_steps[i][:guid] = guid
			end
		  end
            for i in index..(@animation_steps.size - 1) do
              debug_puts "scan..." + i.to_s
              if @animation_steps[i][:type] == "pen_creation_finished" and @animation_steps[i][:guid] == guid
                debug_puts "found pen_creation_finished"

                  debug_puts "found, creating"
                  xml = animation_created_objects[guid][:xml]
                  annotation = Annotation.create_from_xml(xml)
                  if annotation.nil?
                    debug_puts "bad annotation"
                    break
                  end
				  if annotation.is_a? CalligraphicLine
					debug_puts "its a calligraphic line, setting properties"
					@current_pen_annotations[guid] = CalligraphicLine.new
					@current_pen_annotations[guid].line_color = annotation.line_color
                    @current_pen_annotations[guid].line_thickness = annotation.line_thickness
                    @current_pen_annotations[guid].transparency = annotation.transparency
					debug_puts "ok!"
                    @current_pen_annotations[guid].add_point [@animation_steps[index][:x], @animation_steps[index][:y]]
				  elsif annotation.is_a? Line
                    debug_puts "its a line, setting properties"
                    @current_pen_annotations[guid] = Line.new
                    @current_pen_annotations[guid].line_color = annotation.line_color
                    @current_pen_annotations[guid].fill_color = annotation.fill_color
                    @current_pen_annotations[guid].line_thickness = annotation.line_thickness
					@current_pen_annotations[guid].line_dashstyle = annotation.line_dashstyle
                    @current_pen_annotations[guid].transparency = annotation.transparency
                    debug_puts "ok!"
                    @current_pen_annotations[guid].add_point [@animation_steps[index][:x], @animation_steps[index][:y]]
                  elsif annotation.is_a? Shape
                    debug_puts "its a shape, create"
                    @current_pen_annotations[guid] = annotation
                    @rect_width_to_height_ratio[guid] = annotation.size.width/annotation.size.height
                    @current_pen_annotations[guid].position.x = @animation_steps[index][:x]
                    @current_pen_annotations[guid].position.y = @animation_steps[index][:y]
					@current_pen_annotation_initialx[guid] = @animation_steps[index][:x] 
					@current_pen_annotation_initialy[guid] = @animation_steps[index][:y]
                    @current_pen_annotations[guid].size.width = 1
                    @current_pen_annotations[guid].size.height = 1
                    debug_puts "set xy and width height ok!"
                  elsif annotation.is_a? StraightLine
					debug_puts "its a straightline"
					@current_pen_annotations[guid] = annotation
                    @current_pen_annotations[guid].position.x = @animation_steps[index][:x]
                    @current_pen_annotations[guid].position.y = @animation_steps[index][:y]
					@current_pen_annotation_initialx[guid] = @animation_steps[index][:x] 
					@current_pen_annotation_initialy[guid] = @animation_steps[index][:y]
                    @current_pen_annotations[guid].size.width = 1
                    @current_pen_annotations[guid].size.height = 1
				  elsif annotation.respond_to? :clone_empty
                    debug_puts "responds to - clone empty"
                    @current_pen_annotations[guid] = annotation.clone_empty
				  else
				    debug_puts "Unsupported annotation type!!!"
					@current_pen_annotations[guid] = nil
                  end
                break
              end
            end
            debug_puts "finished scanning"
		# Could be a group object, don't  make a default line
        #   if not @current_pen_annotations.has_key? guid
        #     debug_puts "couldn't find pen_creation_finished (!!), making default line"
		#	  debug_puts animation_created_objects[@animation_steps[i][:guid]].inspect
		#	  debug_puts @animation_steps[index].inspect
        #      @current_pen_annotations[guid] = Line.new
        #    end
		    if @current_pen_annotations.has_key? guid
              debug_puts "adding"
              debugger if DEBUG_ON
              @document.current_page.add_annotation @current_pen_annotations[guid]
			else
			  debug_puts "Didn't find"
			  debug_puts @animation_steps[index].inspect
			end
          else
            debug_puts "updating"
			debug_puts @current_pen_annotations[guid].inspect
            if @current_pen_annotations[guid].is_a? Line or @current_pen_annotations[guid].is_a? Ink
              @current_pen_annotations[guid].add_point [@animation_steps[index][:x], @animation_steps[index][:y]]
              debug_puts "updated line or ink"
            end
			if @current_pen_annotations[guid].is_a? CalligraphicLine
              @current_pen_annotations[guid].add_point [@animation_steps[index][:x], @animation_steps[index][:y]]
              debug_puts "updated calligraphic line"
            end
            if @current_pen_annotations[guid].is_a? Shape
				if @current_pen_annotations[guid].is_circle
					r = ((@animation_steps[index][:x] - @current_pen_annotation_initialx[guid])**2 + (@animation_steps[index][:y] - @current_pen_annotation_initialy[guid])**2)**0.5
					@current_pen_annotations[guid].position.x = @current_pen_annotation_initialx[guid] - r
					@current_pen_annotations[guid].position.y = @current_pen_annotation_initialy[guid] - r
					@current_pen_annotations[guid].size.width = r*2
					@current_pen_annotations[guid].size.height =r*2
				elsif @current_pen_annotations[guid].is_always_constrain
					width = @animation_steps[index][:x] - @current_pen_annotation_initialx[guid]
					height = @animation_steps[index][:y] - @current_pen_annotation_initialy[guid]
					if width < 0
						width = -width
						@current_pen_annotations[guid].position.x = @current_pen_annotation_initialx[guid] - width
					else
						@current_pen_annotations[guid].position.x = @current_pen_annotation_initialx[guid];
					end
					if height < 0
						@current_pen_annotations[guid].position.y = @current_pen_annotation_initialy[guid] - width/@rect_width_to_height_ratio[guid]
					else
						@current_pen_annotations[guid].position.y = @current_pen_annotation_initialy[guid]
					end
					@current_pen_annotations[guid].size.width = width
					@current_pen_annotations[guid].size.height = width/@rect_width_to_height_ratio[guid]
				else
					@current_pen_annotations[guid].size.width = @animation_steps[index][:x] - @current_pen_annotation_initialx[guid]
					@current_pen_annotations[guid].size.height =   @animation_steps[index][:y] - @current_pen_annotation_initialy[guid] 
					if @current_pen_annotations[guid].size.width < 0
						@current_pen_annotations[guid].position.x = @current_pen_annotation_initialx[guid] + @current_pen_annotations[guid].size.width
						@current_pen_annotations[guid].size.width = @current_pen_annotations[guid].size.width * -1
					end
					if @current_pen_annotations[guid].size.height < 0
						@current_pen_annotations[guid].position.y = @current_pen_annotation_initialy[guid] + @current_pen_annotations[guid].size.height
						@current_pen_annotations[guid].size.height = @current_pen_annotations[guid].size.height * -1
					end
				end
					debug_puts "updated shape"
            end
            if @current_pen_annotations[guid].is_a? StraightLine
				width = @animation_steps[index][:x] - @current_pen_annotation_initialx[guid]
				height =   @animation_steps[index][:y] - @current_pen_annotation_initialy[guid]
				left = 0
				top = 0
				sl_ends = 1
				if width < 0 && height < 0
					sl_ends = 4
				elsif width < 0 && height > 0
					sl_ends = 2
				elsif width > 0 && height < 0
					sl_ends = 3
				else
					sl_ends = 1
				end
				if width < 0
					width = width * -1
					left = @current_pen_annotation_initialx[guid]-width
				else
					left = @current_pen_annotation_initialx[guid]
				end
				if height < 0
					height = height * -1
					top = @current_pen_annotation_initialy[guid]-height
				else
					top = @current_pen_annotation_initialy[guid]
				end
				@current_pen_annotations[guid].position.x = left
				@current_pen_annotations[guid].position.y = top
				@current_pen_annotations[guid].size.width = width
				@current_pen_annotations[guid].size.height = height
				@current_pen_annotations[guid].straightlineends = sl_ends
				debug_puts "updated straightline"
			end
          end
        when "pen_creation_midgroup"
          if not @current_pen_annotations.has_key? guid
            debug_puts "midgroup gone bad"
          else
            debug_puts "midgroup lifting.."
            #midgroup_pen_strokes << current_pen_annotations[guid]
            #current_pen_annotations[guid] = nil
          end        
        when "pen_creation_finished"
          debugger if DEBUG_ON

          if not @current_pen_annotations.has_key? guid
            debug_puts "recording gone bad"
            debug_puts "adding non pen object"
          else
		    debug_puts "delete_annotation when pen_creation_fnished"
            @document.current_page.delete_annotation @current_pen_annotations[guid]
			@current_pen_annotations.delete(guid)
          end
          #replace
			guid = @animation_steps[index][:guid]  
         
          if animation_created_objects.has_key? guid
            xml = animation_created_objects[guid][:xml]
            debug_puts xml
            annotation = Annotation.create_from_xml(xml)
			debug_puts "-"
			debug_puts "replace add_annoattion"
            debug_puts annotation.inspect unless annotation.nil?
            debugger if DEBUG_ON
            @document.current_page.add_annotation annotation unless annotation.nil?    
		  else
			debug_puts "no replace"
		  end
        when "clicked"
          annotation.do_animation
        when "deleted"
		  debug_puts "delete annotation when deleted"
          @document.current_page.delete_annotation annotation
        when "property_changed"
          case @animation_steps[index][:property]
          when "color"
            debug_puts "changing color to: " + animation_steps[index][:value].inspect + "!!"
            color = []
            animation_steps[index][:value].split(',').each do |color_string|
              color << color_string.to_f
            end
            debug_puts "color arry: " + color.inspect
            annotation.line_color = color
            debug_puts "ok!!!"
          when "fill_color"
            debug_puts "changing fill color to: " + animation_steps[index][:value].inspect + "!!"
             color = []
             animation_steps[index][:value].split(',').each do |color_string|
               color << color_string.to_f
             end
             debug_puts "fill color arry: " + color.inspect
             annotation.fill_color = color
             debug_puts "ok!?!"
          when "txt_image"
            debug_puts "value is nil " if @animation_steps[index][:value].nil?             
            debug_puts "in text image" + @animation_steps[index][:value].to_s      
            annotation.txt_image = @animation_steps[index][:value].to_s
            debug_puts "ok!?!"
          when "transparency"
            debug_puts "changing trans to " + @animation_steps[index][:value].to_s + "!"
            annotation.transparency = @animation_steps[index][:value].to_f
            debug_puts "ok!"
          when "width"
            debug_puts "changing width to " + @animation_steps[index][:value].to_s + "!"
            annotation.line_thickness = @animation_steps[index][:value].to_f
            debug_puts "ok!!"
		  when "mirrored_horizontal"
		  	debug_puts "flipping horizontal"
			annotation.set_property({"mirrored_horizontal" => @animation_steps[index][:value] == "true"})
			debug_puts "ok!"
		  when "mirrored_vertical"
		  	debug_puts "flipping vertical"
			annotation.set_property({"mirrored_vertical" => @animation_steps[index][:value] == "true"})
			debug_puts "ok!"
          end
        when "background_changed"
          debug_puts "changing background color"
          color = []
          @animation_steps[index][:value].split(',').each do |color_string|
            color << color_string.to_f
          end
          debug_puts "bg color arry: " + color.inspect          
          @document.current_page.background = color
          debug_puts "ok!"
		when "annotations_shuffled"
		  debug_puts "annotation shuffled"
		  guids = []
		  @animation_steps[index][:value].split(',').each do |guid|
		  	guids << guid
		  end
		  @document.current_page.put_annotations_order(guids)
		  debug_puts "ok!"
        else
          return true
        end
        
        @current_pen_annotations[guid] = nil unless ["pen_creation", "pen_creation_finished", "pen_creation_midgroup"].include? @animation_steps[index][:type]
     
        #debug_puts "Offsetting by #{annotation.inspect} + #{offset.inspect}"
        debug_puts "moving annotation to + " + [@animation_steps[index][:x], @animation_steps[index][:y]].inspect
        debug_puts "-"
        @document.delay_draw!
        StoppableThread.current.sleep @animation_steps[index][:next_delay] if @animation_steps[index].has_key? :next_delay and (index + 1) < @animation_steps.size
        @document.delay_draw!
        @counter += 1
        return true
      end
#      Document.current.delay_draw!
#      debug_puts "finished playing"
#      GC.enable
#    end
    
#    return a_play_thread
#  end
  
end

# Responsible for munging the events
class Animator
  #include Singleton
    
  def initialize
    @is_recording = false
    @recordings = []
    @original_state = {}
	@original_state_order = []
    @original_bgcolor = [1.0, 1.0, 1.0]
    @newly_added_annotations = {}
    
    #event munger
    @last_event_time = Time.now.to_f
    @last_args_before_event_time = nil
  end

  def is_recording
    return @is_recording
  end
  
  def start_recording
    return nil if @is_recording
    animation = Animation.new
     
    @is_recording = true
    debugger if DEBUG_ON
    state_info = Animator.save_state
    @original_state = state_info[:state]
    @original_bgcolor = state_info[:bgcolor]
	@original_state_order = state_info[:order]
    animation.set_initial_state(@original_state, @original_bgcolor, @original_state_order)
    @recordings << animation
  end
  
  def stop_recording
    return nil unless @is_recording
    @recordings.last.add_event(@last_args_before_event_time) unless @last_args_before_event_time.nil? 
    @is_recording = false
    
    # Store the newly created objects
    @recordings.last.animation_created_objects = @newly_added_annotations
    
  	#KVZ 2010-NOV-29 Added to fix bug#1222176
	Document.current.current_page.backup_page
    #Animator.restore_state(@original_state, @original_bgcolor, @original_state_order)
  end
  
  def Animator.save_state
    #save bg color
    color = Document.current.current_page.background
    #save current positions
    state = {}
	order = []
    Document.current.current_page.annotations.each do |annotation|
      #key = annotation.annotation_id
      key = annotation.properties["guid"]
  	  if not state.has_key? key
        #state[key] = {:position => [annotation.position.x, annotation.position.y], :rotation => annotation.rotation, :size => [annotation.size.width, annotation.size.height]}
        state[key] = {:xml => annotation.xml}
		order << key
      end
    end
    return {:state => state, :bgcolor => color, :order => order}
  end
  
  def Animator.restore_state(state, bgcolor, order)
    #restore positions
    #Document.current.current_page.annotations.each do |annotation|
    #  Document.current.current_page.delete_annotation annotation
    #end
    
    Document.current.current_page.background = bgcolor
    
    Document.current.current_page.clear_page!
    
    order.each do |k|
      xml = state[k][:xml]
      debug_puts "restoring state with xml"
      debug_puts xml
      annotation = Annotation.create_from_xml(xml)
      debugger if DEBUG_ON
      Document.current.current_page.add_annotation annotation unless annotation.nil?
    end
    
    Document.current.delay_draw!
  end
  
  def get_animations
    @recordings
  end
  
  def get_last_animation
    return @recordings.last
  end
  
  def add_recording_event(args)
    return if not @is_recording
    return if args.nil?
    
    if not ["moved", "rotated", "sized"].include?(args[:type])
      debug_puts "recording non moving event: " + args.inspect
      @recordings.last.add_event(args) unless @recordings.last.nil?
      return
    else
        if not @last_args_before_event_time.nil?
          if @last_args_before_event_time[:guid] != args[:guid] or @last_args_before_event_time[:type] != args[:type]  
            @recordings.last.add_event(@last_args_before_event_time) unless @recordings.last.nil?
            @last_event_time = Time.now.to_f
            @last_args_before_event_time = args
          #else
          #  @last_args_before_event_time[:x] += args[:x]
          #  @last_args_before_event_time[:y] += args[:y]
          end
      
          if Time.now.to_f - @last_event_time > 0.0125
            @recordings.last.add_event(@last_args_before_event_time) unless @recordings.last.nil?
            @last_args_before_event_time = nil
            debug_puts "added a move.."
          #else
          #  debug_puts "putting together.."
          end
        else
          @last_event_time = Time.now.to_f
          @last_args_before_event_time = args
      end
    end
  end

  def update_annotation_moved(args)
    args[:type] = "moved"
    args[:guid] = args[:annotation].properties["guid"]
    self.add_recording_event(args)
  end
  
  def update_annotation_rotating(args)
    args[:type] = "rotated"
    args[:guid] = args[:annotation].properties["guid"]   
    self.add_recording_event(args)
  end
  
  def update_annotation_resizing(args)
    args[:type] = "sized"
    args[:guid] = args[:annotation].properties["guid"]  
    self.add_recording_event(args)
  end
  
  def update_annotation_pen_creation(args)
    debug_puts "!pen_creation:" + args.inspect
    args[:type] = "pen_creation"
    self.add_recording_event(args)
  end
  
  def update_annotation_pen_creation_finished(args)
    if args[:annotation].nil?
      args[:type] = "pen_creation_midgroup"
      args[:guid] = "annotation.empty"
      #Force finish because auto groups are bad
      Document.current.current_page.force_group_finish!
    else
      args[:type] = "pen_creation_finished"
      args[:guid] = args[:annotation].properties["guid"] 
      @newly_added_annotations[args[:guid]] = {}
      @newly_added_annotations[args[:guid]][:xml] = args[:annotation].xml
    end
    self.add_recording_event(args)
  end
  
  def update_annotation_clicked(args)
    args[:type] = "clicked"
    args[:guid] = args[:annotation].properties["guid"]
    debug_puts args.inspect
    self.add_recording_event(args)
  end
  
  def update_annotation_deleted(args)
  	debug_puts "deleted!"
  	args[:type] = "deleted"
  	args[:guid] = args[:annotation].properties["guid"]
  	self.add_recording_event(args)
  end
  
  def update_annotation_property_changed(args)
    debug_puts "property changed! " + args.inspect
    args[:type] = "property_changed"
    args[:property] = ""
    args[:value] = ""
    if args.has_key? :color
      args[:property] = "color"
      args[:value] = args[:color]
    elsif args.has_key? :fill_color
      args[:property] = "fill_color"
      args[:value] = args[:fill_color]
    elsif args.has_key? :txt_image
      args[:property] = "txt_image"
      args[:value] = args[:txt_image]
    elsif args.has_key? :transparency
      args[:property] = "transparency"
      args[:value] = args[:transparency]
    elsif args.has_key? :width
      args[:property] = "width"
      args[:value] = args[:width]
	elsif args.has_key? :mirrored_horizontal
	  args[:property] = "mirrored_horizontal"
	  args[:value] = args[:mirrored_horizontal]
	elsif args.has_key? :mirrored_vertical
	  args[:property] = "mirrored_vertical"
	  args[:value] = args[:mirrored_vertical]
    end
    args[:guid] = args[:annotation].properties["guid"]
    self.add_recording_event(args)
  end
  
  def update_page_background_changed(args)
    debug_puts "background changed"
    args[:type] = "background_changed"
    args[:value] = args[:color]
    self.add_recording_event(args)
  end
 
  def update_annotations_shuffled(args)
  	debug_puts "annotations shuffled"
	args[:type] = "annotations_shuffled"
	annotation_guid_string = ""
	args[:annotations].each do |a|
	  annotation_guid_string = annotation_guid_string + a.properties["guid"] + ","
	end
	args[:value] = annotation_guid_string[0..-2]   # trim last comma
	self.add_recording_event(args)
  end
end

class Recorder
  #include Singleton
    
  def Recorder.initialize
    @@animators = []
    @@currently_playing_threads = []
  end
 
  def Recorder.switched_pages
	 Recorder.stop_recording if Recorder.is_recording
	 Recorder.stop_playback
	 @@animators = []
  end
		 
  def Recorder.start_recording
    debug_puts "starting.."
    Recorder.stop_playback

    return unless @@animators.empty?

    animator = Animator.new
    Event.add_observer(:annotation_moved, animator)
    Event.add_observer(:annotation_rotating, animator)
    Event.add_observer(:annotation_resizing, animator)
    Event.add_observer(:annotation_pen_creation, animator)
    Event.add_observer(:annotation_pen_creation_finished, animator)
    Event.add_observer(:annotation_clicked, animator)
    Event.add_observer(:annotation_deleted, animator)
    Event.add_observer(:annotation_property_changed, animator)
    Event.add_observer(:annotations_shuffled, animator)
    Event.add_observer(:page_background_changed, animator)
    animator.start_recording
    @@animators << animator
    debug_puts "okay!"
  end
  
  def Recorder.is_recording
    return false if @@animators.last.nil? or not @@animators.last.respond_to? :is_recording
    return @@animators.last.is_recording
  end
  
  def Recorder.stop_recording
    debug_puts "stopping.."
    Recorder.stop_playback

    return if @@animators.last.nil?
    return unless @@animators.last.is_recording

    @@animators.last.stop_recording
    
    animator = @@animators.last
    Event.remove_observer(:annotation_moved, animator)
    Event.remove_observer(:annotation_rotating, animator)
    Event.remove_observer(:annotation_resizing, animator)
    Event.remove_observer(:annotation_pen_creation, animator)
    Event.remove_observer(:annotation_pen_creation_finished, animator)
    Event.remove_observer(:annotation_clicked, animator)
    Event.remove_observer(:annotation_deleted, animator)
    Event.remove_observer(:annotation_property_changed, animator)
    Event.remove_observer(:page_background_changed, animator)
    
    debugger if DEBUG_ON
        
    debug_puts "stopped, now encoding"
    begin
      Document.current.current_page.set_page_script Recorder.serialize_animation(@@animators.last.get_last_animation)
    rescue Exception => e
      debug_puts "unable to serialize animation " + e.inspect
      debug_puts "backtrace: \n" + e.backtrace.join("\n")
    end
    @@animators = []

    debug_puts "okay!"
    Document.current.draw!
#    StoppableThread.current.sleep 1
    Document.current.draw!
  end
  
  def Recorder.play_recording
    debugger if DEBUG_ON
    page_script = Document.current.current_page.page_script
    return if page_script.nil? or page_script.empty?
    
 #   debug_puts @@currently_playing_threads.inspect
    
    @@currently_playing_threads.each do |t|
#      debug_puts "status: " + t.status.to_s
      if t.nil? or t.should_kill?#or not t.status
        @@currently_playing_threads.delete t
      end
    end
    
    if not @@currently_playing_threads.empty?      
	  debug_puts "there are animations going on.. "
	  @@currently_playing_threads.each do |t|
	    if t.should_pause?
		  debug_puts "there were paused ones so continue them"
		  t.wakeup
		  return
		end
	  end
	  debug_puts "there were no paused ones so its playing so do nothing"
	  return
    end
      
    debug_puts "Loading and playing.."
    animation = Recorder.deserialize_animation(page_script)
    play_thread = animation.play
    debug_puts "done!"
    debug_puts "saving thread"
    @@currently_playing_threads << play_thread
    debug_puts "done. " #+ @@currently_playing_threads.inspect
  end
  
  def Recorder.pause_playback
    #this pauses playback
    debug_puts "stopping.. " #+ @@currently_playing_threads.last.inspect
    @@currently_playing_threads.last.pause_thread unless @@currently_playing_threads.empty?
  end
  
  def Recorder.rewind_playback
    page_script = Document.current.current_page.page_script
    return if page_script.nil?
    
    Recorder.stop_playback
    
    animation = Recorder.deserialize_animation(page_script)
    animation.rewind
  end
  
  def Recorder.stop_playback
    if not @@currently_playing_threads.empty?
      debug_puts "there are animations going on, stopping them"
      @@currently_playing_threads.each do |t|
#        Thread.kill(t) unless t.nil?
		t.set_stop_thread unless t.nil?
      end
    end
  	@@currently_playing_threads = []
  	StoppableThread.stop_all_threads
  end
  
  def Recorder.serialize_animation(animation_object)
    doc = REXML::Document.new
    _animation = doc.add_element "animation"
    _original_state = _animation.add_element "original_state"
    _animation_created = _animation.add_element "animation_created"
    _animation_steps = _animation.add_element "animation_steps"
    
    #save originals
    clr = animation_object.original_bgcolor
    _original_state.attributes["bgcolor"] = clr[0].to_f.to_s + "," + clr[1].to_f.to_s + "," + clr[2].to_f.to_s
    debug_puts animation_object.original_state.inspect
	debug_puts animation_object.original_state_order.inspect
    animation_object.original_state_order.each do |k|
	  debug_puts "using key " + k + " in order.."
      xml = animation_object.original_state[k][:xml]
      if not xml.nil?
        xml = xml.unpack("C*").pack("U*")   #UTF-8 ize XML 
        #debugger if DEBUG_ON
        while xml.include?("<defs")
          defsxml = xml.slice!(/<defs.+?<\/defs>/)
          _newElement = REXML::Document.new defsxml
          _original_state.add_element _newElement.root
        end
        _newElement = REXML::Document.new xml
        _original_state.add_element _newElement.root 
      end
    end

    animation_object.animation_created_objects.each_key do |k|
      xml = animation_object.animation_created_objects[k][:xml]
      if not xml.nil?
        xml = xml.unpack("C*").pack("U*")   #UTF-8 ize XML 
        while xml.include?("<defs")
          defsxml = xml.slice!(/<defs.+?<\/defs>/)
          _newElement = REXML::Document.new defsxml
          _animation_created.add_element _newElement.root
        end
        _newElement = REXML::Document.new xml
        _element = _animation_created.add_element _newElement.root
      end
    end

    animation_object.animation_steps.each_index do |i|
      item = animation_object.animation_steps[i]

	  # Default case if step contains ":type" and ":value"
      _element = _animation_steps.add_element item[:type], { "next_delay" => item[:next_delay] }
      _element.attributes["xml:id"] = item[:guid] if item.has_key? :guid
      _element.attributes["value"] = item[:value] if item.has_key? :value
      #_element.attributes["id"] = item[:annotation].annotation_id if item.has_key? :annotation

	  # Additional attributes and cases..
      case item[:type]
      when "moved"
        _element.attributes["x"] = item[:x].to_s
        _element.attributes["y"] = item[:y].to_s
      when "rotated"
        _element.attributes["angle"] = item[:angle].to_s
      when "sized"
        _element.attributes["width"] = item[:w].to_s
        _element.attributes["height"] = item[:h].to_s
      when "pen_creation"
        _element.attributes["x"] = item[:x].to_s
        _element.attributes["y"] = item[:y].to_s
        _element.attributes["id"] = item[:id].to_s
      when "property_changed"
        _element.attributes["property"] = item[:property].to_s
      when "pen_creation_finished"
        #store pen attributes
      end 
    end

    output = ""
    if defined? REXML::Formatters
			formatter = REXML::Formatters::Default.new(false)
			formatter.write(doc, output)
		else
			doc.write(output)
		end
    return output
  end

  def Recorder.deserialize_animation(xml_string)
     return if xml_string.nil?
     doc = REXML::Document.new(xml_string)
     return if doc.nil?

     animation_object = Animation.new
     original_state = {}
	 original_state_order = []
     original_bgcolor = [1.0, 1.0, 1.0]
     animation_created_objects = {}

     debugger if DEBUG_ON
     
     if doc.elements["animation/original_state"].attributes.has_key? "bgcolor"
       color = []
       doc.elements["animation/original_state"].attributes["bgcolor"].split(',').each do |color_string|
         color << color_string.to_f
       end
       original_bgcolor = color
     end
     
     last_defs_block = ""
     doc.elements["animation/original_state"].each do |annotation|
       id = annotation.attributes["xml:id"]
       raw_string = ''
	   
	   if defined? REXML::Formatters
		   formatter = REXML::Formatters::Default.new(false)
		   formatter.write(annotation, raw_string)
	   else
		   annotation.write(raw_string);
	   end

	   raw_string = raw_string.unpack("U*").pack("C*") #UTF-16 ize XML

	   if id.nil? and raw_string.include? "<defs"
		 last_defs_block += raw_string
	   else
		 if id.nil? or id.length < 1
		   # Just nil and not a def block
		   # this handle cases of annotations in g groups that don't save xml:id to parent group
		   annotation = Annotation.create_from_xml(raw_string)
		   if not annotation.nil?
			 id = annotation.properties["guid"]
		   end
	  	 end 
	   end

	   if not id.nil? and not id.length < 1
         original_state[id] = {}
         original_state[id][:xml] = last_defs_block + raw_string
         original_state_order << id
         last_defs_block = ""
       end
     end

     last_defs_block = ""
     doc.elements["animation/animation_created"].each do |annotation|
       id = annotation.attributes["xml:id"]
			 raw_string = ''
			 if defined? REXML::Formatters
				 formatter = REXML::Formatters::Default.new(false)
				 formatter.write(annotation, raw_string)
			 else
				 annotation.write(raw_string);
			 end
			 raw_string = raw_string.unpack("U*").pack("C*") #UTF-16 ize XML
			 if id == nil and raw_string.include? "<defs"
			   last_defs_block += raw_string
		   else
         animation_created_objects[id] = {}
         animation_created_objects[id][:xml] = last_defs_block + raw_string
         last_defs_block = ""
       end
     end
     
     animation_object.set_initial_state(original_state, original_bgcolor, original_state_order)
     animation_object.animation_created_objects = animation_created_objects
     
     doc.elements["animation/animation_steps"].each do |step|
      animation_step = {}

      animation_step[:type] = step.name
      attributes = step.attributes
      animation_step[:next_delay] = attributes["next_delay"].to_f
      animation_step[:guid] = attributes["xml:id"]
      animation_step[:value] = attributes["value"] if attributes.has_key? "value"
      
      case step.name
      when "moved"
        animation_step[:x] = attributes["x"].to_f
        animation_step[:y] = attributes["y"].to_f
      when "rotated"
        animation_step[:angle] = attributes["angle"].to_f
      when "sized"
        animation_step[:w] = attributes["width"].to_f
        animation_step[:h] = attributes["height"].to_f
      when "pen_creation"
        animation_step[:x] = attributes["x"].to_f
        animation_step[:y] = attributes["y"].to_f
        animation_step[:id] = attributes["id"].to_i
      when "property_changed"
        animation_step[:property] = attributes["property"]
      when "pen_creation_finished"
        #store pen attributes
      end

      animation_object.animation_steps << animation_step
     end

     return animation_object
  end
end

Recorder.initialize

#Animation defines

class ObjectAnimationFactory
  def ObjectAnimationFactory.initialize
    #Set up list of class names, and load
    @@animation_classes = [FlyInAnimation, FlyOutAnimation, ShrinkGrowAnimation, SpinAnimation, FadeInAnimation, FadeOutAnimation, FlipAnimation]
    @@object_animations = {}
    @@object_animations_available_options = {}
    @@animation_classes.each do |animation_class|
      a = animation_class.new
      @@object_animations[a.display_name] = animation_class
      options = nil
      if not a.available_options.nil?
        options = []
        a.available_options.each do |i|
          options << a.display_name_for_option(i)
        end
      end
      @@object_animations_available_options[a.display_name] = options
    end  
  end
  
  def ObjectAnimationFactory.animation(animation_name)
    return nil unless @@object_animations.has_key? animation_name
    return @@object_animations[animation_name].new
  end
  
  def ObjectAnimationFactory.available_animations
    @@object_animations.keys.sort
  end
  
  def ObjectAnimationFactory.available_options_for_animation(animation_name)
    @@object_animations_available_options[animation_name]
  end
  
  def ObjectAnimationFactory.debug
    ObjectAnimationFactory.initialize if @@animation_classes.nil? 
    @@animation_classes.each do |animation_class|
        a = animation_class.new
        debug_puts a.display_name
        if not a.available_options.nil?
          a.available_options.each do |i|
            debug_puts "\t" + a.display_name_for_option(i)
          end
        end
      end
  end
end


class ObjectAnimation
  def ObjectAnimation.initialize
    @@currently_animation_threads = {} # guid -> thread
  end
  
  def ObjectAnimation.stop_animation_for_annotation(a)
    if @@currently_animation_threads.has_key? a.properties["guid"]
      @@currently_animation_threads[a.properties["guid"]].set_stop_thread
      @@currently_animation_threads.delete(a.properties["guid"])
    end
  end
  
  def ObjectAnimation.already_animating?(a)
    return @@currently_animation_threads.has_key? a.properties["guid"]
  end
  
  def display_name
    "BasicAnimation"
  end
  
  def available_options
    nil
  end
  
  def display_name_for_option
    nil
  end
  
  def option_for_annotation(annotation)
    return 0 if annotation.nil?
    return 0 if self.available_options.nil?
    return 1 unless annotation.properties.has_key? "animation_options"
    
    #if there is an invalid option on this annotation and there are options assume the first option
    return 1 if annotation.properties["animation_options"] < 1 && !self.available_options.nil?
    
    return annotation.properties["animation_options"]
  end

  def tick_amount_for_annotation(annotation)
    return 1 if annotation.nil?
    return 1 unless annotation.properties.has_key? "animation_speed"
  
    case annotation.properties["animation_speed"]
    when 0
      return 3.0
    when 1
      return 1.0
    when 2
      return 0.333
    end
  
    return 1
  end

  def repeat_amount_for_annotation(annotation)
    return 1 if annotation.nil?
    return 1 unless annotation.properties.has_key? "animation_repeats"

  	repeats = annotation.properties["animation_repeats"]
	  return 1 if repeats > 10
	  return repeats if repeats >= 0

    case annotation.properties["animation_repeats"]
    when -1
      return 100000
    when -2
      return 100000
    end
  
    return 1
  end

  def max_counter
  	return 100000
  end

  def is_valid
	if @annotation == nil
		return false
	end
	return @annotation.is_valid
  end
  def initialize
	# args for the thread
	@document = nil
	@counter = 0
	@tick_amount = 0
	@repeat_amount = 0
	@number_of_times_played = 0
	@annotation = nil
	@pre_animation_properties = {}
  end
  
  def animation_should_continue(counter)
	return false
  end
  def linear_animation(annotation)
    if annotation==nil
		debug_puts "nil parameter for linear_animation!"
		return false
	end
    Globals.start_flying
	
    if @@currently_animation_threads.has_key? annotation.properties["guid"]
      #debug_puts "already animating, don't do it again, but check first, "
      t = @@currently_animation_threads[annotation.properties["guid"]]
      if t.nil? or t.should_kill?#t.status
        @@currently_animation_threads.delete(annotation.properties["guid"])
      else 
        return false
      end
    end
    
    @document = Document.current
    @counter = 0
    @tick_amount = tick_amount_for_annotation(annotation)
    @repeat_amount = repeat_amount_for_annotation(annotation)
    @number_of_times_played = 0
    
    @pre_animation_properties = {}
    @pre_animation_properties["transparency"] = annotation.transparency
    @pre_animation_properties["position"] = Point[annotation.position.x, annotation.position.y]
    @pre_animation_properties["size"] = Size[annotation.size.width, annotation.size.height]
	@pre_animation_properties["options"] = annotation.properties["animation_options"]
	@annotation = annotation
	
    @@currently_animation_threads[annotation.properties["guid"]] = StoppableThread.new
    @@currently_animation_threads[annotation.properties["guid"]].set_running_object(self)
    return true
  end
    
  def run_once
	    #debug_puts "checking should_stop? - " + StoppableThread.current.should_stop?.inspect
		
        if StoppableThread.current.should_stop?
		  #debug_puts "ending prematurely, yielding max counter " + self.max_counter.to_s
		  
		  wasOkay = animation_should_continue(self.max_counter)
#		   if @annotation.demoing_animation
				#case 1162298
			  if @annotation.demoing_animation or not (self.is_a? SpinAnimation or self.is_a? FadeInAnimation or self.is_a? FadeOutAnimation)
				@annotation.transparency = @pre_animation_properties["transparency"]
				@annotation.position = @pre_animation_properties["position"]
				@annotation.size = @pre_animation_properties["size"]
			  end
#            end
            @document.draw
            return false
          #Thread.kill(self)
        end
        
        @should_continue = animation_should_continue(@counter)
		
        #debug_puts "should continue? " + @should_continue.to_s
        #debug_puts "adding counter"
        
      	@counter += 1
      	debug_puts "tick amount = " + @tick_amount.inspect
      	debug_puts "sleeping for: " + (0.025 / @tick_amount).inspect
        StoppableThread.current.sleep(0.025 / @tick_amount)
        @document.draw
     
        #debugger if DEBUG_ON
        debug_puts "done drawing.."
      
        if not @should_continue
          debug_puts "should not continue"
          @annotation.finished_animating
          @number_of_times_played += 1
          if @number_of_times_played >= @repeat_amount
            debug_puts "stopping animation"
            @@currently_animation_threads.delete(@annotation.properties["guid"]) { debug_puts "couldn't find animation in currently animating list" }
			if self.is_a? FlipAnimation
				self.reset_flip
			end
            if @annotation.demoing_animation
              @annotation.transparency = @pre_animation_properties["transparency"]

			  if @annotation.demoing_animation or not (self.is_a? FlyOutAnimation)
				@annotation.position = @pre_animation_properties["position"]
			  end


			  @annotation.size = @pre_animation_properties["size"]
            end
            @document.draw
            return false
          else
			@repeat_amount = repeat_amount_for_annotation(@annotation)
            debug_puts "yes, go again"
            StoppableThread.current.sleep 0.5
            @counter = 0
          end
        end
      return true
      
  end

  def reset
    debug_puts "Reset before stopping thread"
	if @annotation.demoing_animation
		back_to_start
		if self.is_a? ShrinkGrowAnimation
			@annotation.set_property({"animation_options" => 0})
		end	
	else
		if self.is_a? SlideAnimation or self.is_a? FlipAnimation
			back_to_start
		else 
			if self.is_a? ShrinkGrowAnimation and @should_continue
				@annotation.set_property({"animation_options" => @pre_animation_properties["options"]});
				back_to_start
			end
		end
	end	
  end

  def back_to_start
	@annotation.transparency = @pre_animation_properties["transparency"]
	@annotation.position = @pre_animation_properties["position"]
	@annotation.size = @pre_animation_properties["size"]
  end
   
  def threadless_animation
    counter = 0
    loop do
      counter += 1
      yield counter
      StoppableThread.current.sleep 0.05
      break if counter > 1000
    end
  end
end


# slide_from_side
class SlideAnimation < ObjectAnimation
  
  def display_name
    "Fly In"
  end
  
  def available_options
    1..8
  end
  
  def display_name_for_option(option)
    ["Left", "Right", "Top", "Bottom", "Top Left", "Top Right", "Bottom Left", "Bottom Right"][option - 1]
  end

  def flying_out
    false
  end
  
  def ending_position (annotation)
    x = annotation.position.x
    y = annotation.position.y
    
    case option_for_annotation(annotation)
    when 1, 5, 7
      x = annotation.size.width * -1 - 100
    when 2, 6, 8
      x = 800 + 100
    end
    
    case option_for_annotation(annotation)
    when 3, 5, 6
      y = annotation.size.height * - 1 - 100
    when 4, 7, 8
      y = 1000
    end
    
    return Point[x, y]
  end
  
  def initialize
	@starting_position = nil
	@end_position = nil
	@paht = nil
	
	super
  end
  
  def animate (annotation)
    #should_continue = true
    @starting_position = Point[annotation.position.x, annotation.position.y]
    @end_position = ending_position(annotation)
    @path = (@starting_position - @end_position)
    
    debug_puts "option: " + option_for_annotation(annotation).to_s
    #debug_puts "starting_position" + @starting_position.inspect
    #debug_puts "ending_position" + @end_position.inspect
    
    @path = Point[@path.x, @path.y]
    
    linear_animation(annotation)
   end
   
   def animation_should_continue(counter)
	  #debug_puts "next counter " + counter.to_s
	  should_continue = true
      if counter == 0
         should_continue = true
		 # BUGZID:1149088
         #@starting_position = [@annotation.position.x, @annotation.position.y]
         @annotation.position = @end_position
		 #@annotation.set_property({"original_positionX" => @starting_position.x})
		 #@annotation.set_property({"original_positionY" => @starting_position.y})
		 #debug_puts "placed in corner"
      end
    
      if counter > 20
        #debug_puts "ending"
        @annotation.position = @starting_position
		if self.flying_out
			@annotation.position = @end_position
		else
			@annotation.position = @starting_position
		end
        should_continue = false
      else
        #debug_puts "moving"
        percentage = (counter / 20.0)
        percentage = 1 - percentage if self.flying_out 
        #debug_puts "moving to.. via " + percentage.inspect
        #debug_puts "path: " + @path.inspect + @path.class.to_s
        
        debug_puts (@path * percentage).inspect
        #debug_puts "final pos = " + ((@path * percentage) + @end_position).inspect
        @annotation.position = (@path * percentage) + @end_position
      end
    
      return should_continue
  end
  
end


#fly_in
class FlyInAnimation < SlideAnimation
end


#fly_out
class FlyOutAnimation < SlideAnimation
  def display_name
    "Fly Out"
  end
  
  def flying_out
    true
  end
end


#grow
class ShrinkGrowAnimation < ObjectAnimation
  def display_name
    "Shrink and Grow"
  end
 
  def scale_factor(a)
	if a.properties.has_key? "animation_options"
          debug_puts "Growing animation_options" + a.properties["animation_options"].to_s
  	  return 1.05 if a.properties["animation_options"] == 1
  	end
        debug_puts "Shrinking animation_options" + a.properties["animation_options"].to_s
  	return 0.952380952380 
  end

  def initialize
	super
  end
  
  def animate (a)
    linear_animation(a)
  end
  
  def animation_should_continue(counter)
      @annotation.scale_in_place(scale_factor(@annotation))
      should_continue = true
      if counter > 20
		option_value = @annotation.properties["animation_options"];
        if option_value == 1
          option_value = 0
        else
          option_value = 1
        end
        @annotation.set_property({"animation_options" => option_value});
        should_continue = false
	  end
      return should_continue   
  end
  
end

#shrink
#class ShrinkAnimation < ObjectAnimation
#  def display_name
#    "Shrink and Grow"
#  end
#  
#  def animate (a)
#    linear_animation(a) do |counter|
#      a.scale_in_place(0.952380952380)
#      if counter > 20
#        a.set_property({"animation_type", "Grow"})
#        return false
#      end
#      true
#    end
#  end
#end

#spin
class SpinAnimation < ObjectAnimation
  def display_name
    "Spin"
  end
  
  def available_options
    1..2
  end
  
  def display_name_for_option(option)
    ["Clockwise", "Counter-Clockwise"][option - 1]
  end
  
  def spin_direction(a)
    case option_for_annotation(a)
    when 1
      return 1
    when 2
      return -1
    end
  end
  
  def initialize
	@direction = 1
	@target_rotation = 0.1
	@should_stop_spinning = false
	super
  end


  
  def animate (a)
    @direction = self.spin_direction(a)
   	@target_rotation =  a.rotation

    linear_animation(a)
  end
    
  def animation_should_continue(c) 
	  if c == self.max_counter
		return false
	  end
	  
      should_continue = true
      if c == 0
        @should_stop_spinning = false
        @should_continue = true
      end    
      factor = 0.25
      @annotation.rotation += ((-(factor*c) * (factor*c)) + 100) * @direction unless @should_stop_spinning
	  @annotation.rotation = @annotation.rotation % 360
	  if c > 32
		@should_stop_spinning = true
    	if (@annotation.rotation - @target_rotation).abs > 35
    	    @annotation.rotation += 35
    	else
    	  @annotation.rotation = @target_rotation
    	  should_continue = false
		end
	  end

	  if c > 36
		 @annotation.rotation = @target_rotation
		 should_continue = false
	  end	  	
		
  	  return should_continue
    end
    
end

#fade_in
class FadeInAnimation < ObjectAnimation
  def display_name
    "Fade In"
  end
  
  def animate (a)
    linear_animation(a)
  end
  def animation_should_continue(counter)
	  if counter == max_counter
		return false
	  end
      @annotation.transparency = counter / 75.0
      return counter < 75.0
  end
end

#fade_out
class FadeOutAnimation < ObjectAnimation
  def display_name
    "Fade Out"
  end
  
  def animate (a)
    linear_animation(a)
  end
  def animation_should_continue(counter) 
	  if counter == max_counter
		return false
	  end
      @annotation.transparency = 1 - (counter / 75.0)
      return counter < 75.0
  end
end

#flip
class FlipAnimation < ObjectAnimation
  def display_name
    "Flip around axis"
  end
  
  def available_options
    1..6
  end
  
  def display_name_for_option(option)
    ["Middle Vertical", "Middle Horizontal", "Left Edge", "Right Edge", "Top Edge", "Bottom Edge"][option - 1]
  end
  
  def initialize
	@original_size = nil
	@original_position = nil
	@begin_pos = nil
	@option = 1
	@step_size = 0
	@need_flip = false
	super
  end
  
  def reset_flip
	if @need_flip
		case @option
		when 3
			@annotation.mirror_horizontal!
		when 4
			@annotation.mirror_horizontal!
		when 5
			@annotation.mirror_vertical!
		when 6
			@annotation.mirror_vertical!
		end
	end
  end
  
  def animation_should_continue(counter)
	if counter == max_counter
		self.reset_flip
        return false
	end
	
	should_continue = true
	case @option
	when 1
        if counter < 20
			#BUGZID 1139465, the annotation size should not small than 1
			if @annotation.size.width > @step_size+1
				@annotation.size.width = @annotation.size.width - @step_size
				@annotation.position.x = @annotation.position.x + (@step_size / 2.0)
			end
        end
        #flip!
        @annotation.mirror_horizontal! if counter == 20
		
        if counter > 20
			if @annotation.size.width < @original_size.width - @step_size
				@annotation.size.width = @annotation.size.width + @step_size
				@annotation.position.x = @annotation.position.x - (@step_size / 2.0)
			end
        end
        if counter > 40
          @annotation.size.width = @original_size.width
          @annotation.position.x = @original_position.x
          should_continue = false
        end
    when 2
        if counter < 20
			if @annotation.size.height > @step_size + 1
				@annotation.size.height= @annotation.size.height - @step_size
				@annotation.position.y = @annotation.position.y + (@step_size / 2.0)
			end
        end
        #flip!
        @annotation.mirror_vertical! if counter == 20
        if counter > 20
			@annotation.size.height = @annotation.size.height + @step_size
			@annotation.position.y = @annotation.position.y - (@step_size / 2.0)
        end
        if counter > 40
			@annotation.size.height = @original_size.height
			@annotation.position.y = @original_position.y
          should_continue = false
        end
    when 3
        if counter < 20
			if @annotation.size.width > 1
				@annotation.size.width = @annotation.size.width - @step_size
			end
        end
        #flip!
        if counter == 20
			@annotation.mirror_horizontal!
			@annotation.position.x = @annotation.position.x - @annotation.size.width
			@need_flip = !@need_flip
		end
        if counter > 20
          @annotation.size.width = @annotation.size.width + @step_size
          @annotation.position.x = @annotation.position.x - @step_size
        end
        if counter > 40
          @annotation.size.width = @original_size.width
          @annotation.position.x = @begin_pos.x - @original_size.width
		  @option = 4
		  @begin_pos.x = @annotation.position.x
          should_continue = false
        end
        @annotation.position.x = 2 if @annotation.position.x < 2 
        @annotation.position.x = 798 if @annotation.position.x > 798
    when 4
        if counter < 20
			if @annotation.size.width > 1
				@annotation.size.width = @annotation.size.width - @step_size
				@annotation.position.x = @annotation.position.x + @step_size
			end
        end
        #flip!
        if counter == 20
			@annotation.mirror_horizontal!
			@annotation.position.x = @annotation.position.x + @annotation.size.width
			@need_flip = !@need_flip
        end
        if counter > 20
          @annotation.size.width = @annotation.size.width + @step_size
        end
        
        if counter > 40
          @annotation.size.width = @original_size.width
          @annotation.position.x = @begin_pos.x + @original_size.width
		  @option = 3
		  @begin_pos.x = @annotation.position.x
		  
          should_continue = false
        end
        @annotation.position.x = 2 if @annotation.position.x < 2 
        @annotation.position.x = 798 if @annotation.position.x > 798
    when 5
        if counter < 20
			if @annotation.size.height > 1
				@annotation.size.height= @annotation.size.height - @step_size
			end
        end
        #flip!
        if counter == 20
			@annotation.mirror_vertical!
			@annotation.position.y = @annotation.position.y - @step_size
			@need_flip = !@need_flip
        end
        if counter > 20
			@annotation.size.height = @annotation.size.height + @step_size
			@annotation.position.y = @annotation.position.y - @step_size
		end
        if counter > 40
          @annotation.size.height = @original_size.height
          @annotation.position.y = @begin_pos.y - @original_size.height
          @option = 6
          @begin_pos.y = @annotation.position.y
          should_continue = false
        end
		@annotation.position.y = 2 if @annotation.position.y < 2 
    when 6
        if counter < 20
			if @annotation.size.height > 1
				@annotation.size.height= @annotation.size.height - @step_size
				@annotation.position.y = @annotation.position.y + @step_size
			end
        end
        #flip!
        if counter == 20
			@annotation.mirror_vertical!
			@annotation.position.y = @annotation.position.y + @step_size
			@need_flip = !@need_flip
        end
        if counter > 20
          @annotation.size.height = @annotation.size.height + @step_size
        end
        if counter > 40
          @annotation.size.height = @original_size.height
          @annotation.position.y = @begin_pos.y + @original_size.height
          @option = 5
          @begin_pos.y = @annotation.position.y
          should_continue = false
        end
    end
        return should_continue	
  end
  
  def animate (a)
    @original_size = Size[a.size.width, a.size.height]
    @original_position = Point[a.position.x, a.position.y]
    @begin_pos = Point[a.position.x, a.position.y]
    @option = option_for_annotation(a)
    case @option
    when 1
#      debug_puts "middle horizontal"
      @step_size = ((@original_size.width-1) / 20.0)
    when 2
#      debug_puts "top animation"
      @step_size = ((@original_size.height-1) / 20.0)
    when 3
#      debug_puts "left animation"
      @step_size = ((@original_size.width-1) / 21.0)
    when 4
#      debug_puts "right animation"
      @step_size = ((@original_size.width-1) / 21.0)
    when 5
#      debug_puts "top animation"
      @step_size = ((@original_size.height-1) / 21.0)
    when 6
#      debug_puts "bottom animation"
      @step_size = ((@original_size.height-1) / 21.0)
    end
    debug_puts "step size="+@step_size.to_s
    linear_animation(a)
  end
  
end

ObjectAnimationFactory.initialize
ObjectAnimation.initialize

#puts "Loaded Notebook Animations.. v16"

if __FILE__ == $0  
  sa = SlideAnimation.new
  debug_puts sa.available_options.inspect
  debug_puts sa.display_name_for_option 3

  p1 = Point[3, 4]
  p2 = Point[0, 0]

  debug_puts ((p2 - p1) * 0.5).inspect
  
  ObjectAnimationFactory.debug
  debug_puts ObjectAnimationFactory.available_animations.inspect
  debugger if DEBUG_ON
  debug_puts ObjectAnimationFactory.available_options_for_animation("Fly Out").inspect
  debug_puts ObjectAnimationFactory.animation("Fly In").inspect
end
