require "singleton"

#Event class
#A singleton that represents events in Notebook
#methods
# => add_observer (:event_type, object with update method / object with update_"event_type" method / method)
# => fire_event (:event_type, args)
#     anything that that is observing will get called approprately with args
# => remove_observer() does what it says...

class Event
  include Singleton

  def Event.initialize
    @@event_observers = {:page_loaded => Array.new, :annotation_moved => Array.new} unless defined? @@events_observers
  end
  
  def Event.fire_event(event_type, args)
    return unless @@event_observers.has_key? event_type
    
    @@event_observers[event_type].map do |o|
      o.method(:update.to_s + "_" + event_type.to_s).call(args) if o.respond_to?(:update.to_s + "_" + event_type.to_s)
      o.update(args) if o.respond_to? :update
      o.call(args) if o.respond_to? :call
    end
  end
  
  def Event.add_observer(event_type, o)
    #Allow creation for now
    @@event_observers[event_type] = Array.new unless @@event_observers.has_key? event_type
    
    if o.respond_to? :update or 
       o.respond_to?(:update.to_s + "_" + event_type.to_s) or
       o.respond_to? :call
      @@event_observers[event_type] << o 
    else
      begin
        @@event_observers[event_type] << method(o)  
      rescue NameError => e
        debug_puts "Symbol #{o} does not exist"
      rescue Exception => e
        debug_puts "Unable to add #{o} to observer list"
      end
    end
    #May need this in the future
    #return @@event_observers[event_type].index(@@event_observers[event_type].last)
  end
  
  def Event.remove_observer(event_type, o)
    if o.respond_to? :update or 
       o.respond_to?(:update.to_s + "_" + event_type.to_s)
       o.respond_to? :call
      @@event_observers[event_type].delete(o) { "Observer not found" }
    else
      begin
        @@event_observers[event_type].delete(method(o))  
      rescue Exception => e
        debug_puts "Observer not found"
      end
    end
  end
  
  def Event.debug
    debug_puts @@event_observers.inspect
  end
  
end

Event.initialize

def run_threads
	StoppableThread.threads_run_once
end

page_entered_proc = Proc.new do |args|
  annotations = Document.current.current_page.annotations
  did_adjust_transparency = false

  annotations.each do |ann|
    #App.alert ann.properties.inspect
    if ann.properties.has_key? "animation_occurs"
      #there is an animation
      if ann.properties["animation_occurs"] == 1
        #do animation
	  	#ann.do_animation
		ann.do_delayed_animation
      end
    end
    
    if ann.properties.has_key? "animation_type"
      case ann.properties["animation_type"]
      when "Fade In"
        ann.transparency = 0.02
        did_adjust_transparency = true
      when "Fade Out"
        ann.transparency = 1.0
        did_adjust_transparency = true
	  when "Fly Out"
		if ann.properties.has_key? "original_positionX"
			#there is an original position stored
			ann.position.x = ann.properties["original_positionX"]
		end
		if ann.properties.has_key? "original_positionY"
			#there is an original position stored
			ann.position.y = ann.properties["original_positionY"]
		end
		did_adjust_transparency = true
      end
    end
    Document.current.delay_draw!
  end
  
  Document.current.delay_draw! if did_adjust_transparency
end

Event.add_observer(:page_entered, page_entered_proc)

page_exited_proc = Proc.new do |args|
  StoppableThread.stop_all_threads
  Recorder.switched_pages
end

Event.add_observer(:page_exited, page_exited_proc)

onclick_proc = Proc.new do |args|
  ann = args[:annotation]
  did_stop = false
  
  if ObjectAnimation.already_animating? ann
    #already animating, check if we need to stop
    if ann.properties.has_key? "animation_repeats"
      if ann.properties["animation_repeats"] == -2
        ObjectAnimation.stop_animation_for_annotation(ann)
        did_stop = true
      end
    end
  end
  
  if ann.properties.has_key? "animation_occurs"
    #there is an animation and it is a onclick
    if ann.properties["animation_occurs"] == 0
      #animate
      ann.do_animation unless did_stop
    end
  end
  
end
Event.add_observer(:annotation_clicked, onclick_proc)

#page_entered_proc = Proc.new do debug_puts "Loading .." end
#Event.initialize
#Event.add_observer(:page_entered, page_entered_proc)


if __FILE__ == $0  
  class Animator
    def initialize(name)
      @name = name
    end

    def update(args)
      debug_puts "#{@name} receved an event with args " + args.inspect
    end
    
    def update_annotation_moved(args)
      debug_puts "annotation moved!" + args.inspect
    end

    def update_anotation_rotated(args)
      debug_puts "annotation rotated" + args.inspect
    end
  end

  def test_method(args)
    debug_puts "test method " + args.inspect
  end
  
  #TODO: roll this into a unit test
  Event.initialize
  Event.add_observer(:annotation_moved, :test_method)
  animator = Animator.new("test 1")
  Event.add_observer(:annotation_moved, animator);
  Event.fire_event(:annotation_moved, {:x => 3, :y => 4});
  Event.add_observer(:annotation_moved, Animator.new("test 2"));
  Event.remove_observer(:annotation_moved, :test_method);
  Event.fire_event(:annotation_moved, {:x => 32, :y => 443});
  Event.remove_observer(:annotation_moved, animator)
  Event.fire_event(:annotation_moved, {:x => 312, :y => 5});
  
  Event.add_observer(:blah, Proc.new {|args| debug_puts "a new blah event! #{args}"})
  Event.fire_event(:blah, "with these args")
  Event.fire_event(:not_exist, "args")
  Event.debug
end

