/**
 * Notebook SDK Javascript Library 1.0.0
 * nb-1.0.0.js
 * http://smarttech.com/notebook
 * Copyright 2011 - 2012, SMART Technologies ULC
 * All rights reserved.
 */

nb_load_includes();

(function () {

/**
 * @namespace
 * @name NB
 */
var NB;
NB = this.NB = {};
 
NB.VERSION = "1.0.0";
NB.SERVER = "http://127.0.0.1:25899";
NB.HOSTGUID = "annotation.0";
NB.DEBUGON = false;
NB.HAS = ["nb.js"];

// Flags to set when Widget is ready
NB._fEventsListenerReady = false;
NB._fLocalizationReady = false;
NB._fWindowLoaded = false;
NB._fNBReadyCalled = false;
 
// Populate the above SERVER and GUID with correct values if available.
if (document.location.host != "") {
    NB.SERVER = "http://" + document.location.host;

    // The url is of the form: http://127.0.0.1:25899/widget/annotation.4EUNH7S7HUU49AKC6ST2FL5CA/index.html.
    // Therefore the 5th part when split is the guid.
    var hrefParts = document.location.href.split("/");
    if (hrefParts.length > 4) {
        NB.HOSTGUID = hrefParts[4];
    }
} else {
	if(isDefined(document.location.search) && document.location.search != "") {
		var qs = document.location.search;
		var part = "_port=";
		var port = qs.substring(qs.indexOf(part) + part.length);
		if(port.length > 3) {
			NB.SERVER = "http://127.0.0.1:" + port;
		}
	}
}

DEBUG = function (msg) { if (NB.DEBUGON == true) console.log(msg); };

$(window).load(function() {
	NB._fWindowLoaded = true;
	NB._checkLoaded();
});
 
// Make sure all JS handler is loaded
var _isNBjsAllLoaded = function() {
	return NB._fEventsListenerReady && NB._fLocalizationReady && NB._fWindowLoaded;
};

/**
 * Register callback for when the load is finished.
 * @param {function} callback
 */
NB.ready = function(readyHandler) {
	if (_isNBjsAllLoaded()) {
		if (isDefined(readyHandler) ) {
			NB._fNBReadyCalled = true;
			readyHandler();
		}
	} else {
		NB.onload = readyHandler;
	}
};
 
NB._checkLoaded = function() {
	if (!NB._fNBReadyCalled && _isNBjsAllLoaded()) {
		NB._fNBReadyCalled = true;
		if (NB.onload) {
			NB.onload();
		}
	}
};
 
 
// Events
// Should use websocket normally
NB._processEvents = function () {
    var url = NB.SERVER + "/" + NB.VERSION + "/events/" + NB.HOSTGUID + "/";
    $.get(url, function (data) {
		if(data.length < 1) return;
	var events = JSON.parse(data);
        for (var i in events) {

            if (events[i].length < 1) { continue; }
            var event = events[i];
    DEBUG("_processEvent: ");
		DEBUG(event);
            for (var j in _eventObservers[event.eventName]) {
		DEBUG("  calling: " + _eventObservers[event.eventName][i]);
                try {
                    _eventObservers[event.eventName][j](event);
                } catch (e) {
                    DEBUG("Observer threw an exception");
                    if (NB.DEBUGON) {
						throw e;
					}
                }
            }
        }
    });
};

var _eventObservers = [];
var _eventObserverIds = [];

var startNormalProcessEvents = function() {
	NB._fEventsListenerReady = true;
	NB._checkLoaded();
	var url = NB.SERVER + "/" + NB.VERSION + "/events/" + NB.HOSTGUID + "/";
	$.get(url, function (data) {
		setInterval(NB._processEvents(), 100);
	});
};

//
// Create websocket for events
//
try {
	var socketEventAddr = NB.SERVER.replace("http", "ws") + "/" + NB.VERSION + "/events/" + NB.HOSTGUID + "/";
	console.log("Connecting to Notebook: " + socketEventAddr + " ...");

	var socket;
	var protocol = "http-only";
	// Create new sockets to catch events
	if ( typeof WebSocket !== "undefined" ) {
		socket = new WebSocket(socketEventAddr, protocol);
	} else if ( typeof MozWebSocket !== "undefined" ) {
		socket = new MozWebSocket(socketEventAddr, protocol);
	} else {
		console.log("Unable to connect, websockets not available");
		startNormalProcessEvents();
	}

	var succeeded = false;
	socket.onopen = function() {
		//console.log("Socket open ("+socket.readyState+") on host: "+socketEventAddr);
		console.log("Connected.");
		succeeded = true;
		NB._fEventsListenerReady = true;
		NB._checkLoaded();
	};

	socket.onmessage = function(msg) {
		var evt = JSON.parse(msg.data);
		DEBUG("_processEvent from webSocket: ");
    
		var proc = {};
		proc.evt = evt;
		proc.func = function() {
						var evt = proc.evt;
						DEBUG(evt);
						for (var i in _eventObservers[evt.eventName]) {
						  DEBUG("  calling: " + _eventObservers[evt.eventName][i]);
						  try {
							  _eventObservers[evt.eventName][i](evt);
						  } catch (e) {
							  DEBUG("Observer threw an exception");
							  throw e;
						  }
						}
					};

		setTimeout(proc.func, 100);
	};

	socket.onclose = function() {
		if(!succeeded) console.log("Unable to connect.  Are you running Notebook and do you have the right port? (use ?_port= to specify port for local files)");
		if(succeeded) console.log("Socket closed with host: " + socketEventAddr);
		startNormalProcessEvents();
	};

} catch(exception) {
	console.log("Error: " + exception);
	startNormalProcessEvents();
}

/**
 * Adds an observer to listen to a particular event.
 *
 * Observers are the way we broadcast events to the widgets.
 * This is the only API that the user needs.
 *
 * Events list:
 *
 * - *annotationMovingEvent*
 *   - Sent when an object is moving on the page.
 *   - Event data are the GUID and the new position (x, y) of the object moved.
 * - *annotationRotatedEvent*
 *   - Sent when an object has been rotated.
 *   - Event data are the GUID of the object and the angle of the rotation.
 * - *annotationResizedEvent*
 *   - Sent when an object has been resized.
 *   - Event data are the GUID and the new size (width, height) of the object.
 * - *annotationClickedEvent*
 *   - Sent when an object has been clicked.
 *   - Event data is the GUID of the object clicked.
 * - *annotationDeletedEvent*
 *   - Sent when an object has been deleted.
 *   - Event data is the GUID of the object deleted.
 * - *annotationAddedEvent*
 *   - Sent when an object has been added to the page.
 *   - Event data is the GUID of the object added.
 * - *annotationOrderChangedEvent*
 *   - Sent when the z-order of the objects has changed. In order to get the new z-order of the objects on the page, call {@link NB.page.getObjects}.
 * - *selectionChangedEvent*
 *   - Sent when the selection of objects has changed.
 *   - Event data is the list of GUIDs of the new selected objects.
 * - *annotationDeSelectedEvent*
 *   - Sent when objects are deselected.
 *   - Event data is the list of GUIDs of the objects deselected.
 * - *annotationSelectionAddedEvent*
 *   - Sent when objects are selected.
 *   - Event data is the list of GUIDs of the new objects selected.
 * - *propertyChangedEvent*
 *   - Sent when a property of an object has changed.
 *   - Event data are the GUID and the new pair property/value for the object.
 * - *pageVisibleAreaChanged*
 *   - Sent when the visibale area on the page has changed.
 *   - Event data is the new visible area (x, y, width, height). 
 * - *widgetWillActivate*
 *   - Sent when the widget will be active (for example: when the page that conatins the widget is displayed).
 * - *inkingEvent*
 *   - Sent when an ink object is being created.
 *   - Event data are the GUID of the ink object and its new point (x, y) added.
 * - *inkingFinishedEvent*
 *   - Sent when an ink object is created.
 *   - Event data are the GUID and the position (x, y) of the new ink object.
 * - *annotationFinishedAnimatingEvent*
 *   - Sent when an object being animated ends.
 *   - Event data is the GUID of the object.
 * - *textChangedEvent*
 *   - Sent when text of an object has changed.
 *   - Event data are the GUID of the object and its new text value.
 * - *annotationsUngroupedEvent*
 *   - Sent when objects are ungrouped. 
 * - *annotationsGroupedEvent*
 *   - Sent when objects are grouped.
 * - *languageChangedEvent*
 *   - Sent when the language of the UI has changed.
 *   - Event data are the codes for the languages and the country and if it is a Right-To-Left writing.
 * - *pageEnterEvent*
 *   - Sent when the page is displayed.
 *   - Event data is the GUID of the page entering.
 * - *pageLeaveEvent*
 *   - Sent when the page is leaving.
 *   - Event data is the GUID of the page leaving.
 *
 * Drag and Drop events:
 *
 * - *dndStartEvent*
 *   - Sent when starting to drag objects.
 *   - Event data is a list of GUIDs of objects being dragged. For each one of them, it returns their position (x, y) and the data (id, x, y) of the contact used to drag them.
 * - *dndDragEvent*
 *   - Sent when dragging objects.
 *   - Event data is a list of GUIDs of objects being dragged. For each one of them, it returns their position (x, y) and the data (id, x, y) of the contact used to drag them.
 * - *dndDropEvent*
 *   - Sent when dropping objects.
 *   - Event data is a list of GUIDs of objects being dropped. For each one of them, it returns their position (x, y) and the data (id, x, y) of the contact used to drop them.
 * - *dndAbortEvent*
 *   - Sent when drag and drop operation on objects is aborted. For example: This can happend when object are dropped outside the page.
 *   - Event data is a list of GUIDs of objects that were dragged. For each one of them, it returns the data (id, x, y) of the contact used to drag them.
 *
 * @param {string} eventName Name of the event to listen to; see list above.
 * @param {function} callback Callback function on that event. Event data will be passed as a parameter to this function.
 * @returns {id} Observer ID.
 * @see NB.removeObserver
 */
NB.addObserver = function (eventName, block) {
    if (!isDefined(_eventObservers[eventName])) {
        _eventObservers[eventName] = [];
    }

    if ((_eventObservers[eventName].indexOf(block) < 0)) {
        _eventObservers[eventName].push(block);
		_eventObserverIds.push({"eventName": eventName, "block": block});
        DEBUG("addObserver: " + eventName + ": " + block.toString());
		return _eventObserverIds.length - 1;
    }
	// throw?
};

/**
 * Removes an observer based on its id.
 *
 * @param {int} index Observer ID of returned by {@link NB.addObserver}.
 * @see NB.addObserver
 */
NB.removeObserver = function(index) {
	if(index < _eventObserverIds.length) {
		var e = _eventObserverIds[index];
		var observer = e.block;
		var indexOfObserver = _eventObservers[e.eventName].indexOf(observer);
		if(indexOfObserver > -1) {
			_eventObservers[e.eventName].splice(indexOfObserver, 1);
		}
	}
};

/**
 * Internal widgetWillSuspend event handling.
 * We need to block to make sure things are done saving.
 * Once its done it will fire an event back to NBSDK server.
 * Then thats when we know it is okay to continue saving.
 */
_eventObservers["__NB_Internal_widgetsWillSuspend"] = [function(evt)
{
	DEBUG("__NB_Internal_widgetsWillSuspend: ");
	DEBUG(evt);
	
	var done = function() {	
		var data = {};
		NB.ajax("POST", "events/" + NB.HOSTGUID + "/finishedSuspending", data);
	};

	// Make sure something is listening to this event
	if(!isDefined(_eventObservers["widgetWillSuspend"])) {
		done();
		return;
	}

	if(_eventObservers["widgetWillSuspend"].length < 1) {
		done();
		return;
	}

	for (var i in _eventObservers["widgetWillSuspend"]) {
		DEBUG("  calling: " + _eventObservers["widgetWillSuspend"][i]);
		try {
			_eventObservers["widgetWillSuspend"][i](evt);
		} catch (e) {
			DEBUG("Observer threw an exception");
			throw e;
		}
	}

	// Finished things that were in suspending - post back to NB
	done();
}];

 /**
  * Adds an event to the queues for all event listeners, to allow inter-widget communication.
  *
  * To create data which will be sent to one particular object (if the object registers itself for the event), do:
  *
  * @param {string} eventName The name of the event.
  * @param {object} eventData The data for the event.
  *
  * @example
  * var eventData = {};
  * eventData[GUID-for-that-object] = {};
  * eventData[GUID-for-that-object]["name"] = value; // and so on, to put data under eventData[ann.guid]
  */
NB.postEvent = function(eventName, eventData) {
    var event = { 
        eventName : eventName,
        eventData: eventData};
        NB.ajax("POST", "events/postevent", event);
};


NB._eventObservers = _eventObservers;

// type: 'PUT', 'GET', 'DELETE', ...
// success_fun: function to call when ajax success
// error_func: function to call when ajax fail
// these two functions can be ignore, if you dont call about the feedback
// jsonString can be null, if not needed
// (Should have made this "private")
NB.ajax = function (type, url, jsonData, success_func, error_func) {
    serverPath = NB.SERVER + "/" + NB.VERSION + "/";
    serverPath += url;
    // Stringify the JSON data 
    var jsonString = JSON.stringify(jsonData);

    var ajax_error = function(error) {
		var thePath = serverPath;
		var theData = jsonString;
		DEBUG("NB.ajax " + type + " " + thePath + " with " + theData + " error: ");
        DEBUG(JSON.stringify(error));
        if(isDefined(error_func)) {
            error_func(error);
        }
    };
    
	var ajax_success = function(data) {
		if(isDefined(success_func) && success_func != null) {
			success_func(data);
		}
	};
	
    // Default JSON-request options.
    var params = {
        async: false,
        url: serverPath,
        type: type,
        processData: false,
        success: ajax_success,
        error: ajax_error
    };

    // The issue with setting contentType in the params is that jQuery tries to automatically
    // parse the return data into an object.  In the handlers we might not be setting this at all
    // and returning whatever json::jobject o; o.setNull(); o.toString() returns, which bails and
    // makes it call the error function even though we returned RouteStatusOK.
    // One way I am enforcing this is to say "if you passed in json, handlers MUST return json"
    // Also, if you are GETting data with this function, it MUST be json, so enforce that as well.
	if(isDefined(jsonData) || type == "GET") {
        params["contentType"] = 'application/json';
        params["data"] =  jsonString;
        params["dataType"] = 'json';
	}

    // Make the request.
    $.ajax(params);
};
 
 
var inBatchMode = false;
var batchLevel = 0;
var batchedOperations;

// Sets up the batchedOperations data structure
resetBatchedOperations = function () {
	batchedOperations = {};
	batchedOperations['PUT'] = {};
	batchedOperations['POST'] = {};
	batchedOperations['DELETE'] = {};
	batchedOperations['GET'] = {};
	batchLevel = 0;
} ();
 
 
var mergeBatchOperations = function(operations, data) {
	for(var key in data) {
		if (isDefined(operations[key])) {
			//If the operation for the next level of call already exists, continue merging
			mergeBatchOperations(operations[key], data[key]);
		} else {
			operations[key] = data[key];
		}
	}
};

// For every operation save it into the batched operations map
// If an update to a particular ID already exists merge the records.
var addBatchOperation = function (method, call, id, data) {
	if (!isDefined(batchedOperations[method][call])) {
		batchedOperations[method][call] = {};
	}

	if (!isDefined(batchedOperations[method][call][id])) {
		batchedOperations[method][call][id] = data;
	} else {
		// Already has an entry.. merge
		mergeBatchOperations(batchedOperations[method][call][id], data);
	}
};
 
/**
 * @desc Start a batch operation.
 * Any action defined between NB.beginBatch and {@link NB.endBatch}
 * will be delayed and executed once NB.endBatch gets called.
 * @deprecated You should use {@link NB.batch}.
 * @see NB.endBatch
 * @see NB.batch
 */
NB.beginBatch = function () {
	batchLevel++;
	inBatchMode = true;
};
 
/**
 * @desc End a batch operation.
 * Any action defined between {@link NB.beginBatch} and NB.endBatch
 * will be delayed and executed once NB.endBatch gets called.
 * @deprecated You should use {@link NB.batch}.
 * @see NB.beginBatch
 * @see NB.batch
 */
NB.endBatch = function () {
	batchLevel--;
	if(batchLevel <= 0) {
		inBatchMode = false;
		for (var method in batchedOperations) {
			if (Object.keys(batchedOperations[method]).length > 0) {
				for (var call in batchedOperations[method]) {
				DEBUG("Do a batch " + method + " call for: " + call);
				DEBUG(JSON.stringify(batchedOperations[method][call]));
				NB.ajax(method, call, batchedOperations[method][call]);
				}
			} else {
				DEBUG("No " + method);
			}
			batchedOperations[method] = {};
		}
	}
};

/**
* When manipulating properties of objects, it is immediately reflected in Notebook. 
* This is useful for animations or instantaneous updates. However, this could be slow 
* and is sequential. If you would like to update a large set of properties at once 
* (for example, per frame of a physics simulation), you can use the batch capability.
* 
* Wrap your world update in a function and pass it to NB.batch.
* This will start batching your property changes to object. 
*
* @param {function} block Operation to be batched.
* @param {function} [callback] Callback when batch operation is done.
* @example
* NB.batch(function() {
* 	objects = NB.page.getObjects();
* 	for(i in objects) {
* 		objects[i].rotation = Math.random() * 360.0; 
* 	}
* }, callbackWhenDone());
*/
NB.batch = function(op, callback)
{
	NB.beginBatch();
		try {
			op();
		} catch (e) {
			DEBUG(e);
		}
	NB.endBatch();

	if(isDefined(callback)) {
		callback();
	}
};

NB._ajaxBatchable = function (method, handler, obj, callForObj, propertyTree, success, error) {
	if (!inBatchMode) {
		NB.ajax(method, handler + "/" + obj.guid + callForObj, propertyTree, success, error);
	} else {
	// Batching
		addBatchOperation(method, handler + "/batch" + callForObj, obj.guid, propertyTree);
	}
};
 
NB._mapPropertyToHTTP = function (handler, obj, propertyTree, onSet) {
	prefix = "__$NB";

	propertyTreeArray = propertyTree.split(".");

	var childObj = obj;
	var data = {};
	var childData = data;
	for (var i = 0; i < propertyTreeArray.length - 1; i++) {
		if (!isDefined(childObj[propertyTreeArray[i]])) {
			// Not all objects support all properties.  We only want to map the properties that exist.
			// Ex. Do not map fill for line objects.
			return;
		}

		childObj = childObj[propertyTreeArray[i]];

		if (!isDefined(data[propertyTreeArray[i]])) {
			childData[propertyTreeArray[i]] = {};
		}
		childData = childData[propertyTreeArray[i]];
	}

	var property = propertyTreeArray[propertyTreeArray.length - 1];

	if (!isDefined(childObj[prefix])) childObj[prefix] = {};
	childObj[prefix][property] = childObj[property];

	childObj.__defineGetter__(property, function () {
										   return childObj[prefix][property];
										});

	childObj.__defineSetter__(property, function (newValue) {
										   childObj[prefix][property] = newValue;
										   childData[property] = newValue;
										   
										   if (onSet) {
											onSet();
										   }
										   
										   NB._ajaxBatchable('PUT', handler, obj, '/', data);
										});

    if (onSet) {
        onSet();
	}
};


})();
//
//   nb.addonmanager.js - NBSDK addons APIs
// (

(function() {
  
  if(!isDefined(this.NB)) { throw "NB not defined"; }
  var NB = this.NB;
  
/**
 * @ignore
 * @class 
 * @name NB.addonmanager
 * @desc 
 * This provides methods to handle the addons loaded.
 */
  NB.addonmanager = {};

 var mapPropertyAddonToHTTP = function (obj, propertyTree, onSet) {
	 NB._mapPropertyToHTTP("addon", obj, propertyTree, onSet);
 };
 
/**
 * Get list of addon Ids.
 * @ignore
 * @return {array} Addons Ids [ "addonId1", "addonId2", ... ].
 * @see NB.addonmanager.getAddons
 * @see NB.addonmanager.getAddon
 */
 NB.addonmanager.getAddonIds = function() {
	 var jsonObject;
	 var path = "addons/list/ids";
	 
	 var success = function (data) {
		 jsonObject = data;
	 };
	 var error = function (data) {
		 throw (data);
	 };
	 NB.ajax("GET", path, null, success, error);
	 
	 if ( !isDefined(jsonObject) ) {
		 throw "Invalid data";
	 }
	 
	 var addonIdList = [];
	 for(var i in jsonObject) {
		 addonIdList.push(jsonObject[i]);
	 }
	 
	 return addonIdList;
 };
 
 /**
  * Get addons objects.
  * @ignore
  * @return {array} Addons description.
  * @see NB.addonmanager.getAddonIds
  * @see NB.addonmanager.getAddon
  */
 NB.addonmanager.getAddons = function() {
 
	var jsonObject;
	var path = "addons/list/";

	var success = function (data) {
		jsonObject = data;
	};
	var error = function (data) {
		throw (data);
	};
	NB.ajax("GET", path, null, success, error);

	if ( !isDefined(jsonObject) ) {
		throw "Invalid data";
	}
 
	var addonsList = [];
	for(var guid in jsonObject) {
		addonsList.push(NB._attachAddonProperties(jsonObject[guid]));	
	}
 
	return addonsList;	
 };
 
 /**
  * @ignore
  * @class
  * @name addon
  * @desc Define the properties of an addon.
  */
 /**
  * @ignore
  * @fieldOf addon#
  * @name id
  * @type id
  * @desc Id of the addon.
  */
 /**
  * @ignore
  * @fieldOf addon#
  * @name name
  * @type string
  * @desc Name of the addon.
  */
 /**
  * @ignore
  * @fieldOf addon#
  * @name description
  * @type string
  * @desc Description of the addon.
  */
 /**
  * @ignore
  * @fieldOf addon#
  * @name version
  * @type string
  * @desc Version of the addon.
  */
 /**
  * @ignore
  * @fieldOf addon#
  * @name company
  * @type string
  * @desc Company that issued the addon.
  */
 /**
  * @ignore
  * @fieldOf addon#
  * @name minimum-version
  * @type string
  * @desc Minimum version of Notebook to run the addon.
  */
 /**
  * @ignore
  * @fieldOf addon#
  * @name filename
  * @type string
  * @desc Filename of the addon.
  */
 /**
  * @ignore
  * @fieldOf addon#
  * @name url
  * @type string
  * @desc Url to access the addon.
  */
 /**
  * @ignore
  * @fieldOf addon#
  * @name icon
  * @type string
  * @desc Filename of the icon of the addon.
  */
 /**
  * @ignore
  * @fieldOf addon#
  * @name accessAPI
  * @type string
  * @desc Access level used by the addon.
  */
 /**
  * @ignore
  * @fieldOf addon#
  * @name authors
  * @type array
  * @desc List of the authors who contributed to create the addon.
  */
 /**
  * @ignore
  * @fieldOf addon#
  * @name contact
  * @type addon.contact
  * @desc Contact information for the addon.
  */
 /**
  * @ignore
  * @class
  * @name addon.contact
  * @desc Contact information
  */
 /**
  * @ignore
  * @fieldOf addon.contact#
  * @name phone
  * @type string
  */
 /**
  * @ignore
  * @fieldOf addon.contact#
  * @name email
  * @type string
  */
 /**
  * @ignore
  * @fieldOf addon.contact#
  * @name mail
  * @type string
  */
 
 /**
  * @ignore
  * Get addon object.
  * @param {id} id Id of the addon.
  * @return {addon} Addon description.
  * @see NB.addonmanager.getAddonIds
  * @see NB.addonmanager.getAddons
  */
 NB.addonmanager.getAddon = function(id) {
 
	var jsonObject;
	var path = "addon/"+ id +"/";

	var success = function (data) {
		jsonObject = data;
	};
	var error = function (data) {
		throw (data);
	};
	NB.ajax("GET", path, null, success, error);

	if ( !isDefined(jsonObject) ) {
		throw "Invalid data";
	}

	return NB._attachAddonProperties(jsonObject);
 };
  
  /**
   * @ignore
  * Select an addon in the addon's tab.
  * @param {id} [id] Id of the addon to select. If omitted, the addon calling this method will select itself..
  * @see NB.addonmanager.isSelected
  */
 NB.addonmanager.select = function(id) {

	var path;

	if (isDefined(id)) {
		path = "addons/"+ id +"/select/";
	} else {
		path = "addon/select/";
	}
 
	var error = function (data) {
		throw (data);
	};

	NB.ajax("PUT", path, null, null, error);
 };
 
 // This is for future use (ex: when setting a property to addon)
 NB._attachAddonProperties = function (addon) {
 
	return addon;
 };
 
 /**
  * @ignore
  * Check if the addon calling is selected in the addon's tab.
  * @return {bool} true if the addon is selected, else false.
  * @see NB.addonmanager.select
  */
 NB.addonmanager.isSelected = function() {
	var bRet = false;
	var path = "addon/selected/";

	var success = function (data) {
		bRet = data;
	};
	var error = function (data) {
		throw (data);
	};
	NB.ajax("GET", path, null, success, error);

	return bRet;
 };

  var _onDragOver = null;
  var _onDrop = null;
  var _onDragExited = null;

  var _dndData = null;

  NB.addonmanager.dropState = { eDropNone: "None", eDropCopy: "Copy", eDropMove: "Move" };

  /**
  * @ignore
  * Attach handlers to support dragging onto the addon.
  * @param {function} onDragOver Callback for when a drag is happening. Function is provided one parameter with data and must return true or false if this is a valid drop area.
  * @param {function} onDrop Callback for when a drop has happened. Function is provided one parameter with data.
  * @param {function} onDragExited Callback for when a drag has exited.
  */
  NB.addonmanager.handleDnD = function (onDragOver, onDrop, onDragExited) {
    _onDragOver = onDragOver;
    _onDrop = onDrop;
	_onDragExited = onDragExited;
  };

  NB.addObserver("__NB_Internal_addonDragEnter", function (event) {
    _dndData = event.eventData;

    var dropState = NB.addonmanager.dropState.eDropNone;

    if (_onDragOver != null) {
      dropState = _onDragOver(_dndData);
    }

    NB.ajax('PUT', 'addon/' + NB.HOSTGUID + '/dropState/', dropState);
  });

  NB.addObserver("__NB_Internal_addonDragLeave", function () {
	if(_onDragExited != null) {
	  _onDragExited();
	}
    _dndData = null;
  });

  NB.addObserver("__NB_Internal_addonDragOver", function (event) {
    var dropState = NB.addonmanager.dropState.eDropNone;

    if (_dndData != null && _onDragOver != null) {
      _dndData.x = event.eventData.x;
      _dndData.y = event.eventData.y;

      dropState = _onDragOver(_dndData);
    }

    NB.ajax('PUT', 'addon/' + NB.HOSTGUID + '/dropState/', dropState);
  });

  NB.addObserver("__NB_Internal_addonDrop", function (event) {
    if (_onDrop == null) { return; }

    _onDrop(event.eventData);
  });
 
 })();
//
//   nb.annotation.js - NBSDK annoatation APIs
//

(function () {

  if (!isDefined(this.NB)) { throw "NB not defined"; }
  var NB = this.NB;

  var mapObjectPropertytoHTTP = function (obj, propertyTree, onSet) {
    NB._mapPropertyToHTTP("object", obj, propertyTree, onSet);
  };


  /**
  * Gets an object proxy for a given GUID. See {@link object} for more details.
  *
  * @param {string} guid GUID of the object to get
  * @returns {object} {@link object}
  * @example 
  * // Get an object
  * ann = NB.getObject(guid)
  * // moves ann to (100, 200)
  * ann.x = 100;
  * ann.y = 200;  
  */
  NB.getObject = function (guid) {
    var ann = null;

    var path = "object/" + guid + "/";
    NB.ajax('GET', path, null, function (data) { ann = data; }, function (data) { throw (data); });

    if (ann == null) return null;
    return NB._attachProperties(ann);
  };

  /**
  * Get the widget object which is calling.
  *
  * @returns {object} widget {@link object}
  */
  NB.getHostObject = function() {
	return NB.getObject(NB.HOSTGUID);
  };

  /**
  * Adds an object to the page based on the supplied prototype.
  *
  * Objects are added by generating a prototype using one of the NB.objectPrototype helper functions, then adding
  * the object to the page with addObject.  Once the prototype has been created, the properties can be manipulated
  * before the object is added to the page.
  *
  * @param {NB.objectPrototype} prototype An object returned by one the NB.objectPrototype.* methods.
  * @return {object} Object that was just created
  * @example 
   * <code>
  * // Generate the object prototype
  * var prototype = NB.objectPrototype.shape("hexagon");
  * prototype.stroke.color = [255, 0, 0];
  * prototype.stroke.thickness = 10;
  * // Add the object
  * NB.addObject(prototype);
   * </code>
  */
  NB.addObject = function (prototype) {
	if(prototype["type"] == "file") {
		return NB.file.addAsObjectWithPosition(prototype.token, prototype.x, prototype.y);
	}

    var guid = "";

    function onSuccess(data) {
      guid = data;
    }

    if (typeof (prototype) == "object") {
      NB.ajax('PUT', 'object/add/', prototype, onSuccess, null);
    } else {
      var data = { xml: prototype };
      NB.ajax('PUT', 'object/add/xml', data, onSuccess, null);
    }
    return NB.getObject(guid);
  };

  /**
  * @class
  * @name object
  * @desc 
  * Object Proxy 
  * -----
  * This is a javascript object that represents an object inside of Notebook. It is returned by methods that return "object".
  */

  /**
  * @fieldOf object#
  * @name guid
  * @type id
  * @desc The guid of the object. This is an id that can uniquely identify an object in Notebook.
  */

  NB._attachProperties = function (object) {
    /** 
	* @fieldOf object#
    * @name x
	* @type int
    * @desc x coordinate of the object, setting this property will move the object.
    */
    mapObjectPropertytoHTTP(object, "x");
    /** 
	* @fieldOf object#
    * @name y
	* @type int
    * @desc y coordinate of the object, setting this property will move the object.
    */
    mapObjectPropertytoHTTP(object, "y");
    /**
    * @fieldOf object#
    * @name width
	* @type int
    * @desc width of the object, setting this property will resize the object.
    */
    mapObjectPropertytoHTTP(object, "width");
    /**
	* @fieldOf object#
    * @name height
	* @type int
    * @desc height of the object, setting this property will resize the object.
    */
    mapObjectPropertytoHTTP(object, "height");
    /**
	* @fieldOf object#
    * @name rotation
	* @type int
    * @desc rotation of the object, in degrees, 0 - 360.  Setting this property will resize the object.
    */
    mapObjectPropertytoHTTP(object, "rotation");
    /**
	* @fieldOf object#
    * @name transparency
	* @type double
    * @desc transparency of the object, 0 - 1.0.  0 = fully transparent, 1 = fully opaque.
    * Setting this property will change the transparency of the object.
    */
    mapObjectPropertytoHTTP(object, "transparency");

    /**
	 * @fieldOf object#
	 * @name stroke
	 * @type object.stroke
	 * @desc Properties that pertain to the stroke (outline, border) of the object.
	 */
    /** 
    * @class
    * @name object.stroke
    * @desc Properties that pertain to the stroke (outline, border) of an object.
    */
    mapObjectPropertytoHTTP(object, "stroke.color", function () {
      /**
	  * @fieldOf object.stroke#
      * @name color 
      * @desc Specifies the stroke color of the object.
	  *
	  * Use the color format [r, g, b], where r,g,b are integers 0 - 255.
	  * @type color
      */
      decorateAsColor(object, "stroke.color");
    });

    /**
	* @fieldOf object.stroke#
    * @name thickness 
    * @desc Specifies the stroke thickness of the object, in pixels.
	* @type int
    */
    mapObjectPropertytoHTTP(object, "stroke.thickness");

	/**
	* @fieldOf object#
	* @name fill
	* @type object.fill
	* @desc Properties that pertain to the fill of the object.
	*/
    /** 
    * @class
    * @name object.fill
    * @desc Properties that pertain to the fill of an object.
    */
    mapObjectPropertytoHTTP(object, "fill", function () {
      attachPropertiesForFill(object);
    });

    /** 
	 * @methodOf object#
	 * @name moveForward
	 * @desc Moves the object forward.
	 */
    object.moveForward = function () { NB.page.moveForward([object]); };
    /** 
	 * @methodOf object#
	 * @name moveBackward
	 * @desc Moves the object backward.
	 */
    object.moveBackward = function () { NB.page.moveBackward([object]); };
    /** 
	 * @methodOf object#
	 * @name moveToFront
	 * @desc Moves the object to the front (topmost).
	 */
    object.moveToFront = function () { NB.page.moveToFront([object]); };
    /** 
	 * @methodOf object#
	 * @name moveToBack
	 * @desc Moves the object the back (bottommost).
	 */
    object.moveToBack = function () { NB.page.moveToBack([object]); };

    /**
	* @methodOf object#
	* @name toXML
	* @desc Returns XML representation of the object.
	* return {string} XML representation of the object.
    */
    object.toXML = function () {
      var xml = "";
      path = "object/" + object.guid + "/xml";
      var success = function (data) {
        xml = data;
      };
      var error = function (data) {
        throw (data);
      };
      NB.ajax('GET', path, null, success, error);
      return xml;
    };

	/**
	* @methodOf object#
	* @name update
	* @desc Updates the object with the latest properties from Notebook.
	*
	* You need to call this explicitly to get updated properties from Notebook. 
	*/
	object.update = function () {
		$.extend(true, this, NB.getObject(object.guid));
		return this;
	};

    /**
	* @methodOf object#
	* @name contains
	* @desc Tests if a point is contained within the bounds of the object.
    * @param {double} x x position of a point to test.
    * @param {double} y y position of a point to test.
    * @return {bool} True if the point is contained within the bounds.
    */
    object.contains = function (x, y) {
      return (x > this.x && x < (this.x + this.width) && y > this.y && y < (this.y + this.height));
    };

    /**
	* @methodOf object#
	* @name inflate
	* @desc Inflates the size of the object by a specified amount arount center point of the object.
    * @param {double} horizontal Horizontal amount to inflate.
    * @param {double} vertical Vertical amount to inflate.
    */
    object.inflate = function (width, height) {
      this.x -= width / 2;
      this.y -= height / 2;
      this.width += width;
      this.height += height;
    };

    /**
	* @methodOf object#
	* @name getChildren
	* @desc Get the children of the specified object.
    * @return {array} array of objects. Each object is a child object.
	* @see getChildrenIds
    */
    object.getChildren = function () {
      if (object.guid == "")
        throw ("id of the object incorrect");

      var dataJSON;
      var path = "object/" + object.guid + "/children/";

      var success = function (data) {
        dataJSON = data;
      };
      var error = function (data) {
        throw (data);
      };

      NB.ajax('GET', path, null, success, error);

      // Check json data
      if (!dataJSON) {
        throw "Invalid data";
      }

      var objectList = [];
      for (var childGuid in dataJSON) {
        objectList.push(NB._attachProperties(dataJSON[childGuid]));
      }

      return objectList;

    };

    /**
	* @methodOf object#
	* @name getChildrenIds
	* @desc Get the GUIDs of the children of the specified object.
    * @return {array} array of guids. One guid for one child.
	* @see getChildren
    */
    object.getChildrenIds = function () {
      if (object.guid == "")
        throw ("id of the object incorrect");

      var dataJSON;
      var path = "object/" + object.guid + "/children/guids/";

      var success = function (data) {
        dataJSON = data;
      };
      var error = function (data) {
        throw (data);
      };

      NB.ajax('GET', path, null, success, error);

      // Check json data
      if (!dataJSON || !dataJSON["guids"]) {
        throw "Invalid data";
      }

      return dataJSON["guids"];
    };


    if (isDefined(object.getPoints)) {
      /**
	  * @methodOf object#
	  * @name getPoints
	  * @desc Get the points of the object.
      * @return {array} Representation of the points.
      */
      object.getPoints = function () {
        if (object.guid == "") throw ("id of the object incorrect");

        var dataJSON;
        var path = "object/" + object.guid + "/points/";

        var success = function (data) {
          dataJSON = data;
        };
        var error = function (data) {
          throw (data);
        };

        NB.ajax('GET', path, null, success, error);

        // Check json data
        if (!dataJSON) {
          throw "Invalid data";
        }

        return dataJSON;
      };
    }

    if (isDefined(object.ungroup)) {
      /**
	  * @methodOf object#
	  * @name ungroup
	  * @desc Ungroup the object.
      * @return {object} Representation of the children.
      */
      object.ungroup = function () {
        var groupObject = NB.page.ungroupObjects([object]);

        return groupObject[object.guid];
      };
    }

    if (isDefined(object.getText)) {
      /**
	  * @methodOf object#
	  * @name getText
	  * @desc Get text from an object - both standard text and handwriting recognition from ink object.
      * @param {string} [langId] Language for Handwriting recognition. Possible values are: 
      * 				  { "en" => english US, "es" => spanish Spain, "fr" => french France, "de" => german Germany }.
      * @return {string} Text results.
      */
      object.getText = function (langId) {
        if (isDefined(object.text)) {
          return object.text;
        }

        if (object.guid == "") {
          throw ("id of the object incorrect");
		}

        var lang;
        if (langId === undefined) {
          lang = "default";
        } else {
          lang = langId;
        }

        var jsonReturn;
        var path = "object/" + object.guid + "/text/" + lang;

        var success = function (data) {
          jsonReturn = data;
        };
        var error = function (data) {
		  jsonReturn = "";
        };

        NB.ajax('GET', path, null, success, error);

		if ($.isArray(jsonReturn) && jsonReturn.length > 0) {
			return jsonReturn[0];
		}

        return "";
      };
    }

    if (isDefined(object.text)) {
      /** 
	  * @fieldOf object#
      * @name text
	  * @type string
      * @desc Properties that pertain to the text of an object.
	  * Defined only if object supports text.
      */
      mapObjectPropertytoHTTP(object, "text");

	  /**
	  * @fieldOf object#
	  * @name font
	  * @type object.font
	  * @desc Properties that pertain to the fill of the object.
	  * Defined only if object supports text.
	  */
      /**
	  * @class
	  * @name object.font
	  * @desc Properties that pertain to the font of a text object.
	  * Defined only if object supports text.
	  */

	  /**
	  * @fieldOf object.font#
      * @name color
	  * @type color
      */
      mapObjectPropertytoHTTP(object, "font.color", function () {
        decorateAsColor(object, "font.color");
      });

      /**
	  * @fieldOf object.font#
      * @name name
	  * @type string
      */
      mapObjectPropertytoHTTP(object, "font.name");
      /**
	  * @fieldOf object.font#
      * @name size
	  * @type int
      */
      mapObjectPropertytoHTTP(object, "font.size");
      /**
	  * @fieldOf object.font#
      * @name bold
	  * @type bool
      */
      mapObjectPropertytoHTTP(object, "font.bold");
      /**
	  * @fieldOf object.font#
      * @name italics
	  * @type bool
      */
      mapObjectPropertytoHTTP(object, "font.italics");
      /**
	  * @fieldOf object.font#
      * @name underline
	  * @type bool
      */
      mapObjectPropertytoHTTP(object, "font.underline");
    }

    if (isDefined(object.start)) {
      /**
	  * @fieldOf object#
      * @name start
	  * @type point
      * @desc x, y of the start point of a line object.
	  * Defined only for line objects.
	  * @example
	  * var objs = NB.page.getObjectsByType("line");
	  * for( i in objs ) {
	  *   objs[i].start.x = 10;
	  *   objs[i].start.y = 20;
	  * }
      */
      mapObjectPropertytoHTTP(object, "start.x");
      mapObjectPropertytoHTTP(object, "start.y");
    }

    if (isDefined(object.end)) {
      /**
      * @fieldOf object#
      * @name end
	  * @type point
      * @desc x, y of the end point of a line object.
	  * Defined only for line objects.
	  * @example
	  * var objs = NB.page.getObjectsByType("line");
	  * for( i in objs ) {
	  *   objs[i].end.x = 10;
	  *   objs[i].end.y = 20;
	  * }
      */
      mapObjectPropertytoHTTP(object, "end.x");
      mapObjectPropertytoHTTP(object, "end.y");
    }

    /**
	* @methodOf object#
	* @name setCustomProperty
	* @desc Sets a custom property.
	*
	* This property is saved with the object and the file and can be retrieved by object.getCustomProperty.
    * @param {string} key Key for the data.
    * @param {JSON object} data JSON data.
	* @see object.getCustomProperty
    */
    object.setCustomProperty = function (key, data) {
      NB.ajax('PUT', 'object/' + object.guid + '/customProperty/' + key + '/', data, null, null);
    };

    /**
	* @methodOf object#
	* @name getCustomProperty
	* @desc Reads a custom property.
    * @param {string} key Key for the data.
	* @param {JSON object} defaultData JSON data to return in case of failure.
    * @returns {JSON object} data JSON data.
	* @see object.setCustomProperty
    */
    object.getCustomProperty = function (key, defaultValue) {
      var value = defaultValue;

      var success = function (data) {
        if (data !== null) {
          value = data;
        }
      };

	  // I don't have a way of implementing batch GET right now with data inside of the URL.
	  //NB._ajaxBatchable('GET', 'object', object, '/customProperty/' + key + '/', null);

      NB.ajax('GET', 'object/' + object.guid + '/customProperty/' + key + '/', null, success, null);

      return value;
    };

    /**
	* @methodOf object#
	* @name addObserver
	* @desc Add observer for this object. 
	*
	* The callback will only be called if the event corresponds to this object. If it is a page event the callback will never be called.
    * @param {string} name Name of the event.
    * @param {function} callback Callback when event is sent.
	* @return {id} Observer ID.
	* @see NB.addObserver
	* @see NB.removeObserver
	* @see NB.postEvent
    */
    object.addObserver = function (eventName, block) {
      // Extend the callback to make sure we only call it for the current object
      var callback = function (evt) {
        if (isDefined(evt.eventData.guid) && evt.eventData.guid == object.guid) {
          // Filter event data for the current object
          var newevent = {};
          newevent.eventName = evt.eventName;
		  if (isDefined(evt.eventData[object.guid])) {
		    newevent.eventData = evt.eventData[object.guid];
		  } else {
		    newevent.eventData = evt.eventData;
		  }
          block(newevent);
        }
      };

      return NB.addObserver(eventName, callback);
    };

    object.addDecorator = function (xml, flags) {
      var id = -1;

      var onSuccess = function (data) {
        id = data;
      };

      NB.ajax('PUT', 'object/' + object.guid + '/decorator/', { xml: xml, flags: flags }, onSuccess, null);
	  // This does work but does not return ids because the API does not take a call back
	  //   NB._ajaxBatchable('PUT', "object", object, '/decorator/', {xml: xml, flags: flags}, onSuccess, null);

      return id;
    };

    object.removeDecorator = function (id) {
      NB.ajax('DELETE', 'object/' + object.guid + '/decorator/' + id + '/', null, null, null);
    };
 
	/**
	* @methodOf object#
	* @name invokeLink
	* @desc Invokes the link of the object, if it has one. 
	*
	* Only works if the object is on the current page.
	* @see page.invokeLink 
	*/
	object.invokeLink = function () {
		var url = "object/" + object.guid + "/invoke/link";
		NB.ajax('PUT', url);
	};
 
	/**
	* @methodOf object#
	* @name invokeSound
	* @desc Plays the sound attached to the object, if it has one.
	*
	* Only works if the object is on the current page.
	* @see page.invokeSound
	*/
	object.invokeSound = function () {
		var url = "object/" + object.guid + "/invoke/sound";
		NB.ajax('PUT', url);
	};
 

    /**
    * @fieldOf object#
    * @name animation
    * @type object.animation
    * @desc Properties that pertain to the animation of the object.
    * Defined only if object supports animation.
    */
    /** 
    * @class
    * @name object.animation
    * @desc Properties that pertain to the animation of a particular object.
    */
    if (isDefined(object.animation)) {
      /**
	  * @fieldOf object.animation#
      * @name type
	  * @type string
      * @desc
	  * Effect of the animation.
	  *
      * Possible values are:
	  *
      * - "Fade In"
      * - "Fade Out"
      * - "Flip around axis"
      * - "Fly In"
      * - "Fly Out"
      * - "Shrink and Grow"
      * - "Spin"
      */
      mapObjectPropertytoHTTP(object, "animation.type");
      /**
	  * @fieldOf object.animation#
      * @name direction
	  * @type string
      * @desc
	  * Direction of the animation, if applicable.
	  *
      * Depending of the type of animation, the direction can be set to the following strings:
	  * 
      * - *Fade In* 
      *    - no direction ("")
      * - *Fade Out*
      *    - no direction ("")
      * - *Flip around axis*
      *    - "Middle Vertical"
	  *    - "Middle Horizontal"
	  *    - "Left Edge"
	  *    - "Right Edge"
	  *    - "Top Edge"
	  *    - "Bottom Edge"
      * - *Fly In*
      *    - "Left"
	  *    - "Right"
	  *    - "Top"
      *    - "Bottom"
	  *    - "Top Left"
	  *    - "Top Right"
	  *    - "Bottom Left"
	  *    - "Bottom Right"
      * - *Fly Out*
      *    - "Left"
	  *    - "Right"
	  *    - "Top"
	  *    - "Bottom"
	  *    - "Top Left"
	  *    - "Top Right"
	  *    - "Bottom Left"
	  *    - "Bottom Right" 
      * - *Shrink and Grow*
      *    - no direction ("")
      * - *Spin*
      *    - "Clockwise"
	  *    - "Counter-Clockwise"
      */
      mapObjectPropertytoHTTP(object, "animation.direction");
      /**
	  * @fieldOf object.animation#
      * @name speed
	  * @type string
      * @desc Define the speed of the animation.
	  *
	  * Possible values are:
	  *
	  * - "Fast"
	  * - "Normal"
	  * - "Slow"
      */
      mapObjectPropertytoHTTP(object, "animation.speed");
      /**
	  * @fieldOf object.animation#
      * @name occurs
	  * @type string
      * @desc Define how the animation should start.
	  * 
	  * Possible values are:
	  *
	  * - "When the object is clicked"
	  * - "When the page is entered"
      */
      mapObjectPropertytoHTTP(object, "animation.occurs");
      /**
	  * @fieldOf object.animation#
      * @name repeats
	  * @type string
      * @desc Define how many times the animation should repeat itself.
      * 
	  * Possible values are:
	  *
	  * - "None"
	  * - "Until the object is clicked"
      * - "Forever"
	  * - "Once"
	  * - "Twice"
	  * - "Five times"
      * - "Ten times"
      */
      mapObjectPropertytoHTTP(object, "animation.repeats");

      /**
      * Start animation on the object
      * @param {object} [options] A map containing the following keys with values corrosponding to the fields above. 
      * @example
      * ann = NB.page.getSelectedObject();
      * ann.animation.start({type: "Spin", speed: "Fast"});
      */
      object.animation.start = function (animation) {
        var jsonData = {};

        if (isDefined(animation)) {
          jsonData.animation = animation;
        }

        NB._ajaxBatchable('POST', 'object', object, '/animation/start', jsonData);
      };

      /**
      * Stop animation on the object
      */
      object.animation.stop = function () {
        NB._ajaxBatchable('POST', 'object', object, '/animation/stop', null);
      };
    }

    /**
	* @fieldOf object#
    * @name lockType
    * @type string
    * @desc Define the lock state of the object.
	*
	* Value must be one of the following:
    * - "Unlock"
    * - "Lock In Place"
    * - "Allow Move"
    * - "Allow Move and Rotate"
    */
    mapObjectPropertytoHTTP(object, "lockType");

    /**
    * object._browser.zoom
	*/
    if (isDefined(object._browser) && isDefined(object._browser.zoom)) {
		mapObjectPropertytoHTTP(object, "_browser.zoom");
	}

	/**
	* @methodOf object#
	* @name deleteObject
	* @desc Deletes the object.
	*/
	object.deleteObject = function() {
      NB.ajax('DELETE', 'object/' + object.guid + '/');
	};

    return object;
  };

  function attachPropertiesForFill(obj) {
    /** 
	* @fieldOf object.fill#
    * @name color 
    * @desc Specifies the fill of the object.
	* 
	* Use the color format [r, g, b], where r,g,b are integers 0 - 255.
	* @type color
    */
    mapObjectPropertytoHTTP(obj, "fill.color", function () {
      decorateAsColor(obj, "fill.color");
    });
    /**
	* @fieldOf object.fill#
    * @name gradient 
    * @desc Specifies a linear gradient fill for the object.
    *
    * - Properties:
    *   - *startColor* in the color format [r, g, b], where r,g,b are integers 0 - 255.
    *   - *endColor* same as above.
	*
	* @type object
	* @example
	* var objs = NB.page.getObjectsByType("shape");
	* for( i in objs ) {
	*   objs[i].fill.gradient.startColor = [125, 45, 78];
	*   objs[i].fill.gradient.endColor = [60, 200, 125];
	* }
    */
    mapObjectPropertytoHTTP(obj, "fill.gradient.startColor", function () {
      decorateAsColor(obj, "fill.gradient.startColor");
    });
    mapObjectPropertytoHTTP(obj, "fill.gradient.endColor", function () {
      decorateAsColor(obj, "fill.gradient.endColor");
    });
  }

  function getObjectToDecorate(obj, keyHierarchy) {
    if (keyHierarchy === "") {
      return obj;
    }

    var subObj = obj;
    var keys = keyHierarchy.split(".");
    for (i = 0; i < keys.length; i++) {
      if (!subObj.hasOwnProperty(keys[i])) {
        return null;
      }
      subObj = subObj[keys[i]];
    }

    return subObj;
  }

  function decorateAsColor(o, key) {
    var obj = getObjectToDecorate(o, key);
    if (obj == null) {
      return;
    }

    function getHexComponent(val) {
      hexString = val.toString(16);
      if (hexString.length == 1) {
        hexString = "0" + hexString;
      }
      return hexString;
    }

    obj.getHexString = function () {
      return "#" + getHexComponent(obj[0]) + getHexComponent(obj[1]) + getHexComponent(obj[2]);
    };
  }

})();
//
//   nb.document.js - NBSDK document APIs
//
(function () {

if (!isDefined(this.NB)) {
	throw "NB not defined";
}
var NB = this.NB;

/**
 * @class 
 * @name NB.document
 * @desc 
 * Provides methods to manipulate pages for the current document.
 */
NB.document = {};

var mapPropertyPageToHTTP = function (obj, propertyTree, onSet) {
		NB._mapPropertyToHTTP("page", obj, propertyTree, onSet);
	};

/**
 * Get the current page Id.
 * @return {id} Id of the current {@link page}.
 * @see NB.document.getPageIds
 * @see NB.document.getPage
 */
NB.document.getCurrentPageId = function () {
	var jsonObject;
	var path = "pages/current/id";

	var success = function (data) {
			jsonObject = data;
		};

	NB.ajax("GET", path, null, success, null);

	return jsonObject;
};

/**
 * Get pages Ids.
 * @return {array} List of pages Ids.
 * @see NB.document.getCurrentPageId
 * @see NB.document.getPages
 */
NB.document.getPageIds = function () {
	var jsonObject;
	var path = "pages/list/ids";

	var success = function (data) {
			jsonObject = data;
		};
	var error = function (data) {
			throw (data);
		};
	NB.ajax("GET", path, null, success, error);

	if (!isDefined(jsonObject)) {
		throw "Invalid data";
	}

	var pageIdList = [];
	for (var i in jsonObject) {
		pageIdList.push(jsonObject[i]);
	}

	return pageIdList;
};

/**
 * Get pages objects.
 * @return {array} List of {@link page} objects.
 * @see NB.document.getPage
 * @see NB.document.getPageIds
 * @example
 * var pages = NB.document.getPages()
 * 
 * for( id in pages ) {
 *   alert(pages[id].name);
 * }
 */
NB.document.getPages = function () {

	var jsonObject;
	var path = "pages/list/";

	var success = function (data) {
			jsonObject = data;
		};
	var error = function (data) {
			throw (data);
		};
	NB.ajax("GET", path, null, success, error);

	if (!isDefined(jsonObject)) {
		throw "Invalid data";
	}

	var pagesList = [];
	for (var guid in jsonObject) {
		pagesList.push(NB._attachPageProperties(jsonObject[guid]));
	}

	return pagesList;
};

/**
 * Get page object from its Id.
 * @param {string} id Id of the page.
 * @return {page} Page object, else undefined if not found.
 * @see NB.document.getPages
 * @see NB.document.getCurrentPageId
 */
NB.document.getPage = function (id) {
	var jsonObject;
	var path = "page/" + id + "/";

	var success = function (data) {
			jsonObject = data;
		};
	var error = function (data) {
			throw (data);
		};
	NB.ajax("GET", path, null, success, error);

	if (!isDefined(jsonObject)) {
		return;
	}
	
	if (id != jsonObject.guid) {
		// Page not found but REST handler returned the current page
		return;
	}
	
	return NB._attachPageProperties(jsonObject);
};

/**
 * Add a page to the document.
 * @return {page} New page object added.
 * @see NB.document.deletePage
 */
NB.document.addPage = function () {
	var jsonObject;
	var path = "page/add/";

	var success = function (data) {
			jsonObject = data;
		};
	var error = function (data) {
			throw (data);
		};
	NB.ajax("POST", path, null, success, error);

	if (!isDefined(jsonObject)) {
		throw "Invalid data";
	}

	return NB._attachPageProperties(jsonObject);
};

/**
 * Deletes a page from the document.
 * @param {page} page Page to delete.
 * @see NB.document.addPage
 */
NB.document.deletePage = function (pageToDelete) {
	if (!isDefined(pageToDelete.guid)) throw "Invalid page";
	path = "page/" + pageToDelete.guid + "/";

	var guid = "";

	var success = function (data) {
			guid = data;
		};
	var error = function (data) {
			throw (data);
		};
	NB.ajax("DELETE", path, null, success, error);
};

/**
 * View next page in the document.
 * If the current page is the last one and is not empty, it will create a new page.
 * @see NB.document.viewPreviousPage
 */
NB.document.viewNextPage = function () {
	NB.ajax("PUT", "pages/view/next");
};

/**
 * View previous page in the document.
 * @see NB.document.viewNextPage
 */
NB.document.viewPreviousPage = function () {
	NB.ajax("PUT", "pages/view/previous");
};

NB._attachPageProperties = function (page) {
	/**
	 * @class 
	 * @name page
	 * @desc
	 * This is a javascript object that represents a page in Notebook, returned by NB.document methods like getPage.  
	 * @see NB.document
	 */

	/**
	 * @fieldOf page#
	 * @name name
	 * @type string
	 */
	mapPropertyPageToHTTP(page, "name");

	/**
	* @fieldOf page#
	* @name fill
	* @type page.fill
	* @desc Properties that pertain to the background of the page.
	*/
	/**
	 * @class
	 * @name page.fill
	 * @desc Properties that pertain to the background of a page.
	 */

	/**
	 * @fieldOf page.fill#
	 * @name color
	 * @desc
	 * Define a background color of the page.
	 *
	 * Use the color format [r, g, b], where r,g,b are integers 0 - 255.
	 * @type color
	 */
	mapPropertyPageToHTTP(page, "fill.color");

	/**
	* @methodOf page#
	* @name update
	* @desc Updates the page with the latest properties from Notebook.
	* 
	* You need to call this explicitly to get updated properties from Notebook. 
	*/
	page.update = function () {
		$.extend(true, this, NB.document.getPage(page.guid));
		return this;
	};

	/**
	 * @methodOf page#
	 * @name getObjects
	 * @desc Get the objects on the page.
	 * @return {map} Map of objects. Each key/value pair is a GUID/{@link object} on this page.
	 */
	page.getObjects = function () {
		if (page.guid == "") {
			throw ("id of the page incorrect");
		}

		var objectMap;
		NB.ajax("GET", "page/" + page.guid + "/objects/list/batch/", null, function (data) {
			objectMap = data;
		});

		for (var guid in objectMap) {
			objectMap[guid] = NB._attachProperties(objectMap[guid]);
		}

		return objectMap;
	};

	/**
	 * @methodOf page#
	 * @name getObjectsAsArray
	 * @desc Gets a list of objects on the page as an array.
	 * @return {array} List of objects on the page.
	 */
	page.getObjectsAsArray = function () {
		var objectMap = page.getObjects();

		var objectList = [];
		for (var guid in objectMap) {
			objectList.push(objectMap[guid]);
		}

		return objectList;
	};


	/**
	 * @methodOf page#
	 * @name deleteObject
	 * @desc Deletes an {@link object} on a page.
	 * @param {object} object Object to delete.
	 */
	page.deleteObject = function (anObject) {
		path = "page/" + page.guid + "/object/" + anObject.guid + "/";
		NB.ajax("DELETE", path);
	};

	/**
	 * @methodOf page#
	 * @name deleteObjectWithGuid
	 * @desc Deletes the {@link object} using its own GUID.
	 * @param {string} guid GUID of the object to delete.
	 */
	page.deleteObjectWithGuid = function (guid) {
		path = "page/" + page.guid + "/object/" + guid + "/";
		NB.ajax('DELETE', path);
	};

	/**
	 * @methodOf page#
	 * @name getObjectsByType
	 * @desc Returns the list of objects on the page by type.
	 * @param {string} type One of "ink", "shape", "image", "text", "line".
	 * @returns {array} List of objects on the page that match the type specified.
	 */
	page.getObjectsByType = function (type) {
		var json = [];
		var path = "page/" + page.guid + "/objects/type/" + escape(type) + "/";

		var success = function (data) {
				json = data;
			};
		var error = function (data) {
				throw (data);
			};
		NB.ajax("GET", path, null, success, error);
		return json;
	};

	/**
	 * @methodOf page#
	 * @name getSelectedObject
	 * @desc Returns the single selected object of the current page. If more than one object is selected, it will returns the first one available.
	 * 
	 * The behaviour is undefined if page is not the current visible page.
	 * @returns {object} Selected object or null.
	 * @deprecated You should use {@link page.getSelectedObjects}.
	 */
	page.getSelectedObject = function () {
		objects = page.getSelectedObjects();
		if (objects.length > 0) {
			return objects[0];
		} else {
			return null;
		}
	};

	/**
	 * @methodOf page#
	 * @name getSelectedObjects
	 * @desc Returns the list of selected objects on the current page.
	 * 
	 * The behaviour is undefined if page is not the current visible page.
	 * @returns {array} List of selected objects on the page.
	 */
	page.getSelectedObjects = function () {
		return page._getObjectsAtPath("objects/selected/");
	};

	/**
	 * @methodOf page#
	 * @name getObjectsAtPoint
	 * @desc Returns the list of objects on the page at a given coordinate.
	 * @param {float} x x coordinate.
	 * @param {float} y y coordinate.
	 * @returns {array} List of objects at the specified location.
	 */
	page.getObjectsAtPoint = function (x, y) {
		return page._getObjectsAtPath("page/" + page.guid + "/objects/point/" + x + "/" + y + "/");
	};

	page._getObjectsAtPath = function(path) {
		var guidJSON;

		var success = function (data) {
			guidJSON = data;
		};
		var error = function (data) {
			throw (data);
		};
		NB.ajax('GET', path, null, success, error);

		objectList = [];
		for (var i in guidJSON) {
			objectList.push(NB.getObject(guidJSON[i]));
		}

		return objectList;
	}

	/**
	 * @methodOf page#
	 * @name addObject
	 * @desc Adds an {@link object} to the page using its prototype.
	 * @param {NB.objectPrototype} prototype An object returned by one the NB.objectPrototype.* methods.
	 * @returns {object} Object added.
	 */
	page.addObject = function (obj) {
		return page.addObjectFromJson(obj);
	};

	/**
	* @methodOf page#
	* @name addObjectFromJson
	* @desc Adds an {@link object} to the page using its JSON definition.
	* @param {string} data An Notebook format json string.
	* @returns {object} Object added.
	*/
	page.addObjectFromJson = function (jsonData) {
		var path = "page/" + page.guid + "/object/add/";
		var res;
		var success = function (data) {
				res = data;
			};
		var error = function (data) {
				throw (data);
			};
		NB.ajax('PUT', path, jsonData, success, error);
		return NB.getObject(res);
	};


	/**
	 * @methodOf page#
	 * @name addObjectFromXML
	 * @desc Adds an {@link object} to the page using its XML definition.
	 * @param {string} xml An Notebook format XML string.
	 * @returns {object} Object added.
	 */
	page.addObjectFromXML = function (xml) {
		var jsonData = {};
		jsonData.xml = xml;
		var path = "page/" + page.guid + "/object/add/xml";
		var res;
		var success = function (data) {
				res = data;
			};
		var error = function (data) {
				throw (data);
			};
		NB.ajax('PUT', path, jsonData, success, error);
		return NB.getObject(res);
	};

	/**
	 * @methodOf page#
	 * @name getVisibleArea
	 * @desc Get the visible area of the page. 
	 * 
	 * If the page is not the current page, it will return the current page's visible area.
	 * @returns {object} Rectangle: {left, top, width, height}.
	 */
	page.getVisibleArea = function () {
		var dataJSON = "";
		path = "page/" + page.guid + "/visiblearea/";

		var success = function (data) {
				dataJSON = data;
			};
		var error = function (data) {
				throw (data);
			};
		NB.ajax('GET', path, null, success, error);

		if (dataJSON == "") throw "Invalid data";

		if (dataJSON.length != 4) throw "Invalid data";

		rect = {};
		rect.x = dataJSON[0];
		rect.y = dataJSON[1];
		rect.width = dataJSON[2];
		rect.height = dataJSON[3];

		return rect;
	};

	/**
	 * @methodOf page#
	 * @name moveToFront
	 * @desc Move the object(s) to the front in the z-order.
	 * @param {array} objects List of objects to move.
	 */
	page.moveToFront = function (objects) {
		var path = "page/" + page.guid + "/objects/movetofront/";
		NB.ajax('PUT', path, objects, null, null);
	};

	/**
	 * @methodOf page#
	 * @name moveToBack
	 * @desc Move the object(s) to the back in the z-order.
	 * @param {array} objects List of objects to move.
	 */
	page.moveToBack = function (objects) {
		var path = "page/" + page.guid + "/objects/movetoback/";
		NB.ajax('PUT', path, objects, null, null);
	};

	/**
	 * @methodOf page#
	 * @name moveForward
	 * @desc Move the object(s) forward one position in the z-order.
	 * @param {array} objects List of objects to move.
	 */
	page.moveForward = function (objects) {
		var path = "page/" + page.guid + "/objects/moveforward/";
		NB.ajax('PUT', path, objects, null, null);
	};

	/**
	 * @methodOf page#
	 * @name moveBackward
	 * @desc Move the object(s) backward one position in the z-order.
	 * @param {array} objects List of objects to move.
	 */
	page.moveBackward = function (objects) {
		var path = "page/" + page.guid + "/objects/movebackward/";
		NB.ajax('PUT', path, objects, null, null);
	};

	/**
	 * @methodOf page#
	 * @name groupObjects
	 * @desc Group a list of objects.
	 * @param {array} objects List of objects to group.
	 * @returns {object} The new grouped object.
	 */
	page.groupObjects = function (objects) {
		var dataJSON = [];
		var i;
		for (i in objects) {
			dataJSON.push(objects[i].guid);
		}

		path = "page/" + page.guid + "/objects/group/";

		var success = function (data) {
				dataJSON = data;
			};
		var error = function (data) {
				throw (data);
			};
		NB.ajax('POST', path, dataJSON, success, error);

		return NB.getObject(dataJSON);
	};

	/**
	 * @methodOf page#
	 * @name ungroupObjects
	 * @desc Ungroup a list of objects.
	 * @input {array} objects List of objects to ungroup.
	 * @returns {object} The new objects ungrouped.
	 */
	page.ungroupObjects = function (objects) {
		var dataJSON = [];
		var res = {};
		var i, j;

		for (i in objects) {
			dataJSON.push(objects[i].guid);
		}

		path = "page/" + page.guid + "/objects/ungroup/";

		var success = function (data) {
				dataJSON = data;
			};
		var error = function (data) {
				throw (data);
			};
		NB.ajax('POST', path, dataJSON, success, error);

		// Format response
		for (i in dataJSON) {
			res[i] = [];
			for (j in dataJSON[i]) {
				res[i].push(NB.getObject(dataJSON[i][j]));
			}
		}

		return res;
	};

	return page;
};

})();

//
//   nb.file.js - NBSDK File APIs
//

(function () {

if (!isDefined(this.NB)) { throw "NB not defined"; }

var NB = this.NB;

/**
 * @class 
 * @name NB.file
 * @desc 
 * Provides methods to upload files into SMART Notebook.
 */
NB.file = {};

/**
 * Adds a file upload control to a particular div element.
 * @param {DOM element} control jQuery element to add the file upload control to.
 * @param {function} onSuccess Callback success function.  Will return a token to that function when the file is sucessfully uploaded.
 */
NB.file.addUploadControl = function ($control, onSuccess) {
    var uniqueIdentifier = new Date().getTime();

    var iFrameId = 'iFrame' + uniqueIdentifier;
    var $iFrame = $('<iframe id="' + iFrameId + '" name="' + iFrameId + '" style="display:none;" />');
    $control.append($iFrame);

    var formId = 'form' + uniqueIdentifier;
    var $form = $('<form id="' + formId + '" name="' + formId + '" target="' + iFrameId + '" method="POST" action="' + NB.SERVER + '/' + NB.VERSION + '/file/upload/" enctype="multipart/form-data"></form>');

    $form.append($('<input type="file" id="file' + uniqueIdentifier + '" name="file' + uniqueIdentifier + '" />'));
    $form.append($('<input type="hidden" id="token' + uniqueIdentifier + '" name="token' + uniqueIdentifier + '" />'));

    var $button = $('<input type="button" value="Done" />');
    $button.click(function () {
        NB.file.uploadLocalResource($form, onSuccess);
    });
    $form.append($button);

    $control.append($form);
};

 /**
  * Get the file information.
  * @param {string} token File descriptor.
  * @param {function} onSuccess Callback function. Will return the file info to that function if successful.
  * @see NB.file.addUploadControl
  * @see NB.file.uploadExternalResource
  */
NB.file.getInfo = function (token, onSuccess) {
    $.ajax({
        url: NB.SERVER + '/' + NB.VERSION + '/file/' + token + '/json',
        success: function (jsonData) {
            fileInfo = JSON.parse(jsonData);
            onSuccess(fileInfo);
        },
        error: function (jqXHR, textStatus, errorThrown) {
        }
    });

};


NB.file.uploadLocalResource = function ($form, onSuccess) {
    var uniqueIdentifier = new Date().getTime();
    var token = 'file' + uniqueIdentifier;

    if ($form.children(':file')[0].value == '') {
        return;
    }

    $form.children(':hidden')[0].value = token;
    $form[0].submit();

    var $iFrame = $('#' + $form[0].getAttribute('target'));
    $iFrame.load(function () {
        // There is no current way to tell if the submit succeeded, so we query for the token.
        // Once the web server and api server use the same port, we will have the option to inspect the
        // server response to find out instead.  Then we can provide better error handling.
        NB.file.getInfo(token, function () {
            onSuccess(token);
        });

        $iFrame.unbind();
    });
};

 /**
  * Upload an external resource.
  * @param {string} url Link to load.
  * @param {function} onSuccess Callback success function. Will return a token to that function when the file is sucessfully uploaded.
  */
NB.file.uploadExternalResource = function (urlIn, onSuccess) {
    var uniqueIdentifier = new Date().getTime();
    var token = 'file' + uniqueIdentifier;

    var url = jQuery.trim(urlIn);
    
    if(url.indexOf("://") == -1) {
        // URL is a relative path, prepend our current location
        var path = document.location.href;
        url = path.substring(0, path.lastIndexOf("/")) + "/" + url;
    }
    
    var data = {
        token: token,
        url: url
    };

    $.ajax({
        type: 'POST',
        url: NB.SERVER + '/' + NB.VERSION + '/file/upload/',
        data: JSON.stringify(data),
        success: function () {
            onSuccess(token);
        }
    });

};

 /**
  * Add the uploaded file as an object on the page.
  * @param {string} token File descriptor.
  * @return {object} Object created.
  * @see NB.file.addAsObjectWithPosition
  */
NB.file.addAsObject = function (token) {
    
    var returnID;
    
    NB.ajax("POST",
        'file/add/' + token,
        "",
        function(evt) { returnID = evt; } 
        );
    
    return NB.getObject(returnID);
};

 /**
  * Add the uploaded file as an object on the page at a specific location.
  * @param {string} token File descriptor
  * @param {int} x x position.
  * @param {int} y y position.
  * @return {object} Object created.
  * @see NB.file.addAsObject
  */
NB.file.addAsObjectWithPosition = function (token, x, y) {

    var returnID;

    NB.ajax("POST",
        'file/add/' + token + '?x=' + x + '&y=' + y,
        "",
        function (evt) { returnID = evt; }
        );

    return NB.getObject(returnID);
};

})();
//
// 	nb.helpers.js - NBSDK helpers to implement commonly used functionality
//

(function() {

if (!isDefined(this.NB)) { throw "NB not defined"; }
var NB = this.NB;
/**
 * @class
 * @name NB.helpers
 */
NB.helpers = {};

NB._pointInElement = function(pt, elem) {
	var zoom = NB.page.getVisibleArea().width / 800.0;
	var hostObj = NB.getHostObject();

	var browserOffset = (hostObj.width - $(window).width()) / 2;

	var elemTop = elem.offset().top / zoom; // get div top
	var elemLeft = elem.offset().left / zoom;
	var elemWidth = elem.width() / zoom;
	var elemHeight = elem.height() / zoom;

	var x = pt.x - hostObj.x - browserOffset;
	var y = pt.y - hostObj.y - browserOffset;
	x = x / zoom;
	y = y / zoom;

	DEBUG([x, y, elemLeft, elemTop, elemWidth, elemHeight]);
	return (x > elemLeft && x < (elemWidth + elemLeft) && y > elemTop && y < (elemHeight + elemTop));
};

/**
* Creates a handwriting area inside of a widget.
*
* Available options
* ------
* - *activeBackgroundColor*
*   - Change the background color of the control when writing in it.
*   - Default value: null.
* - *activeBorderColor*
*   - Change the border color of the control when writing in it.
*   - Default value: "#18f".
* - *activeText*
*   - Display the specified text in the control when writing in it.
*   - Default value: null.
* - *onRecognizing*
*   - Callback when writing in the control.
*   - Default value: null.
* - *deleteWhenDone*
*   - Delete the ink object when recognized.
*   - Default value: true.
* - *timeToRecognize*
*   - Set the delay before trying to recognize the ink object.
*   - Default value: 2.
* - *wantsNumbers*
*   - Recognize numbers instead of words.
*   - Default value: false.
*
* @param {elem} elem jQuery object to make a handwriting area, for example a div element.
* @param {function} callback Function called when text is recognized.  Two parameters are passed, the raw Text and the final annotation.
* @param {options} options Map of options (see list above). 
* @example
*    NB.helpers.handwritable($("#area"), function(text, obj) { 	
*        console.log(text); 
*        console.log(obj); 
*        $("#area").text(text); 
*    }, { activeBackgroundColor: "#ddd", timeToRecognize: 1 });
*/
NB.helpers.handwritable = function(elem, callback, options) {
	var defaults = {
		activeBackgroundColor: null,
		activeBorderColor: "#18f",
		activeText: null,
		onRecognizing: null,
		deleteWhenDone: true,
		timeToRecognize: 2,
		wantsNumbers: false
	};

	// Override defaults
	var options = $.extend({}, defaults, options);

	var invalidTimer = -1;
	var currentWaitingTimer = invalidTimer;
	var alreadyInvalid = false;
	var writingInProgress = false;
	var writingStartedAtPoint = { x: 0, y: 0 };
	var inkingEventObserver = -1;	
	var inkingFinishedObserver = -1;

	var elemChangedProps = {};

	inkingEventObserver = NB.addObserver("inkingEvent", function (evt) {
		if(alreadyInvalid) return;

		if(currentWaitingTimer == invalidTimer) {
			if(NB._pointInElement({ x: evt.eventData.x, y: evt.eventData.y }, elem)) {
				if(!writingInProgress) {
					// We are recongizing... 
					writingInProgress = true;
					writingStartedAtPoint = evt.eventData;

					if(options.activeBorderColor) {
						elemChangedProps["border-color"] = elem.css("border-color");	
						elem.css("border-color", options.activeBorderColor);
					}

					if(options.activeBackgroundColor) {
						elemChangedProps["background-color"] = elem.css("background-color");	
						elem.css("background-color", options.activeBackgroundColor);
					}
	
					if(options.activeText) {
						elemChangedProps["text"] = elem.text();
						elem.text(options.activeText);
					}

					if(options.onRecognizing) {
						options.onRecognizing(elem);
					}
					// Redraw?
					// hostObj.x = hostObj.x + 1
				}
			}
		}

		alreadyInvalid = true;
	});

	inkingFinishedObserver = NB.addObserver("inkingFinishedEvent", function (evt) { 
		alreadyInvalid = false;

		if(!writingInProgress) {
			return;
		}

		// if its an inking finished outside of us we don't care, keep the timer going (as long as we have one)
		if(!NB._pointInElement({ x: evt.eventData.x, y: evt.eventData.y }, elem) && currentWaitingTimer != invalidTimer) {
			return;
		}
	
		// Fire a call back after 2 seconds of no waiting
		var secondsToWait = options.timeToRecognize;

		if(currentWaitingTimer != invalidTimer) {
			clearTimeout(currentWaitingTimer);
			currentWaitingTimer = invalidTimer;
		}

		currentWaitingTimer = setTimeout(onFinishedText, secondsToWait * 1000);
		DEBUG(evt.eventData.guid);
	}); 

	var	numFromStringMap = { 
			0: ["o"],  
			1: ["i", "l"], 
			2: ["z"], 
			3: [], 
			4: ["A"], 
			5: [], 
			6: ["b"], 
			7: ["T"], 
			8: [], 
			9: ["g", "q"] 
		};
	
	function normalizeToNumber(string) {
		// For each number look for its replacements and its uppercase equivilants and replace
		for(var num in numFromStringMap) {
			for(var index in numFromStringMap[num]) {
				string = string.replace(numFromStringMap[num][index], num);
				string = string.replace(numFromStringMap[num][index].toUpperCase(), num);
			}
		}

		return parseFloat(string);
	}

	function onFinishedText()
	{
		currentWaitingTimer = invalidTimer;
		writingInProgress = false;

		// Revert Properties
		if(options.activeBorderColor) elem.css("border-color", elemChangedProps["border-color"]);	
		if(options.activeBackgroundColor) elem.css("background-color", elemChangedProps["background-color"]);	
		if(options.activeText) elem.text(elemChangedProps["text"]);	

		// Get the annotation at the point that started the text entry
		var objs = NB.page.getObjectsAtPoint(writingStartedAtPoint.x, writingStartedAtPoint.y);
		if(objs.length < 1) return;
		var obj = objs[objs.length - 1];

		hostObj = NB.getHostObject();

		/*
		if(NB._pointInElement({ x: obj.x, y: obj.y }, elem)) {
			console.log("Got text but didn't hit");
			return;
		}
		*/

		var returnValue = obj.getText();

		// If they wanted numebrs normalize
		if(options.wantsNumbers) {
			returnValue = normalizeToNumber(returnValue);
		}

		// Call their callback function
		callback(returnValue, obj);

		if(options.deleteWhenDone) {
			// animate and remove
			obj.animation.start({ speed: "Fast", repeats: "None", type: "Fade Out" });
			obj.addObserver("annotationFinishedAnimatingEvent", function (evt) {
				var obj = NB.getObject(evt.eventData.guid);
				NB.page.deleteObject(obj);
			}); 
		}

		// hostObj.x = hostObj.x + 1;

		// remove old observers
		//  NB.removeObserver(inkingEventObserver);
		//  NB.removeObserver(inkingFinishedObserver);
	}
};

})();
//
//	nb.localization.js
//

(function () {
 
if (!isDefined(this.NB)) { throw "NB not defined"; }
var NB = this.NB;

/**
 * @class
 * @name NB.localization
 */
NB.localization = {};

var _currentLocale;
var _mapLocalization = {};

 /**
  * @class
  * @name locales.definition
  * @desc Define locale.
  */
 /**
  * @fieldOf locales.definition#
  * @name country
  * @type string
  * @desc Code of the country.
  */
 /**
  * @fieldOf locales.definition#
  * @name lang
  * @type string
  * @desc Code of the language.
  */
 /**
  * @class
  * @name locales.definition.detailed
  * @desc Define detailed properties for locales.
  * @augments locales.definition
  */
 /**
  * @fieldOf locales.definition.detailed#
  * @name isRtl
  * @desc Define if it is read from right to left.
  * @type bool
  */
 /**
  * @class
  * @name locales
  * @desc Define object to send/receive locales.
  */
 /**
  * @fieldOf locales#
  * @name inkRecognition
  * @desc Locales used for ink recognition.
  * @type locales.definition
  */
 /**
  * @fieldOf locales#
  * @name spellCheck
  * @desc Locales used for spell checking.
  * @type locales.definition
  */
 /**
  * @fieldOf locales#
  * @name ui
  * @desc Locales used for the user interface.
  * @type locales.definition.detailed
  */
 
  /**
  * Returns the locales for: UI, Spell check and Ink recognition
  * For a list of the codes returned, see "localization.html" for more details.
  * @return {locales} Object containing all locales settings.
  * @see NB.localize
  * @see NB.localizeFile
  */
  NB.localization.getLocales = function () {
  	var jsonObj;
  	var success = function (data) {
  		jsonObj = data;
  	};
  	var error = function (data) {
  	};
  	NB.ajax('GET', "settings/locales/", null, success, error);
  
  	return jsonObj;
  };
  
  /**
  * Translates paragraph and span elements within the html document that have been marked for translation into the current locale.
  * All translations must be saved under to corresponding JSON file included in the widget. See "localization.html" for more details.
  *
  * Marked paragraphs or span elements must have the following attribute set: rel="localize".
  * @example
  * &lt;p rel="localize"&gt;A paragraph to translate&lt;p&gt;
  * &lt;span rel="localize"&gt;A span to translate&lt;span&gt;
  * @see NB.localize
  * @see NB.localizeFile
  */
  NB.localization.translateHtmlElements = function () {
    // Reload page with new translations
    var i;
    var eltsList = $('p[rel^="localize"], span[rel^="localize"]');
    for (i in eltsList) {
      if (!eltsList[i].hasOwnProperty("englishKey")) {
        eltsList[i].englishKey = eltsList[i].innerHTML;
      }

      if (isDefined(_mapLocalization[eltsList[i].englishKey])) {
        eltsList[i].innerHTML = _mapLocalization[eltsList[i].englishKey];
      } else {
        eltsList[i].innerHTML = eltsList[i].englishKey;
      }
    }
  };

  var loadTranslations = function() {
    var locales = NB.localization.getLocales();
 
    if ( !locales ) {
      return;
    }
 
    // If the locale is different from the previous one, we cleanup the map.
    if ( _currentLocale != locales.ui.lang ) {
      _mapLocalization = {};
    }
 
    _currentLocale = locales.ui.lang;
    if (locales.ui.country.length > 0) {
      _currentLocale = _currentLocale + "_" + locales.ui.country;
    }

    var file = document.location.href.replace(/^.*\//, '');
    file = file.replace(/\..*/, '.json');

    $.ajax({
      url: "./_locales/" + _currentLocale + "/" + file,
      type: 'GET',
      contentType: 'application/json',
      dataType: 'json',
      async: false,
      success: function (jsonData) {
        // Extend the current map with new json object
        $.extend(_mapLocalization, jsonData);

        NB._fLocalizationReady = true;
		NB._checkLoaded();
      },
      error: function (jqXHR, textStatus, errorThrown) {
        NB._fLocalizationReady = true;
		NB._checkLoaded();
        // console.log("Unable to load the file: " + file);
      }
    });
  };

  loadTranslations();

  NB.addObserver("languageChangedEvent", function (event) {
    loadTranslations();
  });

/**
 * Translate a string.
 * @param {string} string String to translate.
 * @see NB.localizeFile
 */
NB.localize = function(originalString) {
	if ( !isDefined(originalString) ) {
		return;
	}
 
	if ( isDefined(_mapLocalization[originalString]) ) {
		return _mapLocalization[originalString];
	}
	
	return originalString;
};
 
/**
 * Get the path of the localized file.
 * @param {string} path Path of the file to localize.
 * @see NB.localize
 */
NB.localizeFile = function(path) {
	if ( !isDefined(path) ) {
		return;
	}
 
	return "./_locales/" + _currentLocale + "/" + path;
};


})();
//
//   nb.page.js - NBSDK page APIs
//

(function() {

/**
 * @class
 * @static
 * @name NB.page
 * @augments page
 * @desc NB.page object describes the current page object.  
 * It is equivalent to NB.document.getPage(NB.document.getCurrentPageID()), with additional 
 * methods for dealing with the current page.
 * @example
 *   var objects = NB.page.getObjects();
 *   var selectedObjects = NB.page.getSelectedObjects();
 *   var objectAdded = NB.page.addObject(NB.objectPrototype.text("Hello World!"));
 * @see page
 */

	var _getCurrentPage = function() {
		var page = NB.document.getPage(NB.document.getCurrentPageId());

		/**
		* @name NB.page.invokeLink
		* @function
		* @desc
		* Invokes the link of an object, if it has one. (Launches browser, page change, etc).  
		* Only works if the annotation is on the current page.
		* @param {string} guid GUID of the object to invoke the link on.
		*/
		page.invokeLink = function (guid) {
			var url = "object/" + guid + "/invoke/link";
			NB.ajax('PUT', url);
		};

		/**
		* @function
		* @name NB.page.invokeSound
		* @desc
		* Plays the sound attached to an object, if it has one.
		* Only works if the annotation is on the current page.
		* @param {string} guid GUID of the object to play the sound on.
		*/
		page.invokeSound = function (guid) {
			var url = "object/" + guid + "/invoke/sound";
			NB.ajax('PUT', url);
		};

		return page;
	};

	var _currentPage = {};
	_currentPage.guid = "";

	NB.__defineGetter__("page", function() {
		if(_currentPage.guid != NB.document.getCurrentPageId()) {
			_currentPage = _getCurrentPage();	
		}
		return _currentPage;
	});

})();

//
//   nb.persist.js - NBSDK page APIs
//

(function () {

    if (!isDefined(this.NB)) { throw "NB not defined"; }

    var NB = this.NB;

    /**
    * @class 
	* @static
    * @name NB.persist
	* @desc
	* NB.persist contains a set of APIs for persisting data to the file for a widget.  
	* Each widget has its own key-value store which they can access through these APIs.  
	* This key-value store is saved along with the file when the user saves, and is per widget.
	* For example, here is a call that saves and loads a score.
	* @example
	*   var globalScore = 0;
	*   NB.ready(function() {
	*     var keys = NB.persist.list();
	*     if (isDefined(keys) && keys.indexOf("score") == -1) {
	*       // score key does not exist, make one
	*       NB.persist.save("score", 0);
	*     }
	*     globalScore = NB.persist.load("score");
	*   });
	*
	*   function onScored() {
	*     globalScore++;
	*     NB.persist.save("score", globalScore);
	*   }
    */

    NB.persist = {};

    /**
    * Saves data for the current widget for the given key.  
	* The data is stored within the document and can later be retrieved using the same key.
    * @param {string} key Key for the data.
    * @param {JSON object} value Javascript object.
	* @see NB.persist.load
    */
    NB.persist.save = function (key, data) {
		var error = function (data) {
			throw (data);
		};
 
        NB.ajax('PUT', 'persist/' + NB.HOSTGUID + '/' + key + '/', JSON.stringify(data), null, error);
    };

    /**
    * Retrieves data for the current widget for a particular key.
    * @param {string} key Key for the data.
    * @returns {JSON object} Value saved under the specified key on success, undefined on failure.
	* @see NB.persist.save
    */
    NB.persist.load = function (key) {
        var value;

        var success = function (data) {
            value = data;
        };

        NB.ajax('GET', 'persist/' + NB.HOSTGUID + '/' + key + '/', null, success, null);

        return value;
    };

    /**
    * Removes data for the current widget for the given key.  The data at 'key' is removed.
    * @param {string} key Key for the data.
    */
    NB.persist.remove = function (key) {
		var error = function (data) {
			throw (data);
		};
 
        NB.ajax('DELETE', 'persist/' + NB.HOSTGUID + '/' + key + '/', null, null, error);
    };

    /**
    * Returns a list of all the keys of data stored within the current document for the widget.
    * @returns {array} Array of keys on success, null on failure.
    */
    NB.persist.list = function () {
        var value;

        var success = function (data) {
            value = data;
        };

        NB.ajax('GET', 'persist/' + NB.HOSTGUID + '/', null, success, null);

        return value;
    };

})();
//
//   nb.prototype.js - NBSDK object prototypes.
//

(function () {

    if (!isDefined(this.NB)) { throw "NB not defined"; }

    var NB = this.NB;

    var prototypeMap = {};

    /**
    * @class 
	* @static
    * @name NB.objectPrototype
	* @desc
	* Provides default prototypes for different kind of objects.
    * Prototypes can be used to simplify creating basic Notebook objects.  Once created, the resulting prototype will
    * have default properties applied to it that can then be easily configured before using {@link NB.addObject}.
	*
    * Optional properties can be used to create commonly styled objects. They can be used to override default properties (x, y, color, ...) and/or 
	* save your own properties.
    */
    NB.objectPrototype = {};

    var getPrototype = function (type, optionalProperties) {
    var prototype = "";
    if (prototypeMap.hasOwnProperty(type)) {
      prototype = prototypeMap[type];
    } else {
      var onSuccess = function(data) {
        prototype = data;
      };

      NB.ajax('GET', 'object/prototype/type/' + type + '/', null, onSuccess, null);
      prototypeMap[type] = prototype;
    }

    var instance = {};
    $.extend(true, instance, prototype, optionalProperties);

    return instance;
  };

	var _makeClass = function(){
	  return function(args){
		if ( this instanceof arguments.callee ) {
		  if ( typeof this.init == "function" )
			this.init.apply( this, args && args.callee ? args : arguments );
		} else
		  return new arguments.callee( arguments );
	  };
	};
  
    /**
	* @function
    * @name NB.objectPrototype.ink
    * @desc
    * Gets the prototype JSON for an ink object.
    * 
    * When creating an ink object, the prototype does not contain any points. Points are manually added
    * afterwards through the addPoint API.
    *
    * @param {object} [properties] Properties and styles to be automatically applied to the prototype.
    * @return {object} Prototype for the {@link object} to be created (supported properties).
    */
    NB.objectPrototype.ink = _makeClass();
	NB.objectPrototype.ink.prototype.init = function (optionalProperties) {
        var inkObj = getPrototype("ink", optionalProperties);

        inkObj.points = "";

        /**
        * @name NB.objectPrototype.ink.addPoint
        * Add a new point to the object.
        * @param {double} X X coordinate of new point.
        * @param {double} Y Y coordinate of new point.
        */
        inkObj.addPoint = function (x, y) {
            this.points += x + "," + y + " ";
        };

        $.extend(this, inkObj);
    };

    /**
    * @function
    * @name NB.objectPrototype.calligraphicInk
    * @desc
    * Gets the prototype JSON for a calligraphic ink object.
    * 
    * When creating a calligraphic ink object, the prototype does not contain any points.  Points are manually added
    * afterwards through the addPoint() API.
	* @example
	*	var line = new NB.objectPrototype.calligraphicInk();
	*	line.stroke.color = [55, 200, 100];
	*	line.stroke.thickness = 7;
	*
	*	for(var i = 0; i < 300; i+=2) {
	*		line.addPoint(i, 100 + (Math.sin(i / 30) * 100));
	*	}
	*
	*	annotation = NB.addObject(line);
    *
    * @param {object} [properties] Properties and styles to be automatically applied to the prototype.
    * @return {object} Prototype for the {@link object} to be created (supported properties).
    */
    NB.objectPrototype.calligraphicInk = _makeClass();
    NB.objectPrototype.calligraphicInk.prototype.init = function (optionalProperties) {
        var inkObj = getPrototype("calligraphicInk", optionalProperties);

        inkObj.points = "";

        /**
        * @name NB.objectPrototype.calligraphicInk.addPoint
        * Add a new point to the object.
        * @param {double} X x coordinate of new point.
        * @param {double} Y y coordinate of new point.
        */
        inkObj.addPoint = function (x, y) {
            this.points += x + "," + y + " ";
        };

		$.extend(this, inkObj);
    };

    /**
    * @function
    * @name NB.objectPrototype.line
    * @desc
    * Gets the prototype JSON for a line object.
    * 
    * @param {double} x1 x1 coordinate.
    * @param {double} y1 y1 coordinate.
    * @param {double} x2 x2 coordinate.
    * @param {double} y2 y2 coordinate.
    * @param {object} [properties] Properties and styles to be automatically applied to the prototype.
    * @return {object} Prototype for the {@link object} to be created (supported properties).
    */
	NB.objectPrototype.line = _makeClass();
    NB.objectPrototype.line.prototype.init = function (startX, startY, endX, endY, optionalProperties) {
        var lineObj = getPrototype("straightLine", optionalProperties);

        lineObj.start = { x: startX, y: startY };
        lineObj.end = { x: endX, y: endY };

		$.extend(this, lineObj);
    };

    /**
    * @function
    * @name NB.objectPrototype.curvedLine
    * @desc
    * Gets the prototype JSON for a curved line object.
    * 
    * @param {double} x1 x1 coordinate.
    * @param {double} y1 y1 coordinate.
    * @param {double} x2 x2 coordinate.
    * @param {double} y2 y2 coordinate.
    * @param {double} rx x radius.
    * @param {double} ry y radius.
    * @param {bool} arcFlag Large arc flag.
    * @param {bool} sweepFlag Sweep flag.
    * @param {object} [properties] Properties and styles to be automatically applied to the prototype.
    * @return {object} Prototype for the {@link object} to be created (supported properties).
    */
	NB.objectPrototype.curvedLine = _makeClass();
    NB.objectPrototype.curvedLine.prototype.init = function (startX, startY, endX, endY, radiusX, radiusY, largeArc, sweep, optionalProperties) {
        var curvedLineObj = getPrototype("curvedLine", optionalProperties);

        curvedLineObj.start = { x: startX, y: startY };
        curvedLineObj.end = { x: endX, y: endY };
        curvedLineObj.rx = radiusX;
        curvedLineObj.ry = radiusY;
        curvedLineObj.largeArc = largeArc;
        curvedLineObj.sweep = sweep;

		$.extend(this, curvedLineObj);
    };

    /**
    * @function
    * @name NB.objectPrototype.ellipse
    * @desc
    * Gets the prototype JSON for an ellipse object.
    * 
    * @param {double} cx Center x coordinate.
    * @param {double} cy Center y coordinate.
    * @param {double} rx X radius.
    * @param {double} ry Y radius.
    * @param {object} [properties] Properties and styles to be automatically applied to the prototype.
    * @return {object} Prototype for the {@link object} to be created (supported properties).
    */
    NB.objectPrototype.ellipse = _makeClass();
    NB.objectPrototype.ellipse.prototype.init = function (centerX, centerY, radiusX, radiusY, optionalProperties) {
        var ellipseObj = getPrototype("ellipse", optionalProperties);

        ellipseObj.cx = centerX;
        ellipseObj.cy = centerY;
        ellipseObj.rx = radiusX;
        ellipseObj.ry = radiusY;

        $.extend(this, ellipseObj);
    };

    /**
    * @function
    * @name NB.objectPrototype.shape
    * @desc
    * Gets the prototype JSON for a shape object.
    * 
    * A shape can be created by explicitly declaring the type by name from a predefined list, or a custom
    * shape can be created by ignoring this parameter.  If a custom shape is created, points are added to
    * the shape using the moveTo(), lineTo(), closeLine() APIs.
    *
    * @param {string} name parameter specifying a predefined shape.  Possible values are: arc, arrow, checkmark, cross, diamond, heart, hexagon, obtuse-triangle, parallelogram, pentagon, rectangle, right-triangle, rounded-rectangle, speech-bubble, star, thought-bubble, trapezoid, triangle.
    * @param {object} [properties] Properties and styles to be automatically applied to the prototype.
    * @return {object} Prototype for the {@link object} to be created (supported properties).
    */
    NB.objectPrototype.shape = _makeClass();
    NB.objectPrototype.shape.prototype.init = function (name, optionalProperties) {
        // There are two versions of the constructor allowed: with or without name.
        if (typeof (name) == "object") {
            optionalProperties = name;
            name = "";
        }

        var polygonObj = getPrototype("shape", optionalProperties);

        if (name != "" && name != undefined) {
            polygonObj.shapeName = name;
        } else {
            polygonObj.path = "";

            /**
			* @function
            * @name NB.objectPrototype.shape.moveTo
			* @desc
            * Move to a new point to start a new path.
            * @param {double} X x coordinate to move to.
            * @param {double} Y y coordinate to move to.
            */
            polygonObj.moveTo = function (x, y) {
                this.path += "M" + x + " " + y + " ";
            };

            /**
			* @function
            * @name NB.objectPrototype.shape.lineTo
			* @desc
            * Draw a line from the current point to the new point
            * @param {double} X x coordinate to draw a line to.
            * @param {double} Y y coordinate to draw a line to.
            */
            polygonObj.lineTo = function (x, y) {
                this.path += "L" + x + " " + y + " ";
            };

            /**
			* @function
            * @name NB.objectPrototype.shape.closeLine
			* @desc
            * Draw a line from the current point to the starting point of the current path.
            */
            polygonObj.closeLine = function () {
                this.path += "z";
            };
        }

		$.extend(this, polygonObj);
    };

    /**
    * @function
    * @name NB.objectPrototype.text
    * @desc
    * Gets the prototype JSON for a text object.
    * 
    * @param {string} text Text to display.
    * @param {object} [properties] Properties and styles to be automatically applied to the prototype.
    * @return {object} Prototype for the {@link object} to be created (supported properties).
    */
	NB.objectPrototype.text = _makeClass();
    NB.objectPrototype.text.prototype.init = function (text, optionalProperties) {
        var textObj = getPrototype("text", optionalProperties);

        textObj.text = text;

		$.extend(this, textObj);
    };

    /**
    * @function
    * @name NB.objectPrototype.file
	* @desc
    * Gets the prototype JSON for an external file.  
	* 
	* This allows you to add arbitary content to Notebook.  If Notebook understands the format it will
	* create an object for it.  Otherwise, it will be added to the attachments and a link object will 
	* be created for it.  Objects can only be added to the current page, using {@link NB.addObject}.
    * @param {path} path relative path to file inside of a widget or an URL to external content.
    * @return {object} Prototype for the {@link object}.
	* @example
	*	var png = new NB.objectPrototype.file("testImage.png");
	*	var webpng = new NB.objectPrototype.file("http://www.someserver.net/anything.png");
	*	var gif = new NB.objectPrototype.file("imgs/test.gif");
	*	var widget = new NB.objectPrototype.file("test.galleryitem");
	*
	*	png.x = 20; png.y = 30;
	*	webpng.x = 40; webpng.y = 40;
	*	gif.x = 100; gif.y = 100;
	*	widget.x = 200; widget.y = 200;
	*
	*	NB.addObject(png);
	*	NB.addObject(webpng);
	*	NB.addObject(gif);
	*	NB.addObject(widget);
    */
	NB.objectPrototype.file = _makeClass();
    NB.objectPrototype.file.prototype.init = function (path) {
		var proto = {};
		proto.x = 0;
		proto.y = 0;
		proto.type = "file";

		var uniqueIdentifier = new Date().getTime();
		var token = 'file' + uniqueIdentifier;

		data = { token: token };

		if (path.indexOf("http://") > -1) {
			data.url = path;
		} else {
			data.path = "widget:" + NB.HOSTGUID + "/" + path;
		}

		NB.ajax("POST", "file/upload/", data);

		proto.token = token;
		$.extend(this, proto);
	};
})();
//
//   nb.settings.js - NBSDK page APIs
//

(function () {

if (!isDefined(this.NB)) { throw "NB not defined"; }

var NB = this.NB;

/**
 * @class 
 * @static
 * @name NB.settings
 */
NB.settings = {};

/**
 * By default, widgets will render everything into Notebook, including body background color.
 * Because you cannot set a rgba CSS property on the body element, we provide an API which you
 * can use to mark areas of your widget to be transparent and click-through.
 * Whenever this color appears in your widget, it will be transparent.
 * @param {int} red
 * @param {int} green
 * @param {int} blue
 * @see NB.settings.disableTransparentColorMask
 */
NB.settings.enableTransparentColorMask = function (red, green, blue) {	
	var data = {};
	data["enable"] = true;
	data["transparentColor"] = [red, green, blue];
	
    NB.ajax('PUT', 'settings/hostWidget/setTransparentColor/', data);
};

/**
 * Disables the color mask.
 * @see NB.settings.enableTransparentColorMask
 */
NB.settings.disableTransparentColorMask = function () {	
	var data = {};
	data["enable"] = false;
	data["transparentColor"] = [255, 255, 0];
	
    NB.ajax('PUT', 'settings/hostWidget/setTransparentColor/', data);
};

/**
 * @name NB.settings.uid
 * @desc 
 * An unique id which uniquely identifies a machine.
 * @type id
 */
NB.settings.__defineGetter__("uid", function() {
	var data; 
	NB.ajax('GET', 'settings/uid/', null, function(d) { data = d; });
	return data["uid"];
});

})();

//
//   nb.ui.js - NBSDK UI APIs
//

(function () {

  if (!isDefined(this.NB)) {
    throw "NB not defined";
  }

  var NB = this.NB;

  /**
  * @class 
  * @static
  * @name NB.ui
  */
  NB.ui = {};

  /**
  * Sets a status message at the top of the current page.
  * @param {string} status Message to be displayed.
  * @returns {id} id of the message that was added.
  * @see NB.ui.clearStatusMessage
  */
  NB.ui.setStatusMessage = function (status) {
    var statusId = -1;
    function onSuccess(data) {
      statusId = data;
    }
    NB.ajax('PUT', "page/current/ui/statusmessage/", status, onSuccess, null);
    return statusId;
  };

  /**
  * Removes a particular status messagefrom the current page.
  * @param {id} id Id of the message to remove.
  * @see NB.ui.setStatusMessage
  */
  NB.ui.clearStatusMessage = function (id) {
    NB.ajax('DELETE', "page/current/ui/statusmessage/" + id + "/", null, null, null);
  };

})();
//
//	nb.undomanager.js
//

(function () {
 
if (!isDefined(this.NB)) { throw "NB not defined"; }
var NB = this.NB;

/**
 * @class
 * @name NB.undomanager
 * @desc 
 * This object provides access to the undo manager of Notebook. Use it to 
 * add a record which can be undone by the user inside the software.
 * By default any undoable action (create ink object, change property, ...) are recorded. 
 * This object helps you to create more complex records.
 */
NB.undomanager = {};

/**
 * Start an undo/redo record. Every undoable actions performed between NB.undomanager.startRecord and {@link NB.undomanager.stopRecord}
 * will be part of only one undo record. When the user will undo the record, all these actions will be undone at the same time.
 * It finalizes the previous record if it has not been done.
 * The auto undo record is temporarly disabled (if enabled) until NB.undomanager.stopRecord is called.
 * @param {number|string} recordType Description of the record. It can be a string or an integer defined in {@link NB.undomanager.undotype}.
 * @see NB.undomanager.stopRecord
 * @example
 * // Update the item "Undo" under the "Edit" menu 
 * // with a predefined string, ex: "Undo Resize"
 * // The predefined string are translated according to the locale of the UI.
 * NB.undomanager.startRecord(NB.undomanager.undotype.eResize);
 * 
 * // or
 * 
 * // Update the item "Undo" under the "Edit" menu 
 * // with your own string, ex: "Undo my undo record".
 * // There is no translation when using your own string.
 * NB.undomanager.startRecord("my undo record"); 
 *
 * // See NB.undomanager.stopRecord for a complete example.
 */
NB.undomanager.startRecord = function( recordType ) {

	var bRet = false;
	var sPath = "undomanager/startrecord/";
    var jsonData = {};

    if ( recordType == undefined ) {
    	jsonData.string = "";
    }
    else {
        var index = this.undotype.indexOf( recordType );
        if ( index >= 0 )
        	jsonData["enum"] = index;
        else
        	jsonData.string = recordType;
    }

	var success = function ( data ) {
            bRet = data;
	};
	var error = function ( data ) {
            throw (data);
	};
	NB.ajax('POST', sPath, jsonData, success, error);
 
	return bRet;
};

/**
 * Stop an undo/redo record. The current record is finalized and saved.
 * The undo/redo record will not be available to the user until this function gets called.
 * {@link NB.undomanager.startRecord} must have been called previously. 
 * @see NB.undomanager.startRecord
 * @example
 * NB.undomanager.startRecord("the widget's record");
 *
 * // Modify fill color of all shapes of the current page
 * // Notebook will perform these actions and add them to the current record.
 * var objs = NB.page.getObjectsByType("shape");
 * for( i in objs ) {
 *   objs[i].fill.color = [255, 125, 0];
 * }
 *
 * // Modify the color of a div (id is "myDiv") inside the current widget
 * // To make this modification undoable,
 * // we need to do it through NB.undomanager.performUndoableAction
 * var oldColor = $('#myDiv').css('background-color');
 * NB.undomanager.performUndoableAction(
 *   function () {
 *     $('#myDiv').css('background-color', '#FF7D00');
 *   },
 *   function () {
 *     $('#myDiv').css('background-color', oldColor);
 *   } );
 *
 * // Finalize the record
 * NB.undomanager.stopRecord();
 */
NB.undomanager.stopRecord = function () {
 
	var bRet = false;
	var sPath = "undomanager/stoprecord/";
	 
	var success = function ( data ) {
        bRet = data;
	};
	var error = function ( data ) {
        throw (data);
	};
	NB.ajax('POST', sPath, null, success, error);
	 
	return bRet;
};

/**
 * Rename the current undo/redo record.
 * {@link NB.undomanager.startRecord} must have been called previously.
 * @param {number|string} recordType Description of the record. It can be a string or an integer defined in {@link NB.undomanager.undotype}.
 * @example
 * // Update the item "Undo" under the "Edit" menu 
 * // with a predefined string, ex: "Undo Resize"
 * // The predefined string are translated according to the locale of the UI.
 * NB.undomanager.startRecord(NB.undomanager.undotype.eResize);
 * 
 * // or
 * 
 * // Update the item "Undo" under the "Edit" menu 
 * // with your own string, ex: "Undo my undo record".
 * // There is no translation when using your own string.
 * NB.undomanager.startRecord("my undo record");
 */
NB.undomanager.renameRecord = function ( recordType ) {
 
	var bRet = false;
	var sPath = "undomanager/renamerecord/";
    var jsonData = {};

    if ( recordType == undefined ) {
    	jsonData.string = "";
    }
    else {
        var index = this.undotype.indexOf( recordType );
        if ( index >= 0 )
        	jsonData["enum"] = index;
        else
        	jsonData.string = recordType;
    }
	
	var success = function ( data ) {
        bRet = data;
	};
	var error = function ( data ) {
        throw (data);
	};
	NB.ajax('POST', sPath, jsonData, success, error);
	 
	return bRet;
};

/**
 * Add a custom undo/redo action to the current record and perform the specified action.
 * Note: startRecord must have been called previously.
 * @param {function} action Functon to execute. 
 * @param {function} callbackUndo Function to call when undo is performed.
 * @example
 * var oldColor = $('#myDiv').css('background-color');
 *
 * NB.undomanager.performUndoableAction(
 *   function() {
 *     // Executed with NB.undomanager.performUndoableAction and when user hit "Redo"
 *     $('#myDiv').css('background-color', '#555');
 *   },
 *   function() {
 *     // Executed when user hit "Undo"
 *     $('#myDiv').css('background-color', oldColor);
 *   } );
 */
NB.undomanager.performUndoableAction = function ( action, callbackUndo ) {
  if (typeof callbackUndo !== "function" || typeof action !== "function")
    throw ("Parameters incorrect");
 
  // Add callback to the map
  var bRet = false;
  var callbacks = _callbackmanager.set(callbackUndo, action);
  var jsonData = {};
  var sPath = "undomanager/addaction/";
 
  jsonData.callbacks = callbacks;
 
  var success = function (data) {
    bRet = data;
  };
  var error = function (data) {
    throw (data);
  };
  NB.ajax('POST', sPath, jsonData, success, error);
 
  action();
 
  return bRet;
};


/**
 * Undo the last record.
 * @see NB.undomanager.redo
 */
NB.undomanager.undo = function () {
 
	var bRet = false;
	var sPath = "undomanager/undo/";

	var success = function ( data ) {
        bRet = data;
	};
	var error = function ( data ) {
        throw (data);
	};
	NB.ajax('POST', sPath, null, success, error);

	return bRet;
};

/**
 * Redo the last record undone.
 * @see NB.undomanager.undo
 */
NB.undomanager.redo = function () {
 
	var bRet = false;
	var sPath = "undomanager/redo/";

	var success = function ( data ) {
        bRet = data;
	};
	var error = function ( data ) {
        throw (data);
	};
	NB.ajax('POST', sPath, null, success, error);

	return bRet;
};

/**
 * Enable or disable the auto undo/redo record. By default the auto record is enabled.
 * @param {bool} enable true to enable, false to disable.
 */
NB.undomanager.setAutoRecord = function ( enable ) { 
	var bRet = false;
	var jsonData = {};
	var sPath = "undomanager/autorecord/";
	
	jsonData.autorecord = enable;
	 
	var success = function ( data ) {
        bRet = data;
	};
	var error = function ( data ) {
        throw (data);
	};
	NB.ajax('POST', sPath, jsonData, success, error);

	return bRet;
};

/**
 * NB.undomanager.callbackmanager
 */
var _callbackmanager = {};
var _callbacks = {};
var _last_id = 0;

/**
 * Save the callbacks for undo/redo with a new id.
 * param {function} undoCallback: Function called when undo is performed.
 * param {function} redoCallback: Function called when redo is performed.
 */
_callbackmanager.set = function( undoCallback, redoCallback ) {
    ++_last_id;
    _callbacks[_last_id] = {};
    _callbacks[_last_id].undo = undoCallback;
    _callbacks[_last_id].redo = redoCallback;
    return _last_id;
};

// Add Undo Observer
NB.addObserver("undoEvent", function(evt) {
    if ( evt.eventData.id !== undefined ) {
        _callbacks[evt.eventData.id].undo();
    }
});

// Add Redo Observer
NB.addObserver("redoEvent", function(evt) {
    if ( evt.eventData.id !== undefined ) {
        _callbacks[evt.eventData.id].redo();
    }
});

/**
 * Defines constants to identify undo/redo records.
 * 
 * List of values:
 *
 * - *eNoUndoRedoName*
 *   - No undo redo name
 * - *eColor*
 *   - Modify stroke color
 * - *eFillColor*
 *   - Modify fill color
 * - *eLineThickness*
 *   - Modify line thickness
 * - *eHyperlink*
 *   - Add/Delete hyperlink
 * - *eLockInPlace*
 *   - Lock/Unlock in place
 * - *eTextChange*
 *   - Modify text
 * - *eTextItalic*
 *   - Text italic
 * - *eTextBold*
 *   - Text bold
 * - *eTextUnderline*
 *   - Text underlined
 * - *eTextFontChange*
 *   - Modify text font
 * - *eLineEnds*
 *   - Modify line ending
 * - *eMirrored*
 *   - Selection mirrored
 * - *eClearPage*
 *   - Clear Page
 * - *eText*
 *   - Add Text
 * - *ePaste*
 *   - Paste clipboard
 * - *eErasing*
 *   - Erasing
 * - *eWriting*
 *   - Add ink annotation
 * - *eShape*
 *   - Add shape annotation
 * - *eLine*
 *   - Add line annotation
 * - *eClone*
 *   - Clone selection
 * - *eBackgroundTransparency*
 *   - Modify background transparency
 * - *eOrder*
 *   - Modify z-order
 * - *eDelete*
 *   - Delete annotation/page
 * - *eUngroup*
 *   - Ungroup annotaitons
 * - *eGroup*
 *   - Group annotations
 * - *eRecognize*
 *   - Handwriting recognization
 * - *eResize*
 *   - Resize annotation
 * - *eRotate*
 *   - Rotate annotation
 * - *eMoveRotationPoint*
 *   - Move rotation point point
 * - *eMove*
 *   - Move selection
 * - *eDrop*
 *   - Drop selection
 * - *eRemoveFromContainer*
 *   - Remove from container
 * - *eTransparency*
 *   - Modify tranparency
 * - *eRenamePage*
 *   - Rename page
 * - *eDragAndDrop*
 *   - Drag and drop
 * - *eDashStyle*
 *   - Modify dash style
 * - *eBackgroundColor*
 *   - Modify background color
 * - *eTableCellColor*
 *   - Modify table cell color
 * - *ePropertyChanges*
 *   - Property changed
 * - *eDeleteAttachment*
 *   - Delete attachment
 * - *eAddAttachment*
 *   - Add attachment
 * - *eAddPage*
 *   - Add page
 * - *eClonePage*
 *   - Clone page
 * - *eDeletePage*
 *   - Delete page
 * - *eDragCopyPage*
 *   - Drag copy page
 * - *eScreenCapture*
 *   - Screen capture
 * - *eScreenShade*
 *   - Screen shade
 * - *eAddTemplate*
 *   - Add template
 * - *eSound*
 *   - Sound
 * - *eVideo*
 *   - Video
 * - *eSoundLaunchType*
 *   - Sound launch type
 * - *eScreenShadeMove*
 *   - Move sreen shade
 * - *eScreenShadeOpen*
 *   - Screen shade open
 * - *eScreenShadeClose*
 *   - Screen shade close
 * - *eInfiniteClone*
 *   - Infinite cloner
 * - *eCut*
 *   - Cut the selection
 * - *eInsertTable*
 *   - Insert table
 * - *eMergeCells*
 *   - Merge cells
 * - *eSplitCell*
 *   - Split cells
 * - *eDeleteColumn*
 *   - Delete column
 * - *eInsertColumn*
 *   - Insert column
 * - *eDeleteRow*
 *   - Delete row
 * - *eInsertRow*
 *   - Insert row
 * - *eMoveSeperator*
 *   - Move separator
 * - *eAdjustCellsSize*
 *   - Adjust cells size
 * - *eInsertQuestion*
 *   - Insert question
 * - *eInsertTitlePage*
 *   - Insert title page
 * - *eInsertAnswerKey*
 *   - Insert answer key
 * - *eEditAnswer*
 *   - Edit answer
 * - *eInsertChoice*
 *   - Insert choice
 * - *eStartTest*
 *   - Start test
 * - *eFinishVoting*
 *   - Finish voting
 * - *eClearResults*
 *   - Clear results
 * - *eAlignAnnotations*
 *   - Align annotations
 * - *eAddPageGroup*
 *   - Add page group
 * - *eThemeApplied*
 *   - Theme applied
 * - *eDeletePageGroup*
 *   - Delete page group
 * - *eReorderPageGroup*
 *   - Reorder page group
 * - *eRenameGroup*
 *   - Rename group
 * - *eTableShade*
 *   - Table shade
 * - *eRemoveCells*
 *   - Remove cells
 * - *eScreenShadeMaximize*
 *   - Screen shade maximize
 * - *ePageRecordingChanged*
 *   - Page reordering changed
 * - *eObjectAnimationChanged*
 *   - Object animation changed
 * - *eThemeDelete*
 *   - Them deleted 
 * - *eInsertProtractor*
 *   - Insert protractor
 * - *eInsertRuler*
 *   - Insert ruler
 * - *eInsertCompass*
 *   - Insert compass
 * - *eRulerWidth*
 *   - Ruler width
 * - *eRulerUnits*
 *   - Ruler units
 * - *eCompassFlipped*
 *   - Compass flipped
 * - *eCompassSpread*
 *   - Compass spread
 * - *eProtractorStyle*
 *   - Protractor style
 * - *eShapeDivision*
 *   - Shape division 
 * - *eReflectShape*
 *   - Reflect shape
 * - *eCompressImage*
 *   - Compress image
 * - *eClearInk*
 *   - Clear ink
 */
NB.undomanager.undotype = ["eNoUndoRedoName",
                           "eColor",
                           "eFillColor",
                           "eLineThickness",
                           "eHyperlink",
                           "eHyperlinkLaunchType",
                           "eLockInPlace",
                           "eTextChange",
                           "eTextItalic",
                           "eTextBold",
                           "eTextUnderline",
                           "eTextFontChange",
                           "eLineEnds",
                           "eMirrored",
                           "eClearPage",
                           "eText",
                           "eEquation",
                           "ePaste",
                           "eErasing",
                           "eWriting",
                           "eShape",
                           "eLine",
                           "eClone",
                           "eBackgroundTransparency",
                           "eOrder",
                           "eDelete",
                           "eUngroup",
                           "eGroup",
                           "eRecognize",
                           "eResize",
                           "eRotate",
                           "eMoveRotationPoint",
                           "eMove",
                           "eDragCopy",
                           "eDrop",
                           "eRemoveFromContainer",
                           "eTransparency",
                           "eRenamePage",
                           "eDragAndDrop",
                           "eDashStyle",
                           "eBackgroundColor",
                           "eTableCellColor",
                           "ePropertyChanges",
                           "eGraphicInsertion",
                           "eInsertHyperlink",
                           "eDeleteAttachment",
                           "eAddAttachment",
                           "eCutAttachment",
                           "eAddPage",
                           "eClonePage",
                           "eDeletePage",
                           "eDragCopyPage",
                           "eScreenCapture",
                           "eScreenShade",
                           "eAddTemplate",
                           "eSound",
                           "eVideo",
                           "eSoundLaunchType",
                           "eScreenShadeMove",
                           "eScreenShadeOpen",
                           "eScreenShadeClose",
                           "eInfiniteClone",
                           "eCut",
                           "eInsertTable",
                           "eMergeCells",
                           "eSplitCell",
                           "eDeleteColumn",
                           "eInsertColumn",
                           "eDeleteRow",
                           "eInsertRow",
                           "eMoveSeperator",
                           "eAdjustCellsSize",
                           "eInsertQuestion",
                           "eInsertTitlePage",
                           "eInsertAnswerKey",
                           "eEditAnswer",
                           "eInsertChoice",
                           "eStartTest",
                           "eFinishVoting",
                           "eClearResults",
                           "ePanelVariableChange",
                           "eAlignAnnotations",
                           "eAddPageGroup",
                           "eThemeApplied",
                           "eDeletePageGroup",
                           "eReorderPageGroup",
                           "eRenameGroup",
                           "eTableShade",
                           "eRemoveCells",
                           "eScreenShadeMaximize",
                           "ePageRecordingChanged",
                           "eObjectAnimationChanged",
                           "eThemeDelete",
                           "eRecognizeMathInk",
                           "eCancelRecognizeMathInk",
                           "eMathAction",
                           "eGenerateGraph",
                           "eGenerateTable",
                           "eConnect",
                           "eDisconnect",
                           "eMoveVertex",
                           "eInsertProtractor",
                           "eInsertRuler",
                           "eInsertCompass",
                           "eRulerWidth",
                           "eRulerUnits",
                           "eCompassFlipped",
                           "eCompassSpread",
                           "eProtractorStyle",
                           "eShapeDivision",
                           "eAddGraphPoint",
                           "eInsertGraph",
                           "eInsertNumberLine",
                           "eReflectShape",
                           "eCompressImage",
                           "eClearInk",
                           "eActiveContent",
                           "eRotate3D",
						   "eEditUrl",
						   "eInsertBrowser"];
 
})();

// _includes.js

/**
 * @constructor
 * @ignore
 */
function nb_load_includes() {
	window.isDefined = function(v) { return (!(typeof(v) == "undefined")); };
	String.prototype.trim = function() { return this.replace(/^\s*|\s*$/g, ""); };

/*!
 * jQuery JavaScript Library v1.6.1
 * http://jquery.com/
 *
 * Copyright 2011, John Resig
 * Dual licensed under the MIT or GPL Version 2 licenses.
 * http://jquery.org/license
 *
 * Includes Sizzle.js
 * http://sizzlejs.com/
 * Copyright 2011, The Dojo Foundation
 * Released under the MIT, BSD, and GPL Licenses.
 *
 * Date: Thu May 12 15:04:36 2011 -0400
 */
(function(a,b){function cy(a){return f.isWindow(a)?a:a.nodeType===9?a.defaultView||a.parentWindow:!1}function cv(a){if(!cj[a]){var b=f("<"+a+">").appendTo("body"),d=b.css("display");b.remove();if(d==="none"||d===""){ck||(ck=c.createElement("iframe"),ck.frameBorder=ck.width=ck.height=0),c.body.appendChild(ck);if(!cl||!ck.createElement)cl=(ck.contentWindow||ck.contentDocument).document,cl.write("<!doctype><html><body></body></html>");b=cl.createElement(a),cl.body.appendChild(b),d=f.css(b,"display"),c.body.removeChild(ck)}cj[a]=d}return cj[a]}function cu(a,b){var c={};f.each(cp.concat.apply([],cp.slice(0,b)),function(){c[this]=a});return c}function ct(){cq=b}function cs(){setTimeout(ct,0);return cq=f.now()}function ci(){try{return new a.ActiveXObject("Microsoft.XMLHTTP")}catch(b){}}function ch(){try{return new a.XMLHttpRequest}catch(b){}}function cb(a,c){a.dataFilter&&(c=a.dataFilter(c,a.dataType));var d=a.dataTypes,e={},g,h,i=d.length,j,k=d[0],l,m,n,o,p;for(g=1;g<i;g++){if(g===1)for(h in a.converters)typeof h=="string"&&(e[h.toLowerCase()]=a.converters[h]);l=k,k=d[g];if(k==="*")k=l;else if(l!=="*"&&l!==k){m=l+" "+k,n=e[m]||e["* "+k];if(!n){p=b;for(o in e){j=o.split(" ");if(j[0]===l||j[0]==="*"){p=e[j[1]+" "+k];if(p){o=e[o],o===!0?n=p:p===!0&&(n=o);break}}}}!n&&!p&&f.error("No conversion from "+m.replace(" "," to ")),n!==!0&&(c=n?n(c):p(o(c)))}}return c}function ca(a,c,d){var e=a.contents,f=a.dataTypes,g=a.responseFields,h,i,j,k;for(i in g)i in d&&(c[g[i]]=d[i]);while(f[0]==="*")f.shift(),h===b&&(h=a.mimeType||c.getResponseHeader("content-type"));if(h)for(i in e)if(e[i]&&e[i].test(h)){f.unshift(i);break}if(f[0]in d)j=f[0];else{for(i in d){if(!f[0]||a.converters[i+" "+f[0]]){j=i;break}k||(k=i)}j=j||k}if(j){j!==f[0]&&f.unshift(j);return d[j]}}function b_(a,b,c,d){if(f.isArray(b))f.each(b,function(b,e){c||bF.test(a)?d(a,e):b_(a+"["+(typeof e=="object"||f.isArray(e)?b:"")+"]",e,c,d)});else if(!c&&b!=null&&typeof b=="object")for(var e in b)b_(a+"["+e+"]",b[e],c,d);else d(a,b)}function b$(a,c,d,e,f,g){f=f||c.dataTypes[0],g=g||{},g[f]=!0;var h=a[f],i=0,j=h?h.length:0,k=a===bU,l;for(;i<j&&(k||!l);i++)l=h[i](c,d,e),typeof l=="string"&&(!k||g[l]?l=b:(c.dataTypes.unshift(l),l=b$(a,c,d,e,l,g)));(k||!l)&&!g["*"]&&(l=b$(a,c,d,e,"*",g));return l}function bZ(a){return function(b,c){typeof b!="string"&&(c=b,b="*");if(f.isFunction(c)){var d=b.toLowerCase().split(bQ),e=0,g=d.length,h,i,j;for(;e<g;e++)h=d[e],j=/^\+/.test(h),j&&(h=h.substr(1)||"*"),i=a[h]=a[h]||[],i[j?"unshift":"push"](c)}}}function bD(a,b,c){var d=b==="width"?bx:by,e=b==="width"?a.offsetWidth:a.offsetHeight;if(c==="border")return e;f.each(d,function(){c||(e-=parseFloat(f.css(a,"padding"+this))||0),c==="margin"?e+=parseFloat(f.css(a,"margin"+this))||0:e-=parseFloat(f.css(a,"border"+this+"Width"))||0});return e}function bn(a,b){b.src?f.ajax({url:b.src,async:!1,dataType:"script"}):f.globalEval((b.text||b.textContent||b.innerHTML||"").replace(bf,"/*$0*/")),b.parentNode&&b.parentNode.removeChild(b)}function bm(a){f.nodeName(a,"input")?bl(a):a.getElementsByTagName&&f.grep(a.getElementsByTagName("input"),bl)}function bl(a){if(a.type==="checkbox"||a.type==="radio")a.defaultChecked=a.checked}function bk(a){return"getElementsByTagName"in a?a.getElementsByTagName("*"):"querySelectorAll"in a?a.querySelectorAll("*"):[]}function bj(a,b){var c;if(b.nodeType===1){b.clearAttributes&&b.clearAttributes(),b.mergeAttributes&&b.mergeAttributes(a),c=b.nodeName.toLowerCase();if(c==="object")b.outerHTML=a.outerHTML;else if(c!=="input"||a.type!=="checkbox"&&a.type!=="radio"){if(c==="option")b.selected=a.defaultSelected;else if(c==="input"||c==="textarea")b.defaultValue=a.defaultValue}else a.checked&&(b.defaultChecked=b.checked=a.checked),b.value!==a.value&&(b.value=a.value);b.removeAttribute(f.expando)}}function bi(a,b){if(b.nodeType===1&&!!f.hasData(a)){var c=f.expando,d=f.data(a),e=f.data(b,d);if(d=d[c]){var g=d.events;e=e[c]=f.extend({},d);if(g){delete e.handle,e.events={};for(var h in g)for(var i=0,j=g[h].length;i<j;i++)f.event.add(b,h+(g[h][i].namespace?".":"")+g[h][i].namespace,g[h][i],g[h][i].data)}}}}function bh(a,b){return f.nodeName(a,"table")?a.getElementsByTagName("tbody")[0]||a.appendChild(a.ownerDocument.createElement("tbody")):a}function X(a,b,c){b=b||0;if(f.isFunction(b))return f.grep(a,function(a,d){var e=!!b.call(a,d,a);return e===c});if(b.nodeType)return f.grep(a,function(a,d){return a===b===c});if(typeof b=="string"){var d=f.grep(a,function(a){return a.nodeType===1});if(S.test(b))return f.filter(b,d,!c);b=f.filter(b,d)}return f.grep(a,function(a,d){return f.inArray(a,b)>=0===c})}function W(a){return!a||!a.parentNode||a.parentNode.nodeType===11}function O(a,b){return(a&&a!=="*"?a+".":"")+b.replace(A,"`").replace(B,"&")}function N(a){var b,c,d,e,g,h,i,j,k,l,m,n,o,p=[],q=[],r=f._data(this,"events");if(!(a.liveFired===this||!r||!r.live||a.target.disabled||a.button&&a.type==="click")){a.namespace&&(n=new RegExp("(^|\\.)"+a.namespace.split(".").join("\\.(?:.*\\.)?")+"(\\.|$)")),a.liveFired=this;var s=r.live.slice(0);for(i=0;i<s.length;i++)g=s[i],g.origType.replace(y,"")===a.type?q.push(g.selector):s.splice(i--,1);e=f(a.target).closest(q,a.currentTarget);for(j=0,k=e.length;j<k;j++){m=e[j];for(i=0;i<s.length;i++){g=s[i];if(m.selector===g.selector&&(!n||n.test(g.namespace))&&!m.elem.disabled){h=m.elem,d=null;if(g.preType==="mouseenter"||g.preType==="mouseleave")a.type=g.preType,d=f(a.relatedTarget).closest(g.selector)[0],d&&f.contains(h,d)&&(d=h);(!d||d!==h)&&p.push({elem:h,handleObj:g,level:m.level})}}}for(j=0,k=p.length;j<k;j++){e=p[j];if(c&&e.level>c)break;a.currentTarget=e.elem,a.data=e.handleObj.data,a.handleObj=e.handleObj,o=e.handleObj.origHandler.apply(e.elem,arguments);if(o===!1||a.isPropagationStopped()){c=e.level,o===!1&&(b=!1);if(a.isImmediatePropagationStopped())break}}return b}}function L(a,c,d){var e=f.extend({},d[0]);e.type=a,e.originalEvent={},e.liveFired=b,f.event.handle.call(c,e),e.isDefaultPrevented()&&d[0].preventDefault()}function F(){return!0}function E(){return!1}function m(a,c,d){var e=c+"defer",g=c+"queue",h=c+"mark",i=f.data(a,e,b,!0);i&&(d==="queue"||!f.data(a,g,b,!0))&&(d==="mark"||!f.data(a,h,b,!0))&&setTimeout(function(){!f.data(a,g,b,!0)&&!f.data(a,h,b,!0)&&(f.removeData(a,e,!0),i.resolve())},0)}function l(a){for(var b in a)if(b!=="toJSON")return!1;return!0}function k(a,c,d){if(d===b&&a.nodeType===1){var e="data-"+c.replace(j,"$1-$2").toLowerCase();d=a.getAttribute(e);if(typeof d=="string"){try{d=d==="true"?!0:d==="false"?!1:d==="null"?null:f.isNaN(d)?i.test(d)?f.parseJSON(d):d:parseFloat(d)}catch(g){}f.data(a,c,d)}else d=b}return d}var c=a.document,d=a.navigator,e=a.location,f=function(){function H(){if(!e.isReady){try{c.documentElement.doScroll("left")}catch(a){setTimeout(H,1);return}e.ready()}}var e=function(a,b){return new e.fn.init(a,b,h)},f=a.jQuery,g=a.$,h,i=/^(?:[^<]*(<[\w\W]+>)[^>]*$|#([\w\-]*)$)/,j=/\S/,k=/^\s+/,l=/\s+$/,m=/\d/,n=/^<(\w+)\s*\/?>(?:<\/\1>)?$/,o=/^[\],:{}\s]*$/,p=/\\(?:["\\\/bfnrt]|u[0-9a-fA-F]{4})/g,q=/"[^"\\\n\r]*"|true|false|null|-?\d+(?:\.\d*)?(?:[eE][+\-]?\d+)?/g,r=/(?:^|:|,)(?:\s*\[)+/g,s=/(webkit)[ \/]([\w.]+)/,t=/(opera)(?:.*version)?[ \/]([\w.]+)/,u=/(msie) ([\w.]+)/,v=/(mozilla)(?:.*? rv:([\w.]+))?/,w=d.userAgent,x,y,z,A=Object.prototype.toString,B=Object.prototype.hasOwnProperty,C=Array.prototype.push,D=Array.prototype.slice,E=String.prototype.trim,F=Array.prototype.indexOf,G={};e.fn=e.prototype={constructor:e,init:function(a,d,f){var g,h,j,k;if(!a)return this;if(a.nodeType){this.context=this[0]=a,this.length=1;return this}if(a==="body"&&!d&&c.body){this.context=c,this[0]=c.body,this.selector=a,this.length=1;return this}if(typeof a=="string"){a.charAt(0)!=="<"||a.charAt(a.length-1)!==">"||a.length<3?g=i.exec(a):g=[null,a,null];if(g&&(g[1]||!d)){if(g[1]){d=d instanceof e?d[0]:d,k=d?d.ownerDocument||d:c,j=n.exec(a),j?e.isPlainObject(d)?(a=[c.createElement(j[1])],e.fn.attr.call(a,d,!0)):a=[k.createElement(j[1])]:(j=e.buildFragment([g[1]],[k]),a=(j.cacheable?e.clone(j.fragment):j.fragment).childNodes);return e.merge(this,a)}h=c.getElementById(g[2]);if(h&&h.parentNode){if(h.id!==g[2])return f.find(a);this.length=1,this[0]=h}this.context=c,this.selector=a;return this}return!d||d.jquery?(d||f).find(a):this.constructor(d).find(a)}if(e.isFunction(a))return f.ready(a);a.selector!==b&&(this.selector=a.selector,this.context=a.context);return e.makeArray(a,this)},selector:"",jquery:"1.6.1",length:0,size:function(){return this.length},toArray:function(){return D.call(this,0)},get:function(a){return a==null?this.toArray():a<0?this[this.length+a]:this[a]},pushStack:function(a,b,c){var d=this.constructor();e.isArray(a)?C.apply(d,a):e.merge(d,a),d.prevObject=this,d.context=this.context,b==="find"?d.selector=this.selector+(this.selector?" ":"")+c:b&&(d.selector=this.selector+"."+b+"("+c+")");return d},each:function(a,b){return e.each(this,a,b)},ready:function(a){e.bindReady(),y.done(a);return this},eq:function(a){return a===-1?this.slice(a):this.slice(a,+a+1)},first:function(){return this.eq(0)},last:function(){return this.eq(-1)},slice:function(){return this.pushStack(D.apply(this,arguments),"slice",D.call(arguments).join(","))},map:function(a){return this.pushStack(e.map(this,function(b,c){return a.call(b,c,b)}))},end:function(){return this.prevObject||this.constructor(null)},push:C,sort:[].sort,splice:[].splice},e.fn.init.prototype=e.fn,e.extend=e.fn.extend=function(){var a,c,d,f,g,h,i=arguments[0]||{},j=1,k=arguments.length,l=!1;typeof i=="boolean"&&(l=i,i=arguments[1]||{},j=2),typeof i!="object"&&!e.isFunction(i)&&(i={}),k===j&&(i=this,--j);for(;j<k;j++)if((a=arguments[j])!=null)for(c in a){d=i[c],f=a[c];if(i===f)continue;l&&f&&(e.isPlainObject(f)||(g=e.isArray(f)))?(g?(g=!1,h=d&&e.isArray(d)?d:[]):h=d&&e.isPlainObject(d)?d:{},i[c]=e.extend(l,h,f)):f!==b&&(i[c]=f)}return i},e.extend({noConflict:function(b){a.$===e&&(a.$=g),b&&a.jQuery===e&&(a.jQuery=f);return e},isReady:!1,readyWait:1,holdReady:function(a){a?e.readyWait++:e.ready(!0)},ready:function(a){if(a===!0&&!--e.readyWait||a!==!0&&!e.isReady){if(!c.body)return setTimeout(e.ready,1);e.isReady=!0;if(a!==!0&&--e.readyWait>0)return;y.resolveWith(c,[e]),e.fn.trigger&&e(c).trigger("ready").unbind("ready")}},bindReady:function(){if(!y){y=e._Deferred();if(c.readyState==="complete")return setTimeout(e.ready,1);if(c.addEventListener)c.addEventListener("DOMContentLoaded",z,!1),a.addEventListener("load",e.ready,!1);else if(c.attachEvent){c.attachEvent("onreadystatechange",z),a.attachEvent("onload",e.ready);var b=!1;try{b=a.frameElement==null}catch(d){}c.documentElement.doScroll&&b&&H()}}},isFunction:function(a){return e.type(a)==="function"},isArray:Array.isArray||function(a){return e.type(a)==="array"},isWindow:function(a){return a&&typeof a=="object"&&"setInterval"in a},isNaN:function(a){return a==null||!m.test(a)||isNaN(a)},type:function(a){return a==null?String(a):G[A.call(a)]||"object"},isPlainObject:function(a){if(!a||e.type(a)!=="object"||a.nodeType||e.isWindow(a))return!1;if(a.constructor&&!B.call(a,"constructor")&&!B.call(a.constructor.prototype,"isPrototypeOf"))return!1;var c;for(c in a);return c===b||B.call(a,c)},isEmptyObject:function(a){for(var b in a)return!1;return!0},error:function(a){throw a},parseJSON:function(b){if(typeof b!="string"||!b)return null;b=e.trim(b);if(a.JSON&&a.JSON.parse)return a.JSON.parse(b);if(o.test(b.replace(p,"@").replace(q,"]").replace(r,"")))return(new Function("return "+b))();e.error("Invalid JSON: "+b)},parseXML:function(b,c,d){a.DOMParser?(d=new DOMParser,c=d.parseFromString(b,"text/xml")):(c=new ActiveXObject("Microsoft.XMLDOM"),c.async="false",c.loadXML(b)),d=c.documentElement,(!d||!d.nodeName||d.nodeName==="parsererror")&&e.error("Invalid XML: "+b);return c},noop:function(){},globalEval:function(b){b&&j.test(b)&&(a.execScript||function(b){a.eval.call(a,b)})(b)},nodeName:function(a,b){return a.nodeName&&a.nodeName.toUpperCase()===b.toUpperCase()},each:function(a,c,d){var f,g=0,h=a.length,i=h===b||e.isFunction(a);if(d){if(i){for(f in a)if(c.apply(a[f],d)===!1)break}else for(;g<h;)if(c.apply(a[g++],d)===!1)break}else if(i){for(f in a)if(c.call(a[f],f,a[f])===!1)break}else for(;g<h;)if(c.call(a[g],g,a[g++])===!1)break;return a},trim:E?function(a){return a==null?"":E.call(a)}:function(a){return a==null?"":(a+"").replace(k,"").replace(l,"")},makeArray:function(a,b){var c=b||[];if(a!=null){var d=e.type(a);a.length==null||d==="string"||d==="function"||d==="regexp"||e.isWindow(a)?C.call(c,a):e.merge(c,a)}return c},inArray:function(a,b){if(F)return F.call(b,a);for(var c=0,d=b.length;c<d;c++)if(b[c]===a)return c;return-1},merge:function(a,c){var d=a.length,e=0;if(typeof c.length=="number")for(var f=c.length;e<f;e++)a[d++]=c[e];else while(c[e]!==b)a[d++]=c[e++];a.length=d;return a},grep:function(a,b,c){var d=[],e;c=!!c;for(var f=0,g=a.length;f<g;f++)e=!!b(a[f],f),c!==e&&d.push(a[f]);return d},map:function(a,c,d){var f,g,h=[],i=0,j=a.length,k=a instanceof e||j!==b&&typeof j=="number"&&(j>0&&a[0]&&a[j-1]||j===0||e.isArray(a));if(k)for(;i<j;i++)f=c(a[i],i,d),f!=null&&(h[h.length]=f);else for(g in a)f=c(a[g],g,d),f!=null&&(h[h.length]=f);return h.concat.apply([],h)},guid:1,proxy:function(a,c){if(typeof c=="string"){var d=a[c];c=a,a=d}if(!e.isFunction(a))return b;var f=D.call(arguments,2),g=function(){return a.apply(c,f.concat(D.call(arguments)))};g.guid=a.guid=a.guid||g.guid||e.guid++;return g},access:function(a,c,d,f,g,h){var i=a.length;if(typeof c=="object"){for(var j in c)e.access(a,j,c[j],f,g,d);return a}if(d!==b){f=!h&&f&&e.isFunction(d);for(var k=0;k<i;k++)g(a[k],c,f?d.call(a[k],k,g(a[k],c)):d,h);return a}return i?g(a[0],c):b},now:function(){return(new Date).getTime()},uaMatch:function(a){a=a.toLowerCase();var b=s.exec(a)||t.exec(a)||u.exec(a)||a.indexOf("compatible")<0&&v.exec(a)||[];return{browser:b[1]||"",version:b[2]||"0"}},sub:function(){function a(b,c){return new a.fn.init(b,c)}e.extend(!0,a,this),a.superclass=this,a.fn=a.prototype=this(),a.fn.constructor=a,a.sub=this.sub,a.fn.init=function(d,f){f&&f instanceof e&&!(f instanceof a)&&(f=a(f));return e.fn.init.call(this,d,f,b)},a.fn.init.prototype=a.fn;var b=a(c);return a},browser:{}}),e.each("Boolean Number String Function Array Date RegExp Object".split(" "),function(a,b){G["[object "+b+"]"]=b.toLowerCase()}),x=e.uaMatch(w),x.browser&&(e.browser[x.browser]=!0,e.browser.version=x.version),e.browser.webkit&&(e.browser.safari=!0),j.test(" ")&&(k=/^[\s\xA0]+/,l=/[\s\xA0]+$/),h=e(c),c.addEventListener?z=function(){c.removeEventListener("DOMContentLoaded",z,!1),e.ready()}:c.attachEvent&&(z=function(){c.readyState==="complete"&&(c.detachEvent("onreadystatechange",z),e.ready())});return e}(),g="done fail isResolved isRejected promise then always pipe".split(" "),h=[].slice;f.extend({_Deferred:function(){var a=[],b,c,d,e={done:function(){if(!d){var c=arguments,g,h,i,j,k;b&&(k=b,b=0);for(g=0,h=c.length;g<h;g++)i=c[g],j=f.type(i),j==="array"?e.done.apply(e,i):j==="function"&&a.push(i);k&&e.resolveWith(k[0],k[1])}return this},resolveWith:function(e,f){if(!d&&!b&&!c){f=f||[],c=1;try{while(a[0])a.shift().apply(e,f)}finally{b=[e,f],c=0}}return this},resolve:function(){e.resolveWith(this,arguments);return this},isResolved:function(){return!!c||!!b},cancel:function(){d=1,a=[];return this}};return e},Deferred:function(a){var b=f._Deferred(),c=f._Deferred(),d;f.extend(b,{then:function(a,c){b.done(a).fail(c);return this},always:function(){return b.done.apply(b,arguments).fail.apply(this,arguments)},fail:c.done,rejectWith:c.resolveWith,reject:c.resolve,isRejected:c.isResolved,pipe:function(a,c){return f.Deferred(function(d){f.each({done:[a,"resolve"],fail:[c,"reject"]},function(a,c){var e=c[0],g=c[1],h;f.isFunction(e)?b[a](function(){h=e.apply(this,arguments),h&&f.isFunction(h.promise)?h.promise().then(d.resolve,d.reject):d[g](h)}):b[a](d[g])})}).promise()},promise:function(a){if(a==null){if(d)return d;d=a={}}var c=g.length;while(c--)a[g[c]]=b[g[c]];return a}}),b.done(c.cancel).fail(b.cancel),delete b.cancel,a&&a.call(b,b);return b},when:function(a){function i(a){return function(c){b[a]=arguments.length>1?h.call(arguments,0):c,--e||g.resolveWith(g,h.call(b,0))}}var b=arguments,c=0,d=b.length,e=d,g=d<=1&&a&&f.isFunction(a.promise)?a:f.Deferred();if(d>1){for(;c<d;c++)b[c]&&f.isFunction(b[c].promise)?b[c].promise().then(i(c),g.reject):--e;e||g.resolveWith(g,b)}else g!==a&&g.resolveWith(g,d?[a]:[]);return g.promise()}}),f.support=function(){var a=c.createElement("div"),b=c.documentElement,d,e,f,g,h,i,j,k,l,m,n,o,p,q,r;a.setAttribute("className","t"),a.innerHTML="   <link/><table></table><a href='/a' style='top:1px;float:left;opacity:.55;'>a</a><input type='checkbox'/>",d=a.getElementsByTagName("*"),e=a.getElementsByTagName("a")[0];if(!d||!d.length||!e)return{};f=c.createElement("select"),g=f.appendChild(c.createElement("option")),h=a.getElementsByTagName("input")[0],j={leadingWhitespace:a.firstChild.nodeType===3,tbody:!a.getElementsByTagName("tbody").length,htmlSerialize:!!a.getElementsByTagName("link").length,style:/top/.test(e.getAttribute("style")),hrefNormalized:e.getAttribute("href")==="/a",opacity:/^0.55$/.test(e.style.opacity),cssFloat:!!e.style.cssFloat,checkOn:h.value==="on",optSelected:g.selected,getSetAttribute:a.className!=="t",submitBubbles:!0,changeBubbles:!0,focusinBubbles:!1,deleteExpando:!0,noCloneEvent:!0,inlineBlockNeedsLayout:!1,shrinkWrapBlocks:!1,reliableMarginRight:!0},h.checked=!0,j.noCloneChecked=h.cloneNode(!0).checked,f.disabled=!0,j.optDisabled=!g.disabled;try{delete a.test}catch(s){j.deleteExpando=!1}!a.addEventListener&&a.attachEvent&&a.fireEvent&&(a.attachEvent("onclick",function b(){j.noCloneEvent=!1,a.detachEvent("onclick",b)}),a.cloneNode(!0).fireEvent("onclick")),h=c.createElement("input"),h.value="t",h.setAttribute("type","radio"),j.radioValue=h.value==="t",h.setAttribute("checked","checked"),a.appendChild(h),k=c.createDocumentFragment(),k.appendChild(a.firstChild),j.checkClone=k.cloneNode(!0).cloneNode(!0).lastChild.checked,a.innerHTML="",a.style.width=a.style.paddingLeft="1px",l=c.createElement("body"),m={visibility:"hidden",width:0,height:0,border:0,margin:0,background:"none"};for(q in m)l.style[q]=m[q];l.appendChild(a),b.insertBefore(l,b.firstChild),j.appendChecked=h.checked,j.boxModel=a.offsetWidth===2,"zoom"in a.style&&(a.style.display="inline",a.style.zoom=1,j.inlineBlockNeedsLayout=a.offsetWidth===2,a.style.display="",a.innerHTML="<div style='width:4px;'></div>",j.shrinkWrapBlocks=a.offsetWidth!==2),a.innerHTML="<table><tr><td style='padding:0;border:0;display:none'></td><td>t</td></tr></table>",n=a.getElementsByTagName("td"),r=n[0].offsetHeight===0,n[0].style.display="",n[1].style.display="none",j.reliableHiddenOffsets=r&&n[0].offsetHeight===0,a.innerHTML="",c.defaultView&&c.defaultView.getComputedStyle&&(i=c.createElement("div"),i.style.width="0",i.style.marginRight="0",a.appendChild(i),j.reliableMarginRight=(parseInt((c.defaultView.getComputedStyle(i,null)||{marginRight:0}).marginRight,10)||0)===0),l.innerHTML="",b.removeChild(l);if(a.attachEvent)for(q in{submit:1,change:1,focusin:1})p="on"+q,r=p in a,r||(a.setAttribute(p,"return;"),r=typeof a[p]=="function"),j[q+"Bubbles"]=r;return j}(),f.boxModel=f.support.boxModel;var i=/^(?:\{.*\}|\[.*\])$/,j=/([a-z])([A-Z])/g;f.extend({cache:{},uuid:0,expando:"jQuery"+(f.fn.jquery+Math.random()).replace(/\D/g,""),noData:{embed:!0,object:"clsid:D27CDB6E-AE6D-11cf-96B8-444553540000",applet:!0},hasData:function(a){a=a.nodeType?f.cache[a[f.expando]]:a[f.expando];return!!a&&!l(a)},data:function(a,c,d,e){if(!!f.acceptData(a)){var g=f.expando,h=typeof c=="string",i,j=a.nodeType,k=j?f.cache:a,l=j?a[f.expando]:a[f.expando]&&f.expando;if((!l||e&&l&&!k[l][g])&&h&&d===b)return;l||(j?a[f.expando]=l=++f.uuid:l=f.expando),k[l]||(k[l]={},j||(k[l].toJSON=f.noop));if(typeof c=="object"||typeof c=="function")e?k[l][g]=f.extend(k[l][g],c):k[l]=f.extend(k[l],c);i=k[l],e&&(i[g]||(i[g]={}),i=i[g]),d!==b&&(i[f.camelCase(c)]=d);if(c==="events"&&!i[c])return i[g]&&i[g].events;return h?i[f.camelCase(c)]:i}},removeData:function(b,c,d){if(!!f.acceptData(b)){var e=f.expando,g=b.nodeType,h=g?f.cache:b,i=g?b[f.expando]:f.expando;if(!h[i])return;if(c){var j=d?h[i][e]:h[i];if(j){delete j[c];if(!l(j))return}}if(d){delete h[i][e];if(!l(h[i]))return}var k=h[i][e];f.support.deleteExpando||h!=a?delete h[i]:h[i]=null,k?(h[i]={},g||(h[i].toJSON=f.noop),h[i][e]=k):g&&(f.support.deleteExpando?delete b[f.expando]:b.removeAttribute?b.removeAttribute(f.expando):b[f.expando]=null)}},_data:function(a,b,c){return f.data(a,b,c,!0)},acceptData:function(a){if(a.nodeName){var b=f.noData[a.nodeName.toLowerCase()];if(b)return b!==!0&&a.getAttribute("classid")===b}return!0}}),f.fn.extend({data:function(a,c){var d=null;if(typeof a=="undefined"){if(this.length){d=f.data(this[0]);if(this[0].nodeType===1){var e=this[0].attributes,g;for(var h=0,i=e.length;h<i;h++)g=e[h].name,g.indexOf("data-")===0&&(g=f.camelCase(g.substring(5)),k(this[0],g,d[g]))}}return d}if(typeof a=="object")return this.each(function(){f.data(this,a)});var j=a.split(".");j[1]=j[1]?"."+j[1]:"";if(c===b){d=this.triggerHandler("getData"+j[1]+"!",[j[0]]),d===b&&this.length&&(d=f.data(this[0],a),d=k(this[0],a,d));return d===b&&j[1]?this.data(j[0]):d}return this.each(function(){var b=f(this),d=[j[0],c];b.triggerHandler("setData"+j[1]+"!",d),f.data(this,a,c),b.triggerHandler("changeData"+j[1]+"!",d)})},removeData:function(a){return this.each(function(){f.removeData(this,a)})}}),f.extend({_mark:function(a,c){a&&(c=(c||"fx")+"mark",f.data(a,c,(f.data(a,c,b,!0)||0)+1,!0))},_unmark:function(a,c,d){a!==!0&&(d=c,c=a,a=!1);if(c){d=d||"fx";var e=d+"mark",g=a?0:(f.data(c,e,b,!0)||1)-1;g?f.data(c,e,g,!0):(f.removeData(c,e,!0),m(c,d,"mark"))}},queue:function(a,c,d){if(a){c=(c||"fx")+"queue";var e=f.data(a,c,b,!0);d&&(!e||f.isArray(d)?e=f.data(a,c,f.makeArray(d),!0):e.push(d));return e||[]}},dequeue:function(a,b){b=b||"fx";var c=f.queue(a,b),d=c.shift(),e;d==="inprogress"&&(d=c.shift()),d&&(b==="fx"&&c.unshift("inprogress"),d.call(a,function(){f.dequeue(a,b)})),c.length||(f.removeData(a,b+"queue",!0),m(a,b,"queue"))}}),f.fn.extend({queue:function(a,c){typeof a!="string"&&(c=a,a="fx");if(c===b)return f.queue(this[0],a);return this.each(function(){var b=f.queue(this,a,c);a==="fx"&&b[0]!=="inprogress"&&f.dequeue(this,a)})},dequeue:function(a){return this.each(function(){f.dequeue(this,a)})},delay:function(a,b){a=f.fx?f.fx.speeds[a]||a:a,b=b||"fx";return this.queue(b,function(){var c=this;setTimeout(function(){f.dequeue(c,b)},a)})},clearQueue:function(a){return this.queue(a||"fx",[])},promise:function(a,c){function m(){--h||d.resolveWith(e,[e])}typeof a!="string"&&(c=a,a=b),a=a||"fx";var d=f.Deferred(),e=this,g=e.length,h=1,i=a+"defer",j=a+"queue",k=a+"mark",l;while(g--)if(l=f.data(e[g],i,b,!0)||(f.data(e[g],j,b,!0)||f.data(e[g],k,b,!0))&&f.data(e[g],i,f._Deferred(),!0))h++,l.done(m);m();return d.promise()}});var n=/[\n\t\r]/g,o=/\s+/,p=/\r/g,q=/^(?:button|input)$/i,r=/^(?:button|input|object|select|textarea)$/i,s=/^a(?:rea)?$/i,t=/^(?:autofocus|autoplay|async|checked|controls|defer|disabled|hidden|loop|multiple|open|readonly|required|scoped|selected)$/i,u=/\:/,v,w;f.fn.extend({attr:function(a,b){return f.access(this,a,b,!0,f.attr)},removeAttr:function(a){return this.each(function(){f.removeAttr(this,a)})},prop:function(a,b){return f.access(this,a,b,!0,f.prop)},removeProp:function(a){a=f.propFix[a]||a;return this.each(function(){try{this[a]=b,delete this[a]}catch(c){}})},addClass:function(a){if(f.isFunction(a))return this.each(function(b){var c=f(this);c.addClass(a.call(this,b,c.attr("class")||""))});if(a&&typeof a=="string"){var b=(a||"").split(o);for(var c=0,d=this.length;c<d;c++){var e=this[c];if(e.nodeType===1)if(!e.className)e.className=a;else{var g=" "+e.className+" ",h=e.className;for(var i=0,j=b.length;i<j;i++)g.indexOf(" "+b[i]+" ")<0&&(h+=" "+b[i]);e.className=f.trim(h)}}}return this},removeClass:function(a){if(f.isFunction(a))return this.each(function(b){var c=f(this);c.removeClass(a.call(this,b,c.attr("class")))});if(a&&typeof a=="string"||a===b){var c=(a||"").split(o);for(var d=0,e=this.length;d<e;d++){var g=this[d];if(g.nodeType===1&&g.className)if(a){var h=(" "+g.className+" ").replace(n," ");for(var i=0,j=c.length;i<j;i++)h=h.replace(" "+c[i]+" "," ");g.className=f.trim(h)}else g.className=""}}return this},toggleClass:function(a,b){var c=typeof a,d=typeof b=="boolean";if(f.isFunction(a))return this.each(function(c){var d=f(this);d.toggleClass(a.call(this,c,d.attr("class"),b),b)});return this.each(function(){if(c==="string"){var e,g=0,h=f(this),i=b,j=a.split(o);while(e=j[g++])i=d?i:!h.hasClass(e),h[i?"addClass":"removeClass"](e)}else if(c==="undefined"||c==="boolean")this.className&&f._data(this,"__className__",this.className),this.className=this.className||a===!1?"":f._data(this,"__className__")||""})},hasClass:function(a){var b=" "+a+" ";for(var c=0,d=this.length;c<d;c++)if((" "+this[c].className+" ").replace(n," ").indexOf(b)>-1)return!0;return!1},val:function(a){var c,d,e=this[0];if(!arguments.length){if(e){c=f.valHooks[e.nodeName.toLowerCase()]||f.valHooks[e.type];if(c&&"get"in c&&(d=c.get(e,"value"))!==b)return d;return(e.value||"").replace(p,"")}return b}var g=f.isFunction(a);return this.each(function(d){var e=f(this),h;if(this.nodeType===1){g?h=a.call(this,d,e.val()):h=a,h==null?h="":typeof h=="number"?h+="":f.isArray(h)&&(h=f.map(h,function(a){return a==null?"":a+""})),c=f.valHooks[this.nodeName.toLowerCase()]||f.valHooks[this.type];if(!c||!("set"in c)||c.set(this,h,"value")===b)this.value=h}})}}),f.extend({valHooks:{option:{get:function(a){var b=a.attributes.value;return!b||b.specified?a.value:a.text}},select:{get:function(a){var b,c=a.selectedIndex,d=[],e=a.options,g=a.type==="select-one";if(c<0)return null;for(var h=g?c:0,i=g?c+1:e.length;h<i;h++){var j=e[h];if(j.selected&&(f.support.optDisabled?!j.disabled:j.getAttribute("disabled")===null)&&(!j.parentNode.disabled||!f.nodeName(j.parentNode,"optgroup"))){b=f(j).val();if(g)return b;d.push(b)}}if(g&&!d.length&&e.length)return f(e[c]).val();return d},set:function(a,b){var c=f.makeArray(b);f(a).find("option").each(function(){this.selected=f.inArray(f(this).val(),c)>=0}),c.length||(a.selectedIndex=-1);return c}}},attrFn:{val:!0,css:!0,html:!0,text:!0,data:!0,width:!0,height:!0,offset:!0},attrFix:{tabindex:"tabIndex"},attr:function(a,c,d,e){var g=a.nodeType;if(!a||g===3||g===8||g===2)return b;if(e&&c in f.attrFn)return f(a)[c](d);if(!("getAttribute"in a))return f.prop(a,c,d);var h,i,j=g!==1||!f.isXMLDoc(a);c=j&&f.attrFix[c]||c,i=f.attrHooks[c],i||(!t.test(c)||typeof d!="boolean"&&d!==b&&d.toLowerCase()!==c.toLowerCase()?v&&(f.nodeName(a,"form")||u.test(c))&&(i=v):i=w);if(d!==b){if(d===null){f.removeAttr(a,c);return b}if(i&&"set"in i&&j&&(h=i.set(a,d,c))!==b)return h;a.setAttribute(c,""+d);return d}if(i&&"get"in i&&j)return i.get(a,c);h=a.getAttribute(c);return h===null?b:h},removeAttr:function(a,b){var c;a.nodeType===1&&(b=f.attrFix[b]||b,f.support.getSetAttribute?a.removeAttribute(b):(f.attr(a,b,""),a.removeAttributeNode(a.getAttributeNode(b))),t.test(b)&&(c=f.propFix[b]||b)in a&&(a[c]=!1))},attrHooks:{type:{set:function(a,b){if(q.test(a.nodeName)&&a.parentNode)f.error("type property can't be changed");else if(!f.support.radioValue&&b==="radio"&&f.nodeName(a,"input")){var c=a.value;a.setAttribute("type",b),c&&(a.value=c);return b}}},tabIndex:{get:function(a){var c=a.getAttributeNode("tabIndex");return c&&c.specified?parseInt(c.value,10):r.test(a.nodeName)||s.test(a.nodeName)&&a.href?0:b}}},propFix:{tabindex:"tabIndex",readonly:"readOnly","for":"htmlFor","class":"className",maxlength:"maxLength",cellspacing:"cellSpacing",cellpadding:"cellPadding",rowspan:"rowSpan",colspan:"colSpan",usemap:"useMap",frameborder:"frameBorder",contenteditable:"contentEditable"},prop:function(a,c,d){var e=a.nodeType;if(!a||e===3||e===8||e===2)return b;var g,h,i=e!==1||!f.isXMLDoc(a);c=i&&f.propFix[c]||c,h=f.propHooks[c];return d!==b?h&&"set"in h&&(g=h.set(a,d,c))!==b?g:a[c]=d:h&&"get"in h&&(g=h.get(a,c))!==b?g:a[c]},propHooks:{}}),w={get:function(a,c){return a[f.propFix[c]||c]?c.toLowerCase():b},set:function(a,b,c){var d;b===!1?f.removeAttr(a,c):(d=f.propFix[c]||c,d in a&&(a[d]=b),a.setAttribute(c,c.toLowerCase()));return c}},f.attrHooks.value={get:function(a,b){if(v&&f.nodeName(a,"button"))return v.get(a,b);return a.value},set:function(a,b,c){if(v&&f.nodeName(a,"button"))return v.set(a,b,c);a.value=b}},f.support.getSetAttribute||(f.attrFix=f.propFix,v=f.attrHooks.name=f.valHooks.button={get:function(a,c){var d;d=a.getAttributeNode(c);return d&&d.nodeValue!==""?d.nodeValue:b},set:function(a,b,c){var d=a.getAttributeNode(c);if(d){d.nodeValue=b;return b}}},f.each(["width","height"],function(a,b){f.attrHooks[b]=f.extend(f.attrHooks[b],{set:function(a,c){if(c===""){a.setAttribute(b,"auto");return c}}})})),f.support.hrefNormalized||f.each(["href","src","width","height"],function(a,c){f.attrHooks[c]=f.extend(f.attrHooks[c],{get:function(a){var d=a.getAttribute(c,2);return d===null?b:d}})}),f.support.style||(f.attrHooks.style={get:function(a){return a.style.cssText.toLowerCase()||b},set:function(a,b){return a.style.cssText=""+b}}),f.support.optSelected||(f.propHooks.selected=f.extend(f.propHooks.selected,{get:function(a){var b=a.parentNode;b&&(b.selectedIndex,b.parentNode&&b.parentNode.selectedIndex)}})),f.support.checkOn||f.each(["radio","checkbox"],function(){f.valHooks[this]={get:function(a){return a.getAttribute("value")===null?"on":a.value}}}),f.each(["radio","checkbox"],function(){f.valHooks[this]=f.extend(f.valHooks[this],{set:function(a,b){if(f.isArray(b))return a.checked=f.inArray(f(a).val(),b)>=0}})});var x=Object.prototype.hasOwnProperty,y=/\.(.*)$/,z=/^(?:textarea|input|select)$/i,A=/\./g,B=/ /g,C=/[^\w\s.|`]/g,D=function(a){return a.replace(C,"\\$&")};f.event={add:function(a,c,d,e){if(a.nodeType!==3&&a.nodeType!==8){if(d===!1)d=E;else if(!d)return;var g,h;d.handler&&(g=d,d=g.handler),d.guid||(d.guid=f.guid++);var i=f._data(a);if(!i)return;var j=i.events,k=i.handle;j||(i.events=j={}),k||(i.handle=k=function(a){return typeof f!="undefined"&&(!a||f.event.triggered!==a.type)?f.event.handle.apply(k.elem,arguments):b}),k.elem=a,c=c.split(" ");var l,m=0,n;while(l=c[m++]){h=g?f.extend({},g):{handler:d,data:e},l.indexOf(".")>-1?(n=l.split("."),l=n.shift(),h.namespace=n.slice(0).sort().join(".")):(n=[],h.namespace=""),h.type=l,h.guid||(h.guid=d.guid);var o=j[l],p=f.event.special[l]||{};if(!o){o=j[l]=[];if(!p.setup||p.setup.call(a,e,n,k)===!1)a.addEventListener?a.addEventListener(l,k,!1):a.attachEvent&&a.attachEvent("on"+l,k)}p.add&&(p.add.call(a,h),h.handler.guid||(h.handler.guid=d.guid)),o.push(h),f.event.global[l]=!0}a=null}},global:{},remove:function(a,c,d,e){if(a.nodeType!==3&&a.nodeType!==8){d===!1&&(d=E);var g,h,i,j,k=0,l,m,n,o,p,q,r,s=f.hasData(a)&&f._data(a),t=s&&s.events;if(!s||!t)return;c&&c.type&&(d=c.handler,c=c.type);if(!c||typeof c=="string"&&c.charAt(0)==="."){c=c||"";for(h in t)f.event.remove(a,h+c);return}c=c.split(" ");while(h=c[k++]){r=h,q=null,l=h.indexOf(".")<0,m=[],l||(m=h.split("."),h=m.shift(),n=new RegExp("(^|\\.)"+f.map(m.slice(0).sort(),D).join("\\.(?:.*\\.)?")+"(\\.|$)")),p=t[h];if(!p)continue;if(!d){for(j=0;j<p.length;j++){q=p[j];if(l||n.test(q.namespace))f.event.remove(a,r,q.handler,j),p.splice(j--,1)}continue}o=f.event.special[h]||{};for(j=e||0;j<p.length;j++){q=p[j];if(d.guid===q.guid){if(l||n.test(q.namespace))e==null&&p.splice(j--,1),o.remove&&o.remove.call(a,q);if(e!=null)break}}if(p.length===0||e!=null&&p.length===1)(!o.teardown||o.teardown.call(a,m)===!1)&&f.removeEvent(a,h,s.handle),g=null,delete t[h]}if(f.isEmptyObject(t)){var u=s.handle;u&&(u.elem=null),delete s.events,delete s.handle,f.isEmptyObject(s)&&f.removeData(a,b,!0)}}},customEvent:{getData:!0,setData:!0,changeData:!0},trigger:function(c,d,e,g){var h=c.type||c,i=[],j;h.indexOf("!")>=0&&(h=h.slice(0,-1),j=!0),h.indexOf(".")>=0&&(i=h.split("."),h=i.shift(),i.sort());if(!!e&&!f.event.customEvent[h]||!!f.event.global[h]){c=typeof c=="object"?c[f.expando]?c:new f.Event(h,c):new f.Event(h),c.type=h,c.exclusive=j,c.namespace=i.join("."),c.namespace_re=new RegExp("(^|\\.)"+i.join("\\.(?:.*\\.)?")+"(\\.|$)");if(g||!e)c.preventDefault(),c.stopPropagation();if(!e){f.each(f.cache,function(){var a=f.expando,b=this[a];b&&b.events&&b.events[h]&&f.event.trigger(c,d,b.handle.elem
)});return}if(e.nodeType===3||e.nodeType===8)return;c.result=b,c.target=e,d=d?f.makeArray(d):[],d.unshift(c);var k=e,l=h.indexOf(":")<0?"on"+h:"";do{var m=f._data(k,"handle");c.currentTarget=k,m&&m.apply(k,d),l&&f.acceptData(k)&&k[l]&&k[l].apply(k,d)===!1&&(c.result=!1,c.preventDefault()),k=k.parentNode||k.ownerDocument||k===c.target.ownerDocument&&a}while(k&&!c.isPropagationStopped());if(!c.isDefaultPrevented()){var n,o=f.event.special[h]||{};if((!o._default||o._default.call(e.ownerDocument,c)===!1)&&(h!=="click"||!f.nodeName(e,"a"))&&f.acceptData(e)){try{l&&e[h]&&(n=e[l],n&&(e[l]=null),f.event.triggered=h,e[h]())}catch(p){}n&&(e[l]=n),f.event.triggered=b}}return c.result}},handle:function(c){c=f.event.fix(c||a.event);var d=((f._data(this,"events")||{})[c.type]||[]).slice(0),e=!c.exclusive&&!c.namespace,g=Array.prototype.slice.call(arguments,0);g[0]=c,c.currentTarget=this;for(var h=0,i=d.length;h<i;h++){var j=d[h];if(e||c.namespace_re.test(j.namespace)){c.handler=j.handler,c.data=j.data,c.handleObj=j;var k=j.handler.apply(this,g);k!==b&&(c.result=k,k===!1&&(c.preventDefault(),c.stopPropagation()));if(c.isImmediatePropagationStopped())break}}return c.result},props:"altKey attrChange attrName bubbles button cancelable charCode clientX clientY ctrlKey currentTarget data detail eventPhase fromElement handler keyCode layerX layerY metaKey newValue offsetX offsetY pageX pageY prevValue relatedNode relatedTarget screenX screenY shiftKey srcElement target toElement view wheelDelta which".split(" "),fix:function(a){if(a[f.expando])return a;var d=a;a=f.Event(d);for(var e=this.props.length,g;e;)g=this.props[--e],a[g]=d[g];a.target||(a.target=a.srcElement||c),a.target.nodeType===3&&(a.target=a.target.parentNode),!a.relatedTarget&&a.fromElement&&(a.relatedTarget=a.fromElement===a.target?a.toElement:a.fromElement);if(a.pageX==null&&a.clientX!=null){var h=a.target.ownerDocument||c,i=h.documentElement,j=h.body;a.pageX=a.clientX+(i&&i.scrollLeft||j&&j.scrollLeft||0)-(i&&i.clientLeft||j&&j.clientLeft||0),a.pageY=a.clientY+(i&&i.scrollTop||j&&j.scrollTop||0)-(i&&i.clientTop||j&&j.clientTop||0)}a.which==null&&(a.charCode!=null||a.keyCode!=null)&&(a.which=a.charCode!=null?a.charCode:a.keyCode),!a.metaKey&&a.ctrlKey&&(a.metaKey=a.ctrlKey),!a.which&&a.button!==b&&(a.which=a.button&1?1:a.button&2?3:a.button&4?2:0);return a},guid:1e8,proxy:f.proxy,special:{ready:{setup:f.bindReady,teardown:f.noop},live:{add:function(a){f.event.add(this,O(a.origType,a.selector),f.extend({},a,{handler:N,guid:a.handler.guid}))},remove:function(a){f.event.remove(this,O(a.origType,a.selector),a)}},beforeunload:{setup:function(a,b,c){f.isWindow(this)&&(this.onbeforeunload=c)},teardown:function(a,b){this.onbeforeunload===b&&(this.onbeforeunload=null)}}}},f.removeEvent=c.removeEventListener?function(a,b,c){a.removeEventListener&&a.removeEventListener(b,c,!1)}:function(a,b,c){a.detachEvent&&a.detachEvent("on"+b,c)},f.Event=function(a,b){if(!this.preventDefault)return new f.Event(a,b);a&&a.type?(this.originalEvent=a,this.type=a.type,this.isDefaultPrevented=a.defaultPrevented||a.returnValue===!1||a.getPreventDefault&&a.getPreventDefault()?F:E):this.type=a,b&&f.extend(this,b),this.timeStamp=f.now(),this[f.expando]=!0},f.Event.prototype={preventDefault:function(){this.isDefaultPrevented=F;var a=this.originalEvent;!a||(a.preventDefault?a.preventDefault():a.returnValue=!1)},stopPropagation:function(){this.isPropagationStopped=F;var a=this.originalEvent;!a||(a.stopPropagation&&a.stopPropagation(),a.cancelBubble=!0)},stopImmediatePropagation:function(){this.isImmediatePropagationStopped=F,this.stopPropagation()},isDefaultPrevented:E,isPropagationStopped:E,isImmediatePropagationStopped:E};var G=function(a){var b=a.relatedTarget;a.type=a.data;try{if(b&&b!==c&&!b.parentNode)return;while(b&&b!==this)b=b.parentNode;b!==this&&f.event.handle.apply(this,arguments)}catch(d){}},H=function(a){a.type=a.data,f.event.handle.apply(this,arguments)};f.each({mouseenter:"mouseover",mouseleave:"mouseout"},function(a,b){f.event.special[a]={setup:function(c){f.event.add(this,b,c&&c.selector?H:G,a)},teardown:function(a){f.event.remove(this,b,a&&a.selector?H:G)}}}),f.support.submitBubbles||(f.event.special.submit={setup:function(a,b){if(!f.nodeName(this,"form"))f.event.add(this,"click.specialSubmit",function(a){var b=a.target,c=b.type;(c==="submit"||c==="image")&&f(b).closest("form").length&&L("submit",this,arguments)}),f.event.add(this,"keypress.specialSubmit",function(a){var b=a.target,c=b.type;(c==="text"||c==="password")&&f(b).closest("form").length&&a.keyCode===13&&L("submit",this,arguments)});else return!1},teardown:function(a){f.event.remove(this,".specialSubmit")}});if(!f.support.changeBubbles){var I,J=function(a){var b=a.type,c=a.value;b==="radio"||b==="checkbox"?c=a.checked:b==="select-multiple"?c=a.selectedIndex>-1?f.map(a.options,function(a){return a.selected}).join("-"):"":f.nodeName(a,"select")&&(c=a.selectedIndex);return c},K=function(c){var d=c.target,e,g;if(!!z.test(d.nodeName)&&!d.readOnly){e=f._data(d,"_change_data"),g=J(d),(c.type!=="focusout"||d.type!=="radio")&&f._data(d,"_change_data",g);if(e===b||g===e)return;if(e!=null||g)c.type="change",c.liveFired=b,f.event.trigger(c,arguments[1],d)}};f.event.special.change={filters:{focusout:K,beforedeactivate:K,click:function(a){var b=a.target,c=f.nodeName(b,"input")?b.type:"";(c==="radio"||c==="checkbox"||f.nodeName(b,"select"))&&K.call(this,a)},keydown:function(a){var b=a.target,c=f.nodeName(b,"input")?b.type:"";(a.keyCode===13&&!f.nodeName(b,"textarea")||a.keyCode===32&&(c==="checkbox"||c==="radio")||c==="select-multiple")&&K.call(this,a)},beforeactivate:function(a){var b=a.target;f._data(b,"_change_data",J(b))}},setup:function(a,b){if(this.type==="file")return!1;for(var c in I)f.event.add(this,c+".specialChange",I[c]);return z.test(this.nodeName)},teardown:function(a){f.event.remove(this,".specialChange");return z.test(this.nodeName)}},I=f.event.special.change.filters,I.focus=I.beforeactivate}f.support.focusinBubbles||f.each({focus:"focusin",blur:"focusout"},function(a,b){function e(a){var c=f.event.fix(a);c.type=b,c.originalEvent={},f.event.trigger(c,null,c.target),c.isDefaultPrevented()&&a.preventDefault()}var d=0;f.event.special[b]={setup:function(){d++===0&&c.addEventListener(a,e,!0)},teardown:function(){--d===0&&c.removeEventListener(a,e,!0)}}}),f.each(["bind","one"],function(a,c){f.fn[c]=function(a,d,e){var g;if(typeof a=="object"){for(var h in a)this[c](h,d,a[h],e);return this}if(arguments.length===2||d===!1)e=d,d=b;c==="one"?(g=function(a){f(this).unbind(a,g);return e.apply(this,arguments)},g.guid=e.guid||f.guid++):g=e;if(a==="unload"&&c!=="one")this.one(a,d,e);else for(var i=0,j=this.length;i<j;i++)f.event.add(this[i],a,g,d);return this}}),f.fn.extend({unbind:function(a,b){if(typeof a=="object"&&!a.preventDefault)for(var c in a)this.unbind(c,a[c]);else for(var d=0,e=this.length;d<e;d++)f.event.remove(this[d],a,b);return this},delegate:function(a,b,c,d){return this.live(b,c,d,a)},undelegate:function(a,b,c){return arguments.length===0?this.unbind("live"):this.die(b,null,c,a)},trigger:function(a,b){return this.each(function(){f.event.trigger(a,b,this)})},triggerHandler:function(a,b){if(this[0])return f.event.trigger(a,b,this[0],!0)},toggle:function(a){var b=arguments,c=a.guid||f.guid++,d=0,e=function(c){var e=(f.data(this,"lastToggle"+a.guid)||0)%d;f.data(this,"lastToggle"+a.guid,e+1),c.preventDefault();return b[e].apply(this,arguments)||!1};e.guid=c;while(d<b.length)b[d++].guid=c;return this.click(e)},hover:function(a,b){return this.mouseenter(a).mouseleave(b||a)}});var M={focus:"focusin",blur:"focusout",mouseenter:"mouseover",mouseleave:"mouseout"};f.each(["live","die"],function(a,c){f.fn[c]=function(a,d,e,g){var h,i=0,j,k,l,m=g||this.selector,n=g?this:f(this.context);if(typeof a=="object"&&!a.preventDefault){for(var o in a)n[c](o,d,a[o],m);return this}if(c==="die"&&!a&&g&&g.charAt(0)==="."){n.unbind(g);return this}if(d===!1||f.isFunction(d))e=d||E,d=b;a=(a||"").split(" ");while((h=a[i++])!=null){j=y.exec(h),k="",j&&(k=j[0],h=h.replace(y,""));if(h==="hover"){a.push("mouseenter"+k,"mouseleave"+k);continue}l=h,M[h]?(a.push(M[h]+k),h=h+k):h=(M[h]||h)+k;if(c==="live")for(var p=0,q=n.length;p<q;p++)f.event.add(n[p],"live."+O(h,m),{data:d,selector:m,handler:e,origType:h,origHandler:e,preType:l});else n.unbind("live."+O(h,m),e)}return this}}),f.each("blur focus focusin focusout load resize scroll unload click dblclick mousedown mouseup mousemove mouseover mouseout mouseenter mouseleave change select submit keydown keypress keyup error".split(" "),function(a,b){f.fn[b]=function(a,c){c==null&&(c=a,a=null);return arguments.length>0?this.bind(b,a,c):this.trigger(b)},f.attrFn&&(f.attrFn[b]=!0)}),function(){function u(a,b,c,d,e,f){for(var g=0,h=d.length;g<h;g++){var i=d[g];if(i){var j=!1;i=i[a];while(i){if(i.sizcache===c){j=d[i.sizset];break}if(i.nodeType===1){f||(i.sizcache=c,i.sizset=g);if(typeof b!="string"){if(i===b){j=!0;break}}else if(k.filter(b,[i]).length>0){j=i;break}}i=i[a]}d[g]=j}}}function t(a,b,c,d,e,f){for(var g=0,h=d.length;g<h;g++){var i=d[g];if(i){var j=!1;i=i[a];while(i){if(i.sizcache===c){j=d[i.sizset];break}i.nodeType===1&&!f&&(i.sizcache=c,i.sizset=g);if(i.nodeName.toLowerCase()===b){j=i;break}i=i[a]}d[g]=j}}}var a=/((?:\((?:\([^()]+\)|[^()]+)+\)|\[(?:\[[^\[\]]*\]|['"][^'"]*['"]|[^\[\]'"]+)+\]|\\.|[^ >+~,(\[\\]+)+|[>+~])(\s*,\s*)?((?:.|\r|\n)*)/g,d=0,e=Object.prototype.toString,g=!1,h=!0,i=/\\/g,j=/\W/;[0,0].sort(function(){h=!1;return 0});var k=function(b,d,f,g){f=f||[],d=d||c;var h=d;if(d.nodeType!==1&&d.nodeType!==9)return[];if(!b||typeof b!="string")return f;var i,j,n,o,q,r,s,t,u=!0,w=k.isXML(d),x=[],y=b;do{a.exec(""),i=a.exec(y);if(i){y=i[3],x.push(i[1]);if(i[2]){o=i[3];break}}}while(i);if(x.length>1&&m.exec(b))if(x.length===2&&l.relative[x[0]])j=v(x[0]+x[1],d);else{j=l.relative[x[0]]?[d]:k(x.shift(),d);while(x.length)b=x.shift(),l.relative[b]&&(b+=x.shift()),j=v(b,j)}else{!g&&x.length>1&&d.nodeType===9&&!w&&l.match.ID.test(x[0])&&!l.match.ID.test(x[x.length-1])&&(q=k.find(x.shift(),d,w),d=q.expr?k.filter(q.expr,q.set)[0]:q.set[0]);if(d){q=g?{expr:x.pop(),set:p(g)}:k.find(x.pop(),x.length===1&&(x[0]==="~"||x[0]==="+")&&d.parentNode?d.parentNode:d,w),j=q.expr?k.filter(q.expr,q.set):q.set,x.length>0?n=p(j):u=!1;while(x.length)r=x.pop(),s=r,l.relative[r]?s=x.pop():r="",s==null&&(s=d),l.relative[r](n,s,w)}else n=x=[]}n||(n=j),n||k.error(r||b);if(e.call(n)==="[object Array]")if(!u)f.push.apply(f,n);else if(d&&d.nodeType===1)for(t=0;n[t]!=null;t++)n[t]&&(n[t]===!0||n[t].nodeType===1&&k.contains(d,n[t]))&&f.push(j[t]);else for(t=0;n[t]!=null;t++)n[t]&&n[t].nodeType===1&&f.push(j[t]);else p(n,f);o&&(k(o,h,f,g),k.uniqueSort(f));return f};k.uniqueSort=function(a){if(r){g=h,a.sort(r);if(g)for(var b=1;b<a.length;b++)a[b]===a[b-1]&&a.splice(b--,1)}return a},k.matches=function(a,b){return k(a,null,null,b)},k.matchesSelector=function(a,b){return k(b,null,null,[a]).length>0},k.find=function(a,b,c){var d;if(!a)return[];for(var e=0,f=l.order.length;e<f;e++){var g,h=l.order[e];if(g=l.leftMatch[h].exec(a)){var j=g[1];g.splice(1,1);if(j.substr(j.length-1)!=="\\"){g[1]=(g[1]||"").replace(i,""),d=l.find[h](g,b,c);if(d!=null){a=a.replace(l.match[h],"");break}}}}d||(d=typeof b.getElementsByTagName!="undefined"?b.getElementsByTagName("*"):[]);return{set:d,expr:a}},k.filter=function(a,c,d,e){var f,g,h=a,i=[],j=c,m=c&&c[0]&&k.isXML(c[0]);while(a&&c.length){for(var n in l.filter)if((f=l.leftMatch[n].exec(a))!=null&&f[2]){var o,p,q=l.filter[n],r=f[1];g=!1,f.splice(1,1);if(r.substr(r.length-1)==="\\")continue;j===i&&(i=[]);if(l.preFilter[n]){f=l.preFilter[n](f,j,d,i,e,m);if(!f)g=o=!0;else if(f===!0)continue}if(f)for(var s=0;(p=j[s])!=null;s++)if(p){o=q(p,f,s,j);var t=e^!!o;d&&o!=null?t?g=!0:j[s]=!1:t&&(i.push(p),g=!0)}if(o!==b){d||(j=i),a=a.replace(l.match[n],"");if(!g)return[];break}}if(a===h)if(g==null)k.error(a);else break;h=a}return j},k.error=function(a){throw"Syntax error, unrecognized expression: "+a};var l=k.selectors={order:["ID","NAME","TAG"],match:{ID:/#((?:[\w\u00c0-\uFFFF\-]|\\.)+)/,CLASS:/\.((?:[\w\u00c0-\uFFFF\-]|\\.)+)/,NAME:/\[name=['"]*((?:[\w\u00c0-\uFFFF\-]|\\.)+)['"]*\]/,ATTR:/\[\s*((?:[\w\u00c0-\uFFFF\-]|\\.)+)\s*(?:(\S?=)\s*(?:(['"])(.*?)\3|(#?(?:[\w\u00c0-\uFFFF\-]|\\.)*)|)|)\s*\]/,TAG:/^((?:[\w\u00c0-\uFFFF\*\-]|\\.)+)/,CHILD:/:(only|nth|last|first)-child(?:\(\s*(even|odd|(?:[+\-]?\d+|(?:[+\-]?\d*)?n\s*(?:[+\-]\s*\d+)?))\s*\))?/,POS:/:(nth|eq|gt|lt|first|last|even|odd)(?:\((\d*)\))?(?=[^\-]|$)/,PSEUDO:/:((?:[\w\u00c0-\uFFFF\-]|\\.)+)(?:\((['"]?)((?:\([^\)]+\)|[^\(\)]*)+)\2\))?/},leftMatch:{},attrMap:{"class":"className","for":"htmlFor"},attrHandle:{href:function(a){return a.getAttribute("href")},type:function(a){return a.getAttribute("type")}},relative:{"+":function(a,b){var c=typeof b=="string",d=c&&!j.test(b),e=c&&!d;d&&(b=b.toLowerCase());for(var f=0,g=a.length,h;f<g;f++)if(h=a[f]){while((h=h.previousSibling)&&h.nodeType!==1);a[f]=e||h&&h.nodeName.toLowerCase()===b?h||!1:h===b}e&&k.filter(b,a,!0)},">":function(a,b){var c,d=typeof b=="string",e=0,f=a.length;if(d&&!j.test(b)){b=b.toLowerCase();for(;e<f;e++){c=a[e];if(c){var g=c.parentNode;a[e]=g.nodeName.toLowerCase()===b?g:!1}}}else{for(;e<f;e++)c=a[e],c&&(a[e]=d?c.parentNode:c.parentNode===b);d&&k.filter(b,a,!0)}},"":function(a,b,c){var e,f=d++,g=u;typeof b=="string"&&!j.test(b)&&(b=b.toLowerCase(),e=b,g=t),g("parentNode",b,f,a,e,c)},"~":function(a,b,c){var e,f=d++,g=u;typeof b=="string"&&!j.test(b)&&(b=b.toLowerCase(),e=b,g=t),g("previousSibling",b,f,a,e,c)}},find:{ID:function(a,b,c){if(typeof b.getElementById!="undefined"&&!c){var d=b.getElementById(a[1]);return d&&d.parentNode?[d]:[]}},NAME:function(a,b){if(typeof b.getElementsByName!="undefined"){var c=[],d=b.getElementsByName(a[1]);for(var e=0,f=d.length;e<f;e++)d[e].getAttribute("name")===a[1]&&c.push(d[e]);return c.length===0?null:c}},TAG:function(a,b){if(typeof b.getElementsByTagName!="undefined")return b.getElementsByTagName(a[1])}},preFilter:{CLASS:function(a,b,c,d,e,f){a=" "+a[1].replace(i,"")+" ";if(f)return a;for(var g=0,h;(h=b[g])!=null;g++)h&&(e^(h.className&&(" "+h.className+" ").replace(/[\t\n\r]/g," ").indexOf(a)>=0)?c||d.push(h):c&&(b[g]=!1));return!1},ID:function(a){return a[1].replace(i,"")},TAG:function(a,b){return a[1].replace(i,"").toLowerCase()},CHILD:function(a){if(a[1]==="nth"){a[2]||k.error(a[0]),a[2]=a[2].replace(/^\+|\s*/g,"");var b=/(-?)(\d*)(?:n([+\-]?\d*))?/.exec(a[2]==="even"&&"2n"||a[2]==="odd"&&"2n+1"||!/\D/.test(a[2])&&"0n+"+a[2]||a[2]);a[2]=b[1]+(b[2]||1)-0,a[3]=b[3]-0}else a[2]&&k.error(a[0]);a[0]=d++;return a},ATTR:function(a,b,c,d,e,f){var g=a[1]=a[1].replace(i,"");!f&&l.attrMap[g]&&(a[1]=l.attrMap[g]),a[4]=(a[4]||a[5]||"").replace(i,""),a[2]==="~="&&(a[4]=" "+a[4]+" ");return a},PSEUDO:function(b,c,d,e,f){if(b[1]==="not")if((a.exec(b[3])||"").length>1||/^\w/.test(b[3]))b[3]=k(b[3],null,null,c);else{var g=k.filter(b[3],c,d,!0^f);d||e.push.apply(e,g);return!1}else if(l.match.POS.test(b[0])||l.match.CHILD.test(b[0]))return!0;return b},POS:function(a){a.unshift(!0);return a}},filters:{enabled:function(a){return a.disabled===!1&&a.type!=="hidden"},disabled:function(a){return a.disabled===!0},checked:function(a){return a.checked===!0},selected:function(a){a.parentNode&&a.parentNode.selectedIndex;return a.selected===!0},parent:function(a){return!!a.firstChild},empty:function(a){return!a.firstChild},has:function(a,b,c){return!!k(c[3],a).length},header:function(a){return/h\d/i.test(a.nodeName)},text:function(a){var b=a.getAttribute("type"),c=a.type;return a.nodeName.toLowerCase()==="input"&&"text"===c&&(b===c||b===null)},radio:function(a){return a.nodeName.toLowerCase()==="input"&&"radio"===a.type},checkbox:function(a){return a.nodeName.toLowerCase()==="input"&&"checkbox"===a.type},file:function(a){return a.nodeName.toLowerCase()==="input"&&"file"===a.type},password:function(a){return a.nodeName.toLowerCase()==="input"&&"password"===a.type},submit:function(a){var b=a.nodeName.toLowerCase();return(b==="input"||b==="button")&&"submit"===a.type},image:function(a){return a.nodeName.toLowerCase()==="input"&&"image"===a.type},reset:function(a){var b=a.nodeName.toLowerCase();return(b==="input"||b==="button")&&"reset"===a.type},button:function(a){var b=a.nodeName.toLowerCase();return b==="input"&&"button"===a.type||b==="button"},input:function(a){return/input|select|textarea|button/i.test(a.nodeName)},focus:function(a){return a===a.ownerDocument.activeElement}},setFilters:{first:function(a,b){return b===0},last:function(a,b,c,d){return b===d.length-1},even:function(a,b){return b%2===0},odd:function(a,b){return b%2===1},lt:function(a,b,c){return b<c[3]-0},gt:function(a,b,c){return b>c[3]-0},nth:function(a,b,c){return c[3]-0===b},eq:function(a,b,c){return c[3]-0===b}},filter:{PSEUDO:function(a,b,c,d){var e=b[1],f=l.filters[e];if(f)return f(a,c,b,d);if(e==="contains")return(a.textContent||a.innerText||k.getText([a])||"").indexOf(b[3])>=0;if(e==="not"){var g=b[3];for(var h=0,i=g.length;h<i;h++)if(g[h]===a)return!1;return!0}k.error(e)},CHILD:function(a,b){var c=b[1],d=a;switch(c){case"only":case"first":while(d=d.previousSibling)if(d.nodeType===1)return!1;if(c==="first")return!0;d=a;case"last":while(d=d.nextSibling)if(d.nodeType===1)return!1;return!0;case"nth":var e=b[2],f=b[3];if(e===1&&f===0)return!0;var g=b[0],h=a.parentNode;if(h&&(h.sizcache!==g||!a.nodeIndex)){var i=0;for(d=h.firstChild;d;d=d.nextSibling)d.nodeType===1&&(d.nodeIndex=++i);h.sizcache=g}var j=a.nodeIndex-f;return e===0?j===0:j%e===0&&j/e>=0}},ID:function(a,b){return a.nodeType===1&&a.getAttribute("id")===b},TAG:function(a,b){return b==="*"&&a.nodeType===1||a.nodeName.toLowerCase()===b},CLASS:function(a,b){return(" "+(a.className||a.getAttribute("class"))+" ").indexOf(b)>-1},ATTR:function(a,b){var c=b[1],d=l.attrHandle[c]?l.attrHandle[c](a):a[c]!=null?a[c]:a.getAttribute(c),e=d+"",f=b[2],g=b[4];return d==null?f==="!=":f==="="?e===g:f==="*="?e.indexOf(g)>=0:f==="~="?(" "+e+" ").indexOf(g)>=0:g?f==="!="?e!==g:f==="^="?e.indexOf(g)===0:f==="$="?e.substr(e.length-g.length)===g:f==="|="?e===g||e.substr(0,g.length+1)===g+"-":!1:e&&d!==!1},POS:function(a,b,c,d){var e=b[2],f=l.setFilters[e];if(f)return f(a,c,b,d)}}},m=l.match.POS,n=function(a,b){return"\\"+(b-0+1)};for(var o in l.match)l.match[o]=new RegExp(l.match[o].source+/(?![^\[]*\])(?![^\(]*\))/.source),l.leftMatch[o]=new RegExp(/(^(?:.|\r|\n)*?)/.source+l.match[o].source.replace(/\\(\d+)/g,n));var p=function(a,b){a=Array.prototype.slice.call(a,0);if(b){b.push.apply(b,a);return b}return a};try{Array.prototype.slice.call(c.documentElement.childNodes,0)[0].nodeType}catch(q){p=function(a,b){var c=0,d=b||[];if(e.call(a)==="[object Array]")Array.prototype.push.apply(d,a);else if(typeof a.length=="number")for(var f=a.length;c<f;c++)d.push(a[c]);else for(;a[c];c++)d.push(a[c]);return d}}var r,s;c.documentElement.compareDocumentPosition?r=function(a,b){if(a===b){g=!0;return 0}if(!a.compareDocumentPosition||!b.compareDocumentPosition)return a.compareDocumentPosition?-1:1;return a.compareDocumentPosition(b)&4?-1:1}:(r=function(a,b){if(a===b){g=!0;return 0}if(a.sourceIndex&&b.sourceIndex)return a.sourceIndex-b.sourceIndex;var c,d,e=[],f=[],h=a.parentNode,i=b.parentNode,j=h;if(h===i)return s(a,b);if(!h)return-1;if(!i)return 1;while(j)e.unshift(j),j=j.parentNode;j=i;while(j)f.unshift(j),j=j.parentNode;c=e.length,d=f.length;for(var k=0;k<c&&k<d;k++)if(e[k]!==f[k])return s(e[k],f[k]);return k===c?s(a,f[k],-1):s(e[k],b,1)},s=function(a,b,c){if(a===b)return c;var d=a.nextSibling;while(d){if(d===b)return-1;d=d.nextSibling}return 1}),k.getText=function(a){var b="",c;for(var d=0;a[d];d++)c=a[d],c.nodeType===3||c.nodeType===4?b+=c.nodeValue:c.nodeType!==8&&(b+=k.getText(c.childNodes));return b},function(){var a=c.createElement("div"),d="script"+(new Date).getTime(),e=c.documentElement;a.innerHTML="<a name='"+d+"'/>",e.insertBefore(a,e.firstChild),c.getElementById(d)&&(l.find.ID=function(a,c,d){if(typeof c.getElementById!="undefined"&&!d){var e=c.getElementById(a[1]);return e?e.id===a[1]||typeof e.getAttributeNode!="undefined"&&e.getAttributeNode("id").nodeValue===a[1]?[e]:b:[]}},l.filter.ID=function(a,b){var c=typeof a.getAttributeNode!="undefined"&&a.getAttributeNode("id");return a.nodeType===1&&c&&c.nodeValue===b}),e.removeChild(a),e=a=null}(),function(){var a=c.createElement("div");a.appendChild(c.createComment("")),a.getElementsByTagName("*").length>0&&(l.find.TAG=function(a,b){var c=b.getElementsByTagName(a[1]);if(a[1]==="*"){var d=[];for(var e=0;c[e];e++)c[e].nodeType===1&&d.push(c[e]);c=d}return c}),a.innerHTML="<a href='#'></a>",a.firstChild&&typeof a.firstChild.getAttribute!="undefined"&&a.firstChild.getAttribute("href")!=="#"&&(l.attrHandle.href=function(a){return a.getAttribute("href",2)}),a=null}(),c.querySelectorAll&&function(){var a=k,b=c.createElement("div"),d="__sizzle__";b.innerHTML="<p class='TEST'></p>";if(!b.querySelectorAll||b.querySelectorAll(".TEST").length!==0){k=function(b,e,f,g){e=e||c;if(!g&&!k.isXML(e)){var h=/^(\w+$)|^\.([\w\-]+$)|^#([\w\-]+$)/.exec(b);if(h&&(e.nodeType===1||e.nodeType===9)){if(h[1])return p(e.getElementsByTagName(b),f);if(h[2]&&l.find.CLASS&&e.getElementsByClassName)return p(e.getElementsByClassName(h[2]),f)}if(e.nodeType===9){if(b==="body"&&e.body)return p([e.body],f);if(h&&h[3]){var i=e.getElementById(h[3]);if(!i||!i.parentNode)return p([],f);if(i.id===h[3])return p([i],f)}try{return p(e.querySelectorAll(b),f)}catch(j){}}else if(e.nodeType===1&&e.nodeName.toLowerCase()!=="object"){var m=e,n=e.getAttribute("id"),o=n||d,q=e.parentNode,r=/^\s*[+~]/.test(b);n?o=o.replace(/'/g,"\\$&"):e.setAttribute("id",o),r&&q&&(e=e.parentNode);try{if(!r||q)return p(e.querySelectorAll("[id='"+o+"'] "+b),f)}catch(s){}finally{n||m.removeAttribute("id")}}}return a(b,e,f,g)};for(var e in a)k[e]=a[e];b=null}}(),function(){var a=c.documentElement,b=a.matchesSelector||a.mozMatchesSelector||a.webkitMatchesSelector||a.msMatchesSelector;if(b){var d=!b.call(c.createElement("div"),"div"),e=!1;try{b.call(c.documentElement,"[test!='']:sizzle")}catch(f){e=!0}k.matchesSelector=function(a,c){c=c.replace(/\=\s*([^'"\]]*)\s*\]/g,"='$1']");if(!k.isXML(a))try{if(e||!l.match.PSEUDO.test(c)&&!/!=/.test(c)){var f=b.call(a,c);if(f||!d||a.document&&a.document.nodeType!==11)return f}}catch(g){}return k(c,null,null,[a]).length>0}}}(),function(){var a=c.createElement("div");a.innerHTML="<div class='test e'></div><div class='test'></div>";if(!!a.getElementsByClassName&&a.getElementsByClassName("e").length!==0){a.lastChild.className="e";if(a.getElementsByClassName("e").length===1)return;l.order.splice(1,0,"CLASS"),l.find.CLASS=function(a,b,c){if(typeof b.getElementsByClassName!="undefined"&&!c)return b.getElementsByClassName(a[1])},a=null}}(),c.documentElement.contains?k.contains=function(a,b){return a!==b&&(a.contains?a.contains(b):!0)}:c.documentElement.compareDocumentPosition?k.contains=function(a,b){return!!(a.compareDocumentPosition(b)&16)}:k.contains=function(){return!1},k.isXML=function(a){var b=(a?a.ownerDocument||a:0).documentElement;return b?b.nodeName!=="HTML":!1};var v=function(a,b){var c,d=[],e="",f=b.nodeType?[b]:b;while(c=l.match.PSEUDO.exec(a))e+=c[0],a=a.replace(l.match.PSEUDO,"");a=l.relative[a]?a+"*":a;for(var g=0,h=f.length;g<h;g++)k(a,f[g],d);return k.filter(e,d)};f.find=k,f.expr=k.selectors,f.expr[":"]=f.expr.filters,f.unique=k.uniqueSort,f.text=k.getText,f.isXMLDoc=k.isXML,f.contains=k.contains}();var P=/Until$/,Q=/^(?:parents|prevUntil|prevAll)/,R=/,/,S=/^.[^:#\[\.,]*$/,T=Array.prototype.slice,U=f.expr.match.POS,V={children:!0,contents:!0,next:!0,prev:!0};f.fn.extend({find:function(a){var b=this,c,d;if(typeof a!="string")return f(a).filter(function(){for(c=0,d=b.length;c<d;c++)if(f.contains(b[c],this))return!0});var e=this.pushStack("","find",a),g,h,i;for(c=0,d=this.length;c<d;c++){g=e.length,f.find(a,this[c],e);if(c>0)for(h=g;h<e.length;h++)for(i=0;i<g;i++)if(e[i]===e[h]){e.splice(h--,1);break}}return e},has:function(a){var b=f(a);return this.filter(function(){for(var a=0,c=b.length;a<c;a++)if(f.contains(this,b[a]))return!0})},not:function(a){return this.pushStack(X(this,a,!1),"not",a)},filter:function(a){return this.pushStack(X(this,a,!0),"filter",a)},is:function(a){return!!a&&(typeof a=="string"?f.filter(a,this).length>0:this.filter(a).length>0)},closest:function(a,b){var c=[],d,e,g=this[0];if(f.isArray(a)){var h,i,j={},k=1;if(g&&a.length){for(d=0,e=a.length;d<e;d++)i=a[d],j[i]||(j[i]=U.test(i)?f(i,b||this.context):i);while(g&&g.ownerDocument&&g!==b){for(i in j)h=j[i],(h.jquery?h.index(g)>-1:f(g).is(h))&&c.push({selector:i,elem:g,level:k});g=g.parentNode,k++}}return c}var l=U.test(a)||typeof a!="string"?f(a,b||this.context):0;for(d=0,e=this.length;d<e;d++){g=this[d];while(g){if(l?l.index(g)>-1:f.find.matchesSelector(g,a)){c.push(g);break}g=g.parentNode;if(!g||!g.ownerDocument||g===b||g.nodeType===11)break}}c=c.length>1?f.unique(c):c;return this.pushStack(c,"closest",a)},index:function(a){if(!a||typeof a=="string")return f.inArray(this[0],a?f(a):this.parent().children());return f.inArray(a.jquery?a[0]:a,this)},add:function(a,b){var c=typeof a=="string"?f(a,b):f.makeArray(a&&a.nodeType?[a]:a),d=f.merge(this.get(),c);return this.pushStack(W(c[0])||W(d[0])?d:f.unique(d))},andSelf:function(){return this.add(this.prevObject)}}),f.each({parent:function(a){var b=a.parentNode;return b&&b.nodeType!==11?b:null},parents:function(a){return f.dir(a,"parentNode")},parentsUntil:function(a,b,c){return f.dir(a,"parentNode",c)},next:function(a){return f.nth(a,2,"nextSibling")},prev:function(a){return f.nth(a,2,"previousSibling")},nextAll:function(a){return f.dir(a,"nextSibling")},prevAll:function(a){return f.dir(a,"previousSibling")},nextUntil:function(a,b,c){return f.dir(a,"nextSibling",c)},prevUntil:function(a,b,c){return f.dir(a,"previousSibling",c)},siblings:function(a){return f.sibling(a.parentNode.firstChild,a)},children:function(a){return f.sibling(a.firstChild)},contents:function(a){return f.nodeName(a,"iframe")?a.contentDocument||a.contentWindow.document:f.makeArray(a.childNodes)}},function(a,b){f.fn[a]=function(c,d){var e=f.map(this,b,c),g=T.call(arguments);P.test(a)||(d=c),d&&typeof d=="string"&&(e=f.filter(d,e)),e=this.length>1&&!V[a]?f.unique(e):e,(this.length>1||R.test(d))&&Q.test(a)&&(e=e.reverse());return this.pushStack(e,a,g.join(","))}}),f.extend({filter:function(a,b,c){c&&(a=":not("+a+")");return b.length===1?f.find.matchesSelector(b[0],a)?[b[0]]:[]:f.find.matches(a,b)},dir:function(a,c,d){var e=[],g=a[c];while(g&&g.nodeType!==9&&(d===b||g.nodeType!==1||!f(g).is(d)))g.nodeType===1&&e.push(g),g=g[c];return e},nth:function(a,b,c,d){b=b||1;var e=0;for(;a;a=a[c])if(a.nodeType===1&&++e===b)break;return a},sibling:function(a,b){var c=[];for(;a;a=a.nextSibling)a.nodeType===1&&a!==b&&c.push(a);return c}});var Y=/ jQuery\d+="(?:\d+|null)"/g,Z=/^\s+/,$=/<(?!area|br|col|embed|hr|img|input|link|meta|param)(([\w:]+)[^>]*)\/>/ig,_=/<([\w:]+)/,ba=/<tbody/i,bb=/<|&#?\w+;/,bc=/<(?:script|object|embed|option|style)/i,bd=/checked\s*(?:[^=]|=\s*.checked.)/i,be=/\/(java|ecma)script/i,bf=/^\s*<!(?:\[CDATA\[|\-\-)/,bg={option:[1,"<select multiple='multiple'>","</select>"],legend:[1,"<fieldset>","</fieldset>"],thead:[1,"<table>","</table>"],tr:[2,"<table><tbody>","</tbody></table>"],td:[3,"<table><tbody><tr>","</tr></tbody></table>"],col:[2,"<table><tbody></tbody><colgroup>","</colgroup></table>"],area:[1,"<map>","</map>"],_default:[0,"",""]};bg.optgroup=bg.option,bg.tbody=bg.tfoot=bg.colgroup=bg.caption=bg.thead,bg.th=bg.td,f.support.htmlSerialize||(bg._default=[1,"div<div>","</div>"]),f.fn.extend({text:function(a){if(f.isFunction(a))return this.each(function(b){var c=f(this);c.text(a.call(this,b,c.text()))});if(typeof a!="object"&&a!==b)return this.empty().append((this[0]&&this[0].ownerDocument||c).createTextNode(a));return f.text(this)},wrapAll:function(a){if(f.isFunction(a))return this.each(function(b){f(this).wrapAll(a.call(this,b))});if(this[0]){var b=f(a,this[0].ownerDocument).eq(0).clone(!0);this[0].parentNode&&b.insertBefore(this[0]),b.map(function(){var a=this;while(a.firstChild&&a.firstChild.nodeType===1)a=a.firstChild;return a}).append(this)}return this},wrapInner:function(a){if(f.isFunction(a))return this.each(function(b){f(this).wrapInner(a.call(this,b))});return this.each(function(){var b=f(this),c=b.contents();c.length?c.wrapAll(a):b.append(a)})},wrap:function(a){return this.each(function(){f(this).wrapAll(a)})},unwrap:function(){return this.parent().each(function(){f.nodeName(this,"body")||f(this).replaceWith(this.childNodes)}).end()},append:function(){return this.domManip(arguments,!0,function(a){this.nodeType===1&&this.appendChild(a)})},prepend:function(){return this.domManip(arguments,!0,function(a){this.nodeType===1&&this.insertBefore(a,this.firstChild)})},before:function(){if(this[0]&&this[0].parentNode)return this.domManip(arguments,!1,function(a){this.parentNode.insertBefore(a,this)});if(arguments.length){var a=f(arguments[0]);a.push.apply(a,this.toArray());return this.pushStack(a,"before",arguments)}},after:function(){if(this[0]&&this[0].parentNode)return this.domManip(arguments,!1,function(a){this.parentNode.insertBefore(a,this.nextSibling)});if(arguments.length){var a=this.pushStack(this,"after",arguments);a.push.apply(a,f(arguments[0]).toArray());return a}},remove:function(a,b){for(var c=0,d;(d=this[c])!=null;c++)if(!a||f.filter(a,[d]).length)!b&&d.nodeType===1&&(f.cleanData(d.getElementsByTagName("*")),f.cleanData([d])),d.parentNode&&d.parentNode.removeChild(d);return this},empty:function(){for(var a=0,b;(b=this[a])!=null;a++){b.nodeType===1&&f.cleanData(b.getElementsByTagName("*"));while(b.firstChild)b.removeChild(b.firstChild)}return this},clone:function(a,b){a=a==null?!1:a,b=b==null?a:b;return this.map(function(){return f.clone(this,a,b)})},html:function(a){if(a===b)return this[0]&&this[0].nodeType===1?this[0].innerHTML.replace(Y,""):null;if(typeof a=="string"&&!bc.test(a)&&(f.support.leadingWhitespace||!Z.test(a))&&!bg[(_.exec(a)||["",""])[1].toLowerCase()]){a=a.replace($,"<$1></$2>");try{for(var c=0,d=this.length;c<d;c++)this[c].nodeType===1&&(f.cleanData(this[c].getElementsByTagName("*")),this[c].innerHTML=a)}catch(e){this.empty().append(a)}}else f.isFunction(a)?this.each(function(b){var c=f(this);c.html(a.call(this,b,c.html()))}):this.empty().append(a);return this},replaceWith:function(a){if(this[0]&&this[0].parentNode){if(f.isFunction(a))return this.each(function(b){var c=f(this),d=c.html();c.replaceWith(a.call(this,b,d))});typeof a!="string"&&(a=f(a).detach());return this.each(function(){var b=this.nextSibling,c=this.parentNode;f(this).remove(),b?f(b).before(a):f(c).append(a)})}return this.length?this.pushStack(f(f.isFunction(a)?a():a),"replaceWith",a):this},detach:function(a){return this.remove(a,!0)},domManip:function(a,c,d){var e,g,h,i,j=a[0],k=[];if(!f.support.checkClone&&arguments.length===3&&typeof j=="string"&&bd.test(j))return this.each(function(){f(this).domManip(a,c,d,!0)});if(f.isFunction(j))return this.each(function(e){var g=f(this);a[0]=j.call(this,e,c?g.html():b),g.domManip(a,c,d)});if(this[0]){i=j&&j.parentNode,f.support.parentNode&&i&&i.nodeType===11&&i.childNodes.length===this.length?e={fragment:i}:e=f.buildFragment(a,this,k),h=e.fragment,h.childNodes.length===1?g=h=h.firstChild:g=h.firstChild;if(g){c=c&&f.nodeName(g,"tr");for(var l=0,m=this.length,n=m-1;l<m;l++)d.call(c?bh(this[l],g):this[l],e.cacheable||m>1&&l<n?f.clone(h,!0,!0):h)}k.length&&f.each(k,bn)}return this}}),f.buildFragment=function(a,b,d){var e,g,h,i=b&&b[0]?b[0].ownerDocument||b[0]:c;a.length===1&&typeof a[0]=="string"&&a[0].length<512&&i===c&&a[0].charAt(0)==="<"&&!bc.test(a[0])&&(f.support.checkClone||!bd.test(a[0]))&&(g=!0,h=f.fragments[a[0]],h&&h!==1&&(e=h)),e||(e=i.createDocumentFragment(),f.clean(a,i,e,d)),g&&(f.fragments[a[0]]=h?e:1);return{fragment:e,cacheable:g}},f.fragments={},f.each({appendTo:"append",prependTo:"prepend",insertBefore:"before",insertAfter:"after",replaceAll:"replaceWith"},function(a,b){f.fn[a]=function(c){var d=[],e=f(c),g=this.length===1&&this[0].parentNode;if(g&&g.nodeType===11&&g.childNodes.length===1&&e.length===1){e[b](this[0]);return this}for(var h=0,i=e.length;h<i;h++){var j=(h>0?this.clone(!0):this).get();f(e[h])[b](j),d=d.concat(j)}return this.pushStack(d,a,e.selector)}}),f.extend({clone:function(a,b,c){var d=a.cloneNode(!0),e,g,h;if((!f.support.noCloneEvent||!f.support.noCloneChecked)&&(a.nodeType===1||a.nodeType===11)&&!f.isXMLDoc(a)){bj(a,d),e=bk(a),g=bk(d);for(h=0;e[h];++h)bj(e[h],g[h])}if(b){bi(a,d);if(c){e=bk(a),g=bk(d);for(h=0;e[h];++h)bi(e[h],g[h])}}return d},clean:function(a,b,d,e){var g;b=b||c,typeof b.createElement=="undefined"&&(b=b.ownerDocument||
b[0]&&b[0].ownerDocument||c);var h=[],i;for(var j=0,k;(k=a[j])!=null;j++){typeof k=="number"&&(k+="");if(!k)continue;if(typeof k=="string")if(!bb.test(k))k=b.createTextNode(k);else{k=k.replace($,"<$1></$2>");var l=(_.exec(k)||["",""])[1].toLowerCase(),m=bg[l]||bg._default,n=m[0],o=b.createElement("div");o.innerHTML=m[1]+k+m[2];while(n--)o=o.lastChild;if(!f.support.tbody){var p=ba.test(k),q=l==="table"&&!p?o.firstChild&&o.firstChild.childNodes:m[1]==="<table>"&&!p?o.childNodes:[];for(i=q.length-1;i>=0;--i)f.nodeName(q[i],"tbody")&&!q[i].childNodes.length&&q[i].parentNode.removeChild(q[i])}!f.support.leadingWhitespace&&Z.test(k)&&o.insertBefore(b.createTextNode(Z.exec(k)[0]),o.firstChild),k=o.childNodes}var r;if(!f.support.appendChecked)if(k[0]&&typeof (r=k.length)=="number")for(i=0;i<r;i++)bm(k[i]);else bm(k);k.nodeType?h.push(k):h=f.merge(h,k)}if(d){g=function(a){return!a.type||be.test(a.type)};for(j=0;h[j];j++)if(e&&f.nodeName(h[j],"script")&&(!h[j].type||h[j].type.toLowerCase()==="text/javascript"))e.push(h[j].parentNode?h[j].parentNode.removeChild(h[j]):h[j]);else{if(h[j].nodeType===1){var s=f.grep(h[j].getElementsByTagName("script"),g);h.splice.apply(h,[j+1,0].concat(s))}d.appendChild(h[j])}}return h},cleanData:function(a){var b,c,d=f.cache,e=f.expando,g=f.event.special,h=f.support.deleteExpando;for(var i=0,j;(j=a[i])!=null;i++){if(j.nodeName&&f.noData[j.nodeName.toLowerCase()])continue;c=j[f.expando];if(c){b=d[c]&&d[c][e];if(b&&b.events){for(var k in b.events)g[k]?f.event.remove(j,k):f.removeEvent(j,k,b.handle);b.handle&&(b.handle.elem=null)}h?delete j[f.expando]:j.removeAttribute&&j.removeAttribute(f.expando),delete d[c]}}}});var bo=/alpha\([^)]*\)/i,bp=/opacity=([^)]*)/,bq=/-([a-z])/ig,br=/([A-Z]|^ms)/g,bs=/^-?\d+(?:px)?$/i,bt=/^-?\d/,bu=/^[+\-]=/,bv=/[^+\-\.\de]+/g,bw={position:"absolute",visibility:"hidden",display:"block"},bx=["Left","Right"],by=["Top","Bottom"],bz,bA,bB,bC=function(a,b){return b.toUpperCase()};f.fn.css=function(a,c){if(arguments.length===2&&c===b)return this;return f.access(this,a,c,!0,function(a,c,d){return d!==b?f.style(a,c,d):f.css(a,c)})},f.extend({cssHooks:{opacity:{get:function(a,b){if(b){var c=bz(a,"opacity","opacity");return c===""?"1":c}return a.style.opacity}}},cssNumber:{zIndex:!0,fontWeight:!0,opacity:!0,zoom:!0,lineHeight:!0,widows:!0,orphans:!0},cssProps:{"float":f.support.cssFloat?"cssFloat":"styleFloat"},style:function(a,c,d,e){if(!!a&&a.nodeType!==3&&a.nodeType!==8&&!!a.style){var g,h,i=f.camelCase(c),j=a.style,k=f.cssHooks[i];c=f.cssProps[i]||i;if(d===b){if(k&&"get"in k&&(g=k.get(a,!1,e))!==b)return g;return j[c]}h=typeof d;if(h==="number"&&isNaN(d)||d==null)return;h==="string"&&bu.test(d)&&(d=+d.replace(bv,"")+parseFloat(f.css(a,c))),h==="number"&&!f.cssNumber[i]&&(d+="px");if(!k||!("set"in k)||(d=k.set(a,d))!==b)try{j[c]=d}catch(l){}}},css:function(a,c,d){var e,g;c=f.camelCase(c),g=f.cssHooks[c],c=f.cssProps[c]||c,c==="cssFloat"&&(c="float");if(g&&"get"in g&&(e=g.get(a,!0,d))!==b)return e;if(bz)return bz(a,c)},swap:function(a,b,c){var d={};for(var e in b)d[e]=a.style[e],a.style[e]=b[e];c.call(a);for(e in b)a.style[e]=d[e]},camelCase:function(a){return a.replace(bq,bC)}}),f.curCSS=f.css,f.each(["height","width"],function(a,b){f.cssHooks[b]={get:function(a,c,d){var e;if(c){a.offsetWidth!==0?e=bD(a,b,d):f.swap(a,bw,function(){e=bD(a,b,d)});if(e<=0){e=bz(a,b,b),e==="0px"&&bB&&(e=bB(a,b,b));if(e!=null)return e===""||e==="auto"?"0px":e}if(e<0||e==null){e=a.style[b];return e===""||e==="auto"?"0px":e}return typeof e=="string"?e:e+"px"}},set:function(a,b){if(!bs.test(b))return b;b=parseFloat(b);if(b>=0)return b+"px"}}}),f.support.opacity||(f.cssHooks.opacity={get:function(a,b){return bp.test((b&&a.currentStyle?a.currentStyle.filter:a.style.filter)||"")?parseFloat(RegExp.$1)/100+"":b?"1":""},set:function(a,b){var c=a.style,d=a.currentStyle;c.zoom=1;var e=f.isNaN(b)?"":"alpha(opacity="+b*100+")",g=d&&d.filter||c.filter||"";c.filter=bo.test(g)?g.replace(bo,e):g+" "+e}}),f(function(){f.support.reliableMarginRight||(f.cssHooks.marginRight={get:function(a,b){var c;f.swap(a,{display:"inline-block"},function(){b?c=bz(a,"margin-right","marginRight"):c=a.style.marginRight});return c}})}),c.defaultView&&c.defaultView.getComputedStyle&&(bA=function(a,c){var d,e,g;c=c.replace(br,"-$1").toLowerCase();if(!(e=a.ownerDocument.defaultView))return b;if(g=e.getComputedStyle(a,null))d=g.getPropertyValue(c),d===""&&!f.contains(a.ownerDocument.documentElement,a)&&(d=f.style(a,c));return d}),c.documentElement.currentStyle&&(bB=function(a,b){var c,d=a.currentStyle&&a.currentStyle[b],e=a.runtimeStyle&&a.runtimeStyle[b],f=a.style;!bs.test(d)&&bt.test(d)&&(c=f.left,e&&(a.runtimeStyle.left=a.currentStyle.left),f.left=b==="fontSize"?"1em":d||0,d=f.pixelLeft+"px",f.left=c,e&&(a.runtimeStyle.left=e));return d===""?"auto":d}),bz=bA||bB,f.expr&&f.expr.filters&&(f.expr.filters.hidden=function(a){var b=a.offsetWidth,c=a.offsetHeight;return b===0&&c===0||!f.support.reliableHiddenOffsets&&(a.style.display||f.css(a,"display"))==="none"},f.expr.filters.visible=function(a){return!f.expr.filters.hidden(a)});var bE=/%20/g,bF=/\[\]$/,bG=/\r?\n/g,bH=/#.*$/,bI=/^(.*?):[ \t]*([^\r\n]*)\r?$/mg,bJ=/^(?:color|date|datetime|email|hidden|month|number|password|range|search|tel|text|time|url|week)$/i,bK=/^(?:about|app|app\-storage|.+\-extension|file|widget):$/,bL=/^(?:GET|HEAD)$/,bM=/^\/\//,bN=/\?/,bO=/<script\b[^<]*(?:(?!<\/script>)<[^<]*)*<\/script>/gi,bP=/^(?:select|textarea)/i,bQ=/\s+/,bR=/([?&])_=[^&]*/,bS=/^([\w\+\.\-]+:)(?:\/\/([^\/?#:]*)(?::(\d+))?)?/,bT=f.fn.load,bU={},bV={},bW,bX;try{bW=e.href}catch(bY){bW=c.createElement("a"),bW.href="",bW=bW.href}bX=bS.exec(bW.toLowerCase())||[],f.fn.extend({load:function(a,c,d){if(typeof a!="string"&&bT)return bT.apply(this,arguments);if(!this.length)return this;var e=a.indexOf(" ");if(e>=0){var g=a.slice(e,a.length);a=a.slice(0,e)}var h="GET";c&&(f.isFunction(c)?(d=c,c=b):typeof c=="object"&&(c=f.param(c,f.ajaxSettings.traditional),h="POST"));var i=this;f.ajax({url:a,type:h,dataType:"html",data:c,complete:function(a,b,c){c=a.responseText,a.isResolved()&&(a.done(function(a){c=a}),i.html(g?f("<div>").append(c.replace(bO,"")).find(g):c)),d&&i.each(d,[c,b,a])}});return this},serialize:function(){return f.param(this.serializeArray())},serializeArray:function(){return this.map(function(){return this.elements?f.makeArray(this.elements):this}).filter(function(){return this.name&&!this.disabled&&(this.checked||bP.test(this.nodeName)||bJ.test(this.type))}).map(function(a,b){var c=f(this).val();return c==null?null:f.isArray(c)?f.map(c,function(a,c){return{name:b.name,value:a.replace(bG,"\r\n")}}):{name:b.name,value:c.replace(bG,"\r\n")}}).get()}}),f.each("ajaxStart ajaxStop ajaxComplete ajaxError ajaxSuccess ajaxSend".split(" "),function(a,b){f.fn[b]=function(a){return this.bind(b,a)}}),f.each(["get","post"],function(a,c){f[c]=function(a,d,e,g){f.isFunction(d)&&(g=g||e,e=d,d=b);return f.ajax({type:c,url:a,data:d,success:e,dataType:g})}}),f.extend({getScript:function(a,c){return f.get(a,b,c,"script")},getJSON:function(a,b,c){return f.get(a,b,c,"json")},ajaxSetup:function(a,b){b?f.extend(!0,a,f.ajaxSettings,b):(b=a,a=f.extend(!0,f.ajaxSettings,b));for(var c in{context:1,url:1})c in b?a[c]=b[c]:c in f.ajaxSettings&&(a[c]=f.ajaxSettings[c]);return a},ajaxSettings:{url:bW,isLocal:bK.test(bX[1]),global:!0,type:"GET",contentType:"application/x-www-form-urlencoded",processData:!0,async:!0,accepts:{xml:"application/xml, text/xml",html:"text/html",text:"text/plain",json:"application/json, text/javascript","*":"*/*"},contents:{xml:/xml/,html:/html/,json:/json/},responseFields:{xml:"responseXML",text:"responseText"},converters:{"* text":a.String,"text html":!0,"text json":f.parseJSON,"text xml":f.parseXML}},ajaxPrefilter:bZ(bU),ajaxTransport:bZ(bV),ajax:function(a,c){function w(a,c,l,m){if(s!==2){s=2,q&&clearTimeout(q),p=b,n=m||"",v.readyState=a?4:0;var o,r,u,w=l?ca(d,v,l):b,x,y;if(a>=200&&a<300||a===304){if(d.ifModified){if(x=v.getResponseHeader("Last-Modified"))f.lastModified[k]=x;if(y=v.getResponseHeader("Etag"))f.etag[k]=y}if(a===304)c="notmodified",o=!0;else try{r=cb(d,w),c="success",o=!0}catch(z){c="parsererror",u=z}}else{u=c;if(!c||a)c="error",a<0&&(a=0)}v.status=a,v.statusText=c,o?h.resolveWith(e,[r,c,v]):h.rejectWith(e,[v,c,u]),v.statusCode(j),j=b,t&&g.trigger("ajax"+(o?"Success":"Error"),[v,d,o?r:u]),i.resolveWith(e,[v,c]),t&&(g.trigger("ajaxComplete",[v,d]),--f.active||f.event.trigger("ajaxStop"))}}typeof a=="object"&&(c=a,a=b),c=c||{};var d=f.ajaxSetup({},c),e=d.context||d,g=e!==d&&(e.nodeType||e instanceof f)?f(e):f.event,h=f.Deferred(),i=f._Deferred(),j=d.statusCode||{},k,l={},m={},n,o,p,q,r,s=0,t,u,v={readyState:0,setRequestHeader:function(a,b){if(!s){var c=a.toLowerCase();a=m[c]=m[c]||a,l[a]=b}return this},getAllResponseHeaders:function(){return s===2?n:null},getResponseHeader:function(a){var c;if(s===2){if(!o){o={};while(c=bI.exec(n))o[c[1].toLowerCase()]=c[2]}c=o[a.toLowerCase()]}return c===b?null:c},overrideMimeType:function(a){s||(d.mimeType=a);return this},abort:function(a){a=a||"abort",p&&p.abort(a),w(0,a);return this}};h.promise(v),v.success=v.done,v.error=v.fail,v.complete=i.done,v.statusCode=function(a){if(a){var b;if(s<2)for(b in a)j[b]=[j[b],a[b]];else b=a[v.status],v.then(b,b)}return this},d.url=((a||d.url)+"").replace(bH,"").replace(bM,bX[1]+"//"),d.dataTypes=f.trim(d.dataType||"*").toLowerCase().split(bQ),d.crossDomain==null&&(r=bS.exec(d.url.toLowerCase()),d.crossDomain=!(!r||r[1]==bX[1]&&r[2]==bX[2]&&(r[3]||(r[1]==="http:"?80:443))==(bX[3]||(bX[1]==="http:"?80:443)))),d.data&&d.processData&&typeof d.data!="string"&&(d.data=f.param(d.data,d.traditional)),b$(bU,d,c,v);if(s===2)return!1;t=d.global,d.type=d.type.toUpperCase(),d.hasContent=!bL.test(d.type),t&&f.active++===0&&f.event.trigger("ajaxStart");if(!d.hasContent){d.data&&(d.url+=(bN.test(d.url)?"&":"?")+d.data),k=d.url;if(d.cache===!1){var x=f.now(),y=d.url.replace(bR,"$1_="+x);d.url=y+(y===d.url?(bN.test(d.url)?"&":"?")+"_="+x:"")}}(d.data&&d.hasContent&&d.contentType!==!1||c.contentType)&&v.setRequestHeader("Content-Type",d.contentType),d.ifModified&&(k=k||d.url,f.lastModified[k]&&v.setRequestHeader("If-Modified-Since",f.lastModified[k]),f.etag[k]&&v.setRequestHeader("If-None-Match",f.etag[k])),v.setRequestHeader("Accept",d.dataTypes[0]&&d.accepts[d.dataTypes[0]]?d.accepts[d.dataTypes[0]]+(d.dataTypes[0]!=="*"?", */*; q=0.01":""):d.accepts["*"]);for(u in d.headers)v.setRequestHeader(u,d.headers[u]);if(d.beforeSend&&(d.beforeSend.call(e,v,d)===!1||s===2)){v.abort();return!1}for(u in{success:1,error:1,complete:1})v[u](d[u]);p=b$(bV,d,c,v);if(!p)w(-1,"No Transport");else{v.readyState=1,t&&g.trigger("ajaxSend",[v,d]),d.async&&d.timeout>0&&(q=setTimeout(function(){v.abort("timeout")},d.timeout));try{s=1,p.send(l,w)}catch(z){status<2?w(-1,z):f.error(z)}}return v},param:function(a,c){var d=[],e=function(a,b){b=f.isFunction(b)?b():b,d[d.length]=encodeURIComponent(a)+"="+encodeURIComponent(b)};c===b&&(c=f.ajaxSettings.traditional);if(f.isArray(a)||a.jquery&&!f.isPlainObject(a))f.each(a,function(){e(this.name,this.value)});else for(var g in a)b_(g,a[g],c,e);return d.join("&").replace(bE,"+")}}),f.extend({active:0,lastModified:{},etag:{}});var cc=f.now(),cd=/(\=)\?(&|$)|\?\?/i;f.ajaxSetup({jsonp:"callback",jsonpCallback:function(){return f.expando+"_"+cc++}}),f.ajaxPrefilter("json jsonp",function(b,c,d){var e=b.contentType==="application/x-www-form-urlencoded"&&typeof b.data=="string";if(b.dataTypes[0]==="jsonp"||b.jsonp!==!1&&(cd.test(b.url)||e&&cd.test(b.data))){var g,h=b.jsonpCallback=f.isFunction(b.jsonpCallback)?b.jsonpCallback():b.jsonpCallback,i=a[h],j=b.url,k=b.data,l="$1"+h+"$2";b.jsonp!==!1&&(j=j.replace(cd,l),b.url===j&&(e&&(k=k.replace(cd,l)),b.data===k&&(j+=(/\?/.test(j)?"&":"?")+b.jsonp+"="+h))),b.url=j,b.data=k,a[h]=function(a){g=[a]},d.always(function(){a[h]=i,g&&f.isFunction(i)&&a[h](g[0])}),b.converters["script json"]=function(){g||f.error(h+" was not called");return g[0]},b.dataTypes[0]="json";return"script"}}),f.ajaxSetup({accepts:{script:"text/javascript, application/javascript, application/ecmascript, application/x-ecmascript"},contents:{script:/javascript|ecmascript/},converters:{"text script":function(a){f.globalEval(a);return a}}}),f.ajaxPrefilter("script",function(a){a.cache===b&&(a.cache=!1),a.crossDomain&&(a.type="GET",a.global=!1)}),f.ajaxTransport("script",function(a){if(a.crossDomain){var d,e=c.head||c.getElementsByTagName("head")[0]||c.documentElement;return{send:function(f,g){d=c.createElement("script"),d.async="async",a.scriptCharset&&(d.charset=a.scriptCharset),d.src=a.url,d.onload=d.onreadystatechange=function(a,c){if(c||!d.readyState||/loaded|complete/.test(d.readyState))d.onload=d.onreadystatechange=null,e&&d.parentNode&&e.removeChild(d),d=b,c||g(200,"success")},e.insertBefore(d,e.firstChild)},abort:function(){d&&d.onload(0,1)}}}});var ce=a.ActiveXObject?function(){for(var a in cg)cg[a](0,1)}:!1,cf=0,cg;f.ajaxSettings.xhr=a.ActiveXObject?function(){return!this.isLocal&&ch()||ci()}:ch,function(a){f.extend(f.support,{ajax:!!a,cors:!!a&&"withCredentials"in a})}(f.ajaxSettings.xhr()),f.support.ajax&&f.ajaxTransport(function(c){if(!c.crossDomain||f.support.cors){var d;return{send:function(e,g){var h=c.xhr(),i,j;c.username?h.open(c.type,c.url,c.async,c.username,c.password):h.open(c.type,c.url,c.async);if(c.xhrFields)for(j in c.xhrFields)h[j]=c.xhrFields[j];c.mimeType&&h.overrideMimeType&&h.overrideMimeType(c.mimeType),!c.crossDomain&&!e["X-Requested-With"]&&(e["X-Requested-With"]="XMLHttpRequest");try{for(j in e)h.setRequestHeader(j,e[j])}catch(k){}h.send(c.hasContent&&c.data||null),d=function(a,e){var j,k,l,m,n;try{if(d&&(e||h.readyState===4)){d=b,i&&(h.onreadystatechange=f.noop,ce&&delete cg[i]);if(e)h.readyState!==4&&h.abort();else{j=h.status,l=h.getAllResponseHeaders(),m={},n=h.responseXML,n&&n.documentElement&&(m.xml=n),m.text=h.responseText;try{k=h.statusText}catch(o){k=""}!j&&c.isLocal&&!c.crossDomain?j=m.text?200:404:j===1223&&(j=204)}}}catch(p){e||g(-1,p)}m&&g(j,k,m,l)},!c.async||h.readyState===4?d():(i=++cf,ce&&(cg||(cg={},f(a).unload(ce)),cg[i]=d),h.onreadystatechange=d)},abort:function(){d&&d(0,1)}}}});var cj={},ck,cl,cm=/^(?:toggle|show|hide)$/,cn=/^([+\-]=)?([\d+.\-]+)([a-z%]*)$/i,co,cp=[["height","marginTop","marginBottom","paddingTop","paddingBottom"],["width","marginLeft","marginRight","paddingLeft","paddingRight"],["opacity"]],cq,cr=a.webkitRequestAnimationFrame||a.mozRequestAnimationFrame||a.oRequestAnimationFrame;f.fn.extend({show:function(a,b,c){var d,e;if(a||a===0)return this.animate(cu("show",3),a,b,c);for(var g=0,h=this.length;g<h;g++)d=this[g],d.style&&(e=d.style.display,!f._data(d,"olddisplay")&&e==="none"&&(e=d.style.display=""),e===""&&f.css(d,"display")==="none"&&f._data(d,"olddisplay",cv(d.nodeName)));for(g=0;g<h;g++){d=this[g];if(d.style){e=d.style.display;if(e===""||e==="none")d.style.display=f._data(d,"olddisplay")||""}}return this},hide:function(a,b,c){if(a||a===0)return this.animate(cu("hide",3),a,b,c);for(var d=0,e=this.length;d<e;d++)if(this[d].style){var g=f.css(this[d],"display");g!=="none"&&!f._data(this[d],"olddisplay")&&f._data(this[d],"olddisplay",g)}for(d=0;d<e;d++)this[d].style&&(this[d].style.display="none");return this},_toggle:f.fn.toggle,toggle:function(a,b,c){var d=typeof a=="boolean";f.isFunction(a)&&f.isFunction(b)?this._toggle.apply(this,arguments):a==null||d?this.each(function(){var b=d?a:f(this).is(":hidden");f(this)[b?"show":"hide"]()}):this.animate(cu("toggle",3),a,b,c);return this},fadeTo:function(a,b,c,d){return this.filter(":hidden").css("opacity",0).show().end().animate({opacity:b},a,c,d)},animate:function(a,b,c,d){var e=f.speed(b,c,d);if(f.isEmptyObject(a))return this.each(e.complete,[!1]);a=f.extend({},a);return this[e.queue===!1?"each":"queue"](function(){e.queue===!1&&f._mark(this);var b=f.extend({},e),c=this.nodeType===1,d=c&&f(this).is(":hidden"),g,h,i,j,k,l,m,n,o;b.animatedProperties={};for(i in a){g=f.camelCase(i),i!==g&&(a[g]=a[i],delete a[i]),h=a[g],f.isArray(h)?(b.animatedProperties[g]=h[1],h=a[g]=h[0]):b.animatedProperties[g]=b.specialEasing&&b.specialEasing[g]||b.easing||"swing";if(h==="hide"&&d||h==="show"&&!d)return b.complete.call(this);c&&(g==="height"||g==="width")&&(b.overflow=[this.style.overflow,this.style.overflowX,this.style.overflowY],f.css(this,"display")==="inline"&&f.css(this,"float")==="none"&&(f.support.inlineBlockNeedsLayout?(j=cv(this.nodeName),j==="inline"?this.style.display="inline-block":(this.style.display="inline",this.style.zoom=1)):this.style.display="inline-block"))}b.overflow!=null&&(this.style.overflow="hidden");for(i in a)k=new f.fx(this,b,i),h=a[i],cm.test(h)?k[h==="toggle"?d?"show":"hide":h]():(l=cn.exec(h),m=k.cur(),l?(n=parseFloat(l[2]),o=l[3]||(f.cssNumber[i]?"":"px"),o!=="px"&&(f.style(this,i,(n||1)+o),m=(n||1)/k.cur()*m,f.style(this,i,m+o)),l[1]&&(n=(l[1]==="-="?-1:1)*n+m),k.custom(m,n,o)):k.custom(m,h,""));return!0})},stop:function(a,b){a&&this.queue([]),this.each(function(){var a=f.timers,c=a.length;b||f._unmark(!0,this);while(c--)a[c].elem===this&&(b&&a[c](!0),a.splice(c,1))}),b||this.dequeue();return this}}),f.each({slideDown:cu("show",1),slideUp:cu("hide",1),slideToggle:cu("toggle",1),fadeIn:{opacity:"show"},fadeOut:{opacity:"hide"},fadeToggle:{opacity:"toggle"}},function(a,b){f.fn[a]=function(a,c,d){return this.animate(b,a,c,d)}}),f.extend({speed:function(a,b,c){var d=a&&typeof a=="object"?f.extend({},a):{complete:c||!c&&b||f.isFunction(a)&&a,duration:a,easing:c&&b||b&&!f.isFunction(b)&&b};d.duration=f.fx.off?0:typeof d.duration=="number"?d.duration:d.duration in f.fx.speeds?f.fx.speeds[d.duration]:f.fx.speeds._default,d.old=d.complete,d.complete=function(a){d.queue!==!1?f.dequeue(this):a!==!1&&f._unmark(this),f.isFunction(d.old)&&d.old.call(this)};return d},easing:{linear:function(a,b,c,d){return c+d*a},swing:function(a,b,c,d){return(-Math.cos(a*Math.PI)/2+.5)*d+c}},timers:[],fx:function(a,b,c){this.options=b,this.elem=a,this.prop=c,b.orig=b.orig||{}}}),f.fx.prototype={update:function(){this.options.step&&this.options.step.call(this.elem,this.now,this),(f.fx.step[this.prop]||f.fx.step._default)(this)},cur:function(){if(this.elem[this.prop]!=null&&(!this.elem.style||this.elem.style[this.prop]==null))return this.elem[this.prop];var a,b=f.css(this.elem,this.prop);return isNaN(a=parseFloat(b))?!b||b==="auto"?0:b:a},custom:function(a,b,c){function h(a){return d.step(a)}var d=this,e=f.fx,g;this.startTime=cq||cs(),this.start=a,this.end=b,this.unit=c||this.unit||(f.cssNumber[this.prop]?"":"px"),this.now=this.start,this.pos=this.state=0,h.elem=this.elem,h()&&f.timers.push(h)&&!co&&(cr?(co=1,g=function(){co&&(cr(g),e.tick())},cr(g)):co=setInterval(e.tick,e.interval))},show:function(){this.options.orig[this.prop]=f.style(this.elem,this.prop),this.options.show=!0,this.custom(this.prop==="width"||this.prop==="height"?1:0,this.cur()),f(this.elem).show()},hide:function(){this.options.orig[this.prop]=f.style(this.elem,this.prop),this.options.hide=!0,this.custom(this.cur(),0)},step:function(a){var b=cq||cs(),c=!0,d=this.elem,e=this.options,g,h;if(a||b>=e.duration+this.startTime){this.now=this.end,this.pos=this.state=1,this.update(),e.animatedProperties[this.prop]=!0;for(g in e.animatedProperties)e.animatedProperties[g]!==!0&&(c=!1);if(c){e.overflow!=null&&!f.support.shrinkWrapBlocks&&f.each(["","X","Y"],function(a,b){d.style["overflow"+b]=e.overflow[a]}),e.hide&&f(d).hide();if(e.hide||e.show)for(var i in e.animatedProperties)f.style(d,i,e.orig[i]);e.complete.call(d)}return!1}e.duration==Infinity?this.now=b:(h=b-this.startTime,this.state=h/e.duration,this.pos=f.easing[e.animatedProperties[this.prop]](this.state,h,0,1,e.duration),this.now=this.start+(this.end-this.start)*this.pos),this.update();return!0}},f.extend(f.fx,{tick:function(){for(var a=f.timers,b=0;b<a.length;++b)a[b]()||a.splice(b--,1);a.length||f.fx.stop()},interval:13,stop:function(){clearInterval(co),co=null},speeds:{slow:600,fast:200,_default:400},step:{opacity:function(a){f.style(a.elem,"opacity",a.now)},_default:function(a){a.elem.style&&a.elem.style[a.prop]!=null?a.elem.style[a.prop]=(a.prop==="width"||a.prop==="height"?Math.max(0,a.now):a.now)+a.unit:a.elem[a.prop]=a.now}}}),f.expr&&f.expr.filters&&(f.expr.filters.animated=function(a){return f.grep(f.timers,function(b){return a===b.elem}).length});var cw=/^t(?:able|d|h)$/i,cx=/^(?:body|html)$/i;"getBoundingClientRect"in c.documentElement?f.fn.offset=function(a){var b=this[0],c;if(a)return this.each(function(b){f.offset.setOffset(this,a,b)});if(!b||!b.ownerDocument)return null;if(b===b.ownerDocument.body)return f.offset.bodyOffset(b);try{c=b.getBoundingClientRect()}catch(d){}var e=b.ownerDocument,g=e.documentElement;if(!c||!f.contains(g,b))return c?{top:c.top,left:c.left}:{top:0,left:0};var h=e.body,i=cy(e),j=g.clientTop||h.clientTop||0,k=g.clientLeft||h.clientLeft||0,l=i.pageYOffset||f.support.boxModel&&g.scrollTop||h.scrollTop,m=i.pageXOffset||f.support.boxModel&&g.scrollLeft||h.scrollLeft,n=c.top+l-j,o=c.left+m-k;return{top:n,left:o}}:f.fn.offset=function(a){var b=this[0];if(a)return this.each(function(b){f.offset.setOffset(this,a,b)});if(!b||!b.ownerDocument)return null;if(b===b.ownerDocument.body)return f.offset.bodyOffset(b);f.offset.initialize();var c,d=b.offsetParent,e=b,g=b.ownerDocument,h=g.documentElement,i=g.body,j=g.defaultView,k=j?j.getComputedStyle(b,null):b.currentStyle,l=b.offsetTop,m=b.offsetLeft;while((b=b.parentNode)&&b!==i&&b!==h){if(f.offset.supportsFixedPosition&&k.position==="fixed")break;c=j?j.getComputedStyle(b,null):b.currentStyle,l-=b.scrollTop,m-=b.scrollLeft,b===d&&(l+=b.offsetTop,m+=b.offsetLeft,f.offset.doesNotAddBorder&&(!f.offset.doesAddBorderForTableAndCells||!cw.test(b.nodeName))&&(l+=parseFloat(c.borderTopWidth)||0,m+=parseFloat(c.borderLeftWidth)||0),e=d,d=b.offsetParent),f.offset.subtractsBorderForOverflowNotVisible&&c.overflow!=="visible"&&(l+=parseFloat(c.borderTopWidth)||0,m+=parseFloat(c.borderLeftWidth)||0),k=c}if(k.position==="relative"||k.position==="static")l+=i.offsetTop,m+=i.offsetLeft;f.offset.supportsFixedPosition&&k.position==="fixed"&&(l+=Math.max(h.scrollTop,i.scrollTop),m+=Math.max(h.scrollLeft,i.scrollLeft));return{top:l,left:m}},f.offset={initialize:function(){var a=c.body,b=c.createElement("div"),d,e,g,h,i=parseFloat(f.css(a,"marginTop"))||0,j="<div style='position:absolute;top:0;left:0;margin:0;border:5px solid #000;padding:0;width:1px;height:1px;'><div></div></div><table style='position:absolute;top:0;left:0;margin:0;border:5px solid #000;padding:0;width:1px;height:1px;' cellpadding='0' cellspacing='0'><tr><td></td></tr></table>";f.extend(b.style,{position:"absolute",top:0,left:0,margin:0,border:0,width:"1px",height:"1px",visibility:"hidden"}),b.innerHTML=j,a.insertBefore(b,a.firstChild),d=b.firstChild,e=d.firstChild,h=d.nextSibling.firstChild.firstChild,this.doesNotAddBorder=e.offsetTop!==5,this.doesAddBorderForTableAndCells=h.offsetTop===5,e.style.position="fixed",e.style.top="20px",this.supportsFixedPosition=e.offsetTop===20||e.offsetTop===15,e.style.position=e.style.top="",d.style.overflow="hidden",d.style.position="relative",this.subtractsBorderForOverflowNotVisible=e.offsetTop===-5,this.doesNotIncludeMarginInBodyOffset=a.offsetTop!==i,a.removeChild(b),f.offset.initialize=f.noop},bodyOffset:function(a){var b=a.offsetTop,c=a.offsetLeft;f.offset.initialize(),f.offset.doesNotIncludeMarginInBodyOffset&&(b+=parseFloat(f.css(a,"marginTop"))||0,c+=parseFloat(f.css(a,"marginLeft"))||0);return{top:b,left:c}},setOffset:function(a,b,c){var d=f.css(a,"position");d==="static"&&(a.style.position="relative");var e=f(a),g=e.offset(),h=f.css(a,"top"),i=f.css(a,"left"),j=(d==="absolute"||d==="fixed")&&f.inArray("auto",[h,i])>-1,k={},l={},m,n;j?(l=e.position(),m=l.top,n=l.left):(m=parseFloat(h)||0,n=parseFloat(i)||0),f.isFunction(b)&&(b=b.call(a,c,g)),b.top!=null&&(k.top=b.top-g.top+m),b.left!=null&&(k.left=b.left-g.left+n),"using"in b?b.using.call(a,k):e.css(k)}},f.fn.extend({position:function(){if(!this[0])return null;var a=this[0],b=this.offsetParent(),c=this.offset(),d=cx.test(b[0].nodeName)?{top:0,left:0}:b.offset();c.top-=parseFloat(f.css(a,"marginTop"))||0,c.left-=parseFloat(f.css(a,"marginLeft"))||0,d.top+=parseFloat(f.css(b[0],"borderTopWidth"))||0,d.left+=parseFloat(f.css(b[0],"borderLeftWidth"))||0;return{top:c.top-d.top,left:c.left-d.left}},offsetParent:function(){return this.map(function(){var a=this.offsetParent||c.body;while(a&&!cx.test(a.nodeName)&&f.css(a,"position")==="static")a=a.offsetParent;return a})}}),f.each(["Left","Top"],function(a,c){var d="scroll"+c;f.fn[d]=function(c){var e,g;if(c===b){e=this[0];if(!e)return null;g=cy(e);return g?"pageXOffset"in g?g[a?"pageYOffset":"pageXOffset"]:f.support.boxModel&&g.document.documentElement[d]||g.document.body[d]:e[d]}return this.each(function(){g=cy(this),g?g.scrollTo(a?f(g).scrollLeft():c,a?c:f(g).scrollTop()):this[d]=c})}}),f.each(["Height","Width"],function(a,c){var d=c.toLowerCase();f.fn["inner"+c]=function(){return this[0]?parseFloat(f.css(this[0],d,"padding")):null},f.fn["outer"+c]=function(a){return this[0]?parseFloat(f.css(this[0],d,a?"margin":"border")):null},f.fn[d]=function(a){var e=this[0];if(!e)return a==null?null:this;if(f.isFunction(a))return this.each(function(b){var c=f(this);c[d](a.call(this,b,c[d]()))});if(f.isWindow(e)){var g=e.document.documentElement["client"+c];return e.document.compatMode==="CSS1Compat"&&g||e.document.body["client"+c]||g}if(e.nodeType===9)return Math.max(e.documentElement["client"+c],e.body["scroll"+c],e.documentElement["scroll"+c],e.body["offset"+c],e.documentElement["offset"+c]);if(a===b){var h=f.css(e,d),i=parseFloat(h);return f.isNaN(i)?h:i}return this.css(d,typeof a=="string"?a:a+"px")}}),a.jQuery=a.$=f})(window);

}
